/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdio.h>
#include <float.h>

#include "landscape.h"
#include "config.h"

landscape::landscape()
{
	int x, y;

	sel_sqr = 0;
	sel_tri = 0;
	first_d = 0;

	for (y = 0; y < 256; y++)
	{
		for (x = 0; x < 256; x++)
		{
			sqr[y][x].set(float(x * 16), float(y * 16), 128.0);
		}
	}
}

landscape::~landscape()
{
}

void landscape::draw(const bbox *box, const vector *dir)
{
	int x, xl, xh, y, yl, yh;
	texture tex;

	xl = int(floor(box->min.x() / 16.0));
	yl = int(floor(box->min.y() / 16.0));
	xh = int(ceil(box->max.x() / 16.0));
	yh = int(ceil(box->max.y() / 16.0));

	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh > 255) xh = 255;
	if (yh > 255) yh = 255;

	glDisable(GL_TEXTURE_2D);
	glColor3ub(127, 127, 127);

	for (y = yl; y <= yh; y++)
	{
		for (x = xl ; x <= xh; x++)
		{
			sqr[y][x].draw(this, dir);
		}
	}

	glColor3ub(255, 255, 255);
	glEnable(GL_TEXTURE_2D);

	tex.draw(this);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glDisable(GL_CULL_FACE);

	for (drawable *ptr = first_d; ptr; ptr = ptr->next_d)
	{
		ptr->tex.glbind();
		ptr->draw(this);
	}
	first_d = 0;

	glEnable(GL_CULL_FACE);
	glDisable(GL_BLEND);
}

void landscape::queue(drawable *tri)
{
	drawable **ptr;
	drawable *next;

	for (ptr = &first_d; (next = *ptr); ptr = &next->next_d)
	{
		if (next->dist < tri->dist)
			break;
	}

	tri->next_d = next;
	*ptr = tri;
}

int landscape::load(const char *file)
{
	FILE *fp;
	unsigned i, j;
	texture texs[256];
	char name[1024];

	sprintf(name, "maps/%s", file);

	if ( !(fp = fopen(name, "r")) )
		return -1;

	if ( fread(name, 1, 4, fp) != 4 ||
	     strncmp(name, "CJPL", 4) )
	{
		fclose(fp);
		return -1;
	}

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	for (j = 1; j <= i; j++)
	{
		if (texs[j].load(fp))
		{
			fclose(fp);
			return -1;
		}
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (sqr[j][i].load(fp, texs))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	fclose(fp);

	return 0;
}

int landscape::save(const char *file)
{
	FILE *fp;
	unsigned i, j;
	texture tex;
	char name[1024];

	sprintf(name, "maps/%s", file);

	if ( !(fp = fopen(name, "w")) )
		return -1;

	if (fwrite("CJPL", 1, 4, fp) != 4)
	{
		fclose(fp);
		return -1;
	}

	i = tex.reindex();
	if (fputc(i, fp) != int(i))
	{
		fclose(fp);
		return -1;
	}

	if (tex.first())
	{
		do
		{
			if (tex.index())
			{
				if (tex.save(fp))
				{
					fclose(fp);
					return -1;
				}
			}
		}
		while (tex.next());
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (sqr[j][i].save(fp))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	if (fclose(fp))
		return -1;

	return 0;
}

square *landscape::select_square(const vector *org, const vector *dir,
				 const bbox *box)
{
	int x, xl, xh, y, yl, yh;
	float sel_n = 0;

	sel_sqr = 0;
	sel_tri = 0;

	xl = int(floor(box->min.x() / 16.0));
	yl = int(floor(box->min.y() / 16.0));
	xh = int(ceil(box->max.x() / 16.0));
	yh = int(ceil(box->max.y() / 16.0));

	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh > 255) xh = 255;
	if (yh > 255) yh = 255;

	for (y = yl; y <= yh; y++)
	{
		for (x = xl ; x < xh; x++)
		{
			vector vec;
			float n;
			square *sq = &sqr[y][x];

			if ( (n = vec.isect_z_line(sq, org, dir)) > 0
			     && (!sel_sqr || n < sel_n) )
			{
				float xf = sq->x();
				float yf = sq->y();

				if (vec.x() >= xf &&
				    vec.x() <= xf + 16.0 &&
				    vec.y() >= yf &&
				    vec.y() <= yf + 16.0)
				{
					sel_sqr = sq;
					sel_n = n;
				}
			}
		}
	}

	return sel_sqr;
}

triangle *landscape::select_next_triangle()
{
	if (sel_sqr)
	{
		if (sel_tri) sel_tri = sel_tri->next;
		if (!sel_tri) sel_tri = sel_sqr->tris;
	}

	return sel_tri;
}

square::square()
{
	tris = 0;
}

square::~square()
{
	triangle *tri, *next;

	for (tri = tris; tri; tri = next)
	{
		next = tri->next;
		delete tri;
	}
}

void square::draw(landscape *land, const vector *dir)
{
	if (land->selected_square() != this)
	{
		glBegin(GL_LINE_LOOP);
		glvertex();
		set_x(x() + 16);
		glvertex();
		set_y(y() + 16);
		glvertex();
		set_x(x() - 16);
		glvertex();
		set_y(y() - 16);
		glEnd();

	}

	for (triangle *tri = tris; tri; tri = tri->next)
	{
		tri->queue(land, this, dir);
	}
}

void square::highlight(landscape *land)
{
	triangle *tri;

	glDisable(GL_TEXTURE_2D);

	glBegin(GL_LINE_LOOP);
	glvertex();
	set_x(x() + 16);
	glvertex();
	set_y(y() + 16);
	glvertex();
	set_x(x() - 16);
	glvertex();
	set_y(y() - 16);
	glEnd();

	if ( (tri = land->selected_triangle()) )
	{
		tri->highlight();
	}
}

int square::load(FILE *fp, texture *texs)
{
	unsigned i, j;
	triangle *tri, **ptr;

	if ( (i = fgetc(fp)) > 255)
		return -1;
	set_z(float(i));

	if ( (i = fgetc(fp)) > 255)
		return -1;

	ptr = &tris;

	for (j = 0; j < i; j++)
	{
		if ( !(tri = *ptr) )
		{
			*ptr = tri = new triangle;
		}

		if (tri->load(fp, texs))
			return -1;

		ptr = &tri->next;
	}

	while ( (tri = *ptr) )
	{
		*ptr = tri->next;
		delete tri;
	}

	return 0;
}

int square::save(FILE *fp)
{
	unsigned i;
	triangle *tri;
	
	i = (unsigned char)(z());

	if (fputc(i, fp) != int(i))
		return -1;

	i = triangles();
	if (fputc(i, fp) != int(i))
		return -1;

	for (tri = tris; tri; tri = tri->next)
	{
		if (tri->save(fp))
			return -1;
	}

	return 0;
}

unsigned square::triangles()
{
	unsigned i = 0;
	triangle *tri;

	for (tri = tris; tri; tri = tri->next)
	{
		i++;
	}

	return i;
}

void square::set_height(float f)
{
	if (f < 0.0) f = 0.0;
	if (f > 255.0) f = 255.0;
	set_z(rint(f));
}

void square::fix_vec(vector *vec)
{
	float i = vec->x() - x();
	float j = vec->y() - y();
	float k = vec->z() - z();

	if (i < 0.0) i = 0.0;
	if (i > 16.0) i = 16.0;

	if (j < 0.0) j = 0.0;
	if (j > 16.0) j = 16.0;

	if (k < -128.0) k = -128.0;
	if (k > 127.0) k = 127.0;

	vec->set(rint(i), rint(j), rint(k));
}

triangle *square::new_triangle(const vector *vec)
{
	triangle *tri = new triangle;

	tri->vtx[0].set(vec);
	tri->vtx[1].set(vec);
	tri->vtx[2].set(vec);

	tri->vtx[1].v = 1.0;
	tri->vtx[2].u = 1.0;

	tri->tex.get_default();
	tri->next = tris;

	return tris = tri;
}

void square::delete_triangle(triangle *old)
{
	triangle *tri, **ptr;

	for (ptr = &tris; (tri = *ptr); ptr = &tri->next)
	{
		if (tri == old)
		{
			*ptr = tri->next;
			break;
		}
	}

	delete old;
}

triangle *square::select_triangle(const vector *org, const vector *dir)
{
	triangle *tri, *sel = 0;
	float n, sel_n = 0;

	for (tri = tris; tri; tri = tri->next)
	{
		if ((n = tri->isect_line(this, org, dir)) > 0
		    && (!sel || n < sel_n))
		{
			sel = tri;
			sel_n = n;
		}
	}

	return sel;
}

void square::copy(square *sqr)
{
	triangle *t, *tri, **ptr;

	ptr = &tris;

	for (t = sqr->tris; t; t = t->next)
	{
		if ( !(tri = *ptr) )
		{
			*ptr = tri = new triangle;
		}

		tri->copy(t);

		ptr = &tri->next;
	}

	while ( (tri = *ptr) )
	{
		*ptr = tri->next;
		delete tri;
	}
}

void square::clear()
{
	triangle *tri, *next;

	for (tri = tris; tri; tri = next)
	{
		next = tri->next;
		delete tri;
	}

	tris = 0;
}

triangle::triangle()
{
	next = 0;
}

void triangle::queue(landscape *land, const vector *v, const vector *dir)
{
	unsigned i;	

	for (i = 0; i < 3; i++)
	{
		vec[i].add(&vtx[i], v);
	}

	if (tex.has_alpha())
	{
		vector c;

		c.set(&vec[0]);
		c.add(&vec[1]);
		c.add(&vec[2]);
		dist = c.dot(dir);

		land->queue(this);
	}
	else
	{
		tex.queue(this);
	}
}

void triangle::draw(landscape *land)
{
	unsigned i;

	glBegin(GL_TRIANGLES);

	for (i = 0; i < 3; i++)
	{
		vtx[i].gltexture();
		vec[i].glvertex();
	}

	glEnd();

	if (land->selected_triangle() != this &&
	    !tex.has_alpha())
	{
		glBegin(GL_LINE_LOOP);

		for (i = 0; i < 3; i++)
		{
			vtx[i].gltexture();
			vec[i].glvertex();
		}

		glEnd();
	}
}

void triangle::highlight()
{
	unsigned i;

	glBegin(GL_LINE_LOOP);

	for (i = 0; i < 3; i++)
	{
		vec[i].glvertex();
	}

	glEnd();
}

int triangle::load(FILE *fp, texture *texs)
{
	unsigned i;

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	tex.copy(&texs[i]);

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].load(fp))
			return -1;
	}

	return 0;
}

int triangle::save(FILE *fp)
{
	unsigned i;

	i = tex.index();
	if (fputc(i, fp) != int(i))
	{
		fclose(fp);
		return -1;
	}

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].save(fp))
			return -1;
	}

	return 0;
}

int triangle::select_vertex(const vector *sqr,
			    const vector *org, const vector *dir)
{
	int i, sel = -1;
	float n, sel_n = 1.0;
	vector v;

	for (i=0; i<3; i++)
	{
		v.add(sqr, &vtx[i]);
		if (v.isect_point_line(&v, org, dir) > 0 &&
		    (n = v.val2()) <= sel_n)
		{
			sel = i;
			sel_n = n;
		}
	}

	return sel;
}

float triangle::isect_line(const vector *sqr,
			   const vector *org, const vector *dir)
{
	vector vn, vo, va, vb;
	float a, b, n;

	va.sub(&vtx[1], &vtx[0]);
	vb.sub(&vtx[2], &vtx[0]);
	vn.cross(&va, &vb);

	if (fabs(n = vn.dot(dir)) < FLT_EPSILON)
		return 0;

	vo.add(&vtx[0], sqr);

	va.cross(dir);
	if ( (a = (va.dot(&vo) - va.dot(org)) / n) < 0.0 )
		return 0;

	vb.cross(dir);
	if ( (b = (vb.dot(org) - vb.dot(&vo)) / n) < 0.0 )
		return 0;

	if ( (a + b) > 1.0 )
		return 0;

	return (vn.dot(&vo) - vn.dot(org)) / n;
}

void triangle::copy(triangle *tri)
{
	int i;

	for (i = 0; i < 3; i++)
	{
		vtx[i].set(&tri->vtx[i]);
		vtx[i].u = tri->vtx[i].u;
		vtx[i].v = tri->vtx[i].v;
	}

	tex.copy(&tri->tex);
}

vertex::vertex()
{
	set(0.0, 0.0, 0.0);
	u = v = 0.0;
}

int vertex::load(FILE *fp)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	if (fread(&data, 1, 5, fp) != 5)
		return -1;

	set(float(data.x), float(data.y), float(data.z));

	u = float(data.u) / 64.0;
	v = float(data.v) / 64.0;

	return 0;
}

int vertex::save(FILE *fp)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	data.x = signed(x());
	data.y = signed(y());
	data.z = signed(z());

	data.u = unsigned(u * 64.0);
	data.v = unsigned(v * 64.0);

	if (fwrite(&data, 1, 5, fp) != 5)
		return -1;

	return 0;
}
