/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <png.h>

#include "config.h"
#include "texture.h"
#include "landscape.h"

drawable::~drawable()
{
}

void drawable::draw(landscape *)
{
}

texture_data::texture_data() : node<texture_data>(this)
{
	count = 0;

	tex = 0;
	name = 0;
	index = 0;
	first_d = 0;
}

texture_data::~texture_data()
{
	if (name) delete name;
	if (tex) glDeleteTextures(1, &tex);
}

int texture_data::load(const char *file)
{
	FILE *fp;
	char fname[1024];

	sprintf(fname, "textures/%s", file);

	if ( !(fp = fopen(fname, "rb")) )
		return -1;

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	png_infop info_ptr = png_create_info_struct(png_ptr);

	if (setjmp(png_ptr->jmpbuf))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, 0);
		fclose(fp);
		return -1;
	}

	png_init_io(png_ptr, fp);

	png_read_info(png_ptr, info_ptr);
	png_set_strip_16(png_ptr);
	png_set_expand(png_ptr);
	png_read_update_info(png_ptr, info_ptr);

	png_uint_32 width = png_get_image_width(png_ptr, info_ptr);
	png_uint_32 height = png_get_image_height(png_ptr, info_ptr);
	png_uint_32 rowbytes = png_get_rowbytes(png_ptr, info_ptr);
	png_byte channels = png_get_channels(png_ptr, info_ptr);

	GLenum fmt;

	switch(channels)
	{
	case 1:
		fmt = GL_LUMINANCE;
		break;
	case 2:
		fmt = GL_LUMINANCE_ALPHA;
		break;
	case 3:
		fmt = GL_RGB;
		break;
	case 4:
		fmt = GL_RGBA;
		break;
	default:
		png_destroy_read_struct(&png_ptr, &info_ptr, 0);
		fclose(fp);
		return -1;
	}

	png_bytep pixels = (png_bytep)alloca(height * rowbytes);
	png_bytep *rows = (png_bytep *)alloca(height * sizeof(png_bytep));

	for (png_uint_32 row = 0; row < height; row++)
	{
		rows[row] = &pixels[row * rowbytes];
	}

	png_read_image(png_ptr, rows);

	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, 0);

	fclose(fp);

	strcpy(name = new char[strlen(file) + 1], file);

	if (!tex) glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D, tex);
	glTexImage2D(GL_TEXTURE_2D, 0, channels, width, height, 0,
		     fmt, GL_UNSIGNED_BYTE, pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	sizex = width; sizey = height;
	alpha = !(channels & 1);

	return 0;
}

void texture_data::draw(landscape *land)
{
	drawable *d;

	if ( (d = first_d) )
	{
		glbind();

		do
		{
			d->draw(land);
		}
		while ( (d = d->next_d) );

		first_d = 0;
	}
}

list<texture_data> texture::lh;
texture_data *texture::def = 0;

void texture::init()
{
	DIR *dir;
	struct dirent *ent;
	struct stat buf;
	char fname[1024];

	if (!(dir = opendir("textures")))
		return;

	while ( (ent = readdir(dir)) )
	{
		sprintf(fname, "textures/%s", ent->d_name);
		stat(fname, &buf);
		if (S_ISREG(buf.st_mode) && !find(ent->d_name))
		{
			load(ent->d_name);
		}
	}

	closedir(dir);

	set_default();
	set(0);
}

int texture::load(const char *file)
{
	texture_data *data = new texture_data;

	if (data->load(file))
	{
		delete data;
		return -1;
	}

	lh.add_tail(data);
	set(data);

	return 0;
}

int texture::load(FILE *fp)
{
	unsigned i;
	char fname[256];

	if ( (i = fgetc(fp)) > 255 )
		return -1;

	if (i)
	{
		if (fread(fname, 1, i, fp) != i)
			return -1;
	}
	fname[i] = 0;

	if (!find(fname))
		return -1;

	return 0;
}

int texture::save(FILE *fp)
{
	const char *str;
	unsigned n;

	if (!(str = name()))
		return -1;

	n = strlen(str);

	if (n > 255)
		return -1;

	if (fputc(n, fp) != int(n))
		return -1;

	if (fwrite(str, 1, n, fp) != n)
		return -1;

	return 0;
}

int texture::find(const char *name)
{
	texture_data *data;

	for (data = lh.first(); data; data = data->next())
	{
		if ( data->filename() &&
		     !strcmp(data->filename(), name) )
		{
			set(data);
			return 1;
		}
	}

	return 0;
}

unsigned texture::reindex()
{
	texture_data *data;
	unsigned n = 0;

	for (data = lh.first(); data; data = data->next())
	{
		data->index = data->count ? ++n : 0;
	}	

	return n;
}

void texture::draw(landscape *land)
{
	texture_data *data;

	for (data = lh.first(); data; data = data->next())
	{
		data->draw(land);
	}	
}
