/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>

#include <GL/gl.h>

#include "config.h"
#include "misc.h"
#include "camera.h"

void camera::init()
{
	width = cfg.get_float("width");
	height = cfg.get_float("height");
	nearv.set_z(cfg.get_float("near"));
	farv.set_z(cfg.get_float("far"));
	set_fov(cfg.get_float("fov"));
}

void camera::set_size(float w, float h)
{
	width = w;
	height = h;
	set_fov(fov);
}

void camera::set_fov(float f)
{
	float fx, fy;

	fov = f;

	if (height > width)
	{
		fx = f * width / height;
		fy = f;
	}
	else
	{
		fx = f;
		fy = f * height / width;
	}

	nearv.set_x(nearv.z() * fx);
	farv.set_x(farv.z() * fx);

	hx = 1.0 / sqrt((fx * fx) + 1.0);
	hz = hx * fx;

	nearv.set_y(nearv.z() * fy);
	farv.set_y(farv.z() * fy);

	vy = 1.0 / sqrt((fy * fy) + 1.0);
	vz = vy * fy;
}

void camera::perspective()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-nearv.x(), nearv.x(), -nearv.y(), nearv.y(),
		  nearv.z(), farv.z());
	glMatrixMode(GL_MODELVIEW);
}

void camera::ortho()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, o_width(), o_height(), 0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void camera::min_max(vector *v)
{
	v->transform(&ctow);
	box.min_max(v);
}

void camera::make()
{
	vector x, y, z;

	z.sub(&pos, &tgt); z.unit();
	y.set(0.0, 0.0, 1.0);
	x.cross(&y, &z); x.unit();
	y.cross(&z, &x); y.unit();

	ctow.set(&x, &y, &z, &pos);
}

void camera::load()
{
	vector x;

	make();

	wtoc.invert(&ctow);
	wtoc.glload();

	box.min.set(&pos);
	box.max.set(&pos);

	x.set(farv.x(), farv.y(), -farv.z()); min_max(&x);
	x.set(-farv.x(), farv.y(), -farv.z()); min_max(&x);
	x.set(farv.x(), -farv.y(), -farv.z()); min_max(&x);
	x.set(-farv.x(), -farv.y(), -farv.z()); min_max(&x);
}

int camera::check_bsphere(vector *pos, float size)
{
	vector v;

	v.transform(pos, &wtoc);

	if ((v.z() + nearv.z()) > size)
		return 1;

	if (-(v.z() + farv.z()) > size)
		return 1;

	float a = v.z() * hz;
	float b = v.x() * hx;

	if ((a + b) > size)
		return 1;

	if ((a - b) > size)
		return 1;

	a = v.z() * vz;
	b = v.y() * vy;

	if ((a + b) > size)
		return 1;

	if ((a - b) > size)
		return 1;

	return 0;
}
