/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>

#include "model.h"
#include "entity.h"
#include "gstring.h"
#include "camera.h"
#include "exec.h"
#include "misc.h"

entity *entity::array[65536];
list<entity> entity::lh;
unsigned entity::last_id = 0;

entity::entity(unsigned i) : node<entity>(this), aud(&origin)
{
	mdl = 0;
	origin.set(0.0, 0.0, 0.0);
	velocity.set(0.0, 0.0, 0.0);
	angles.set(0.0, 0.0, 0.0);
	avelocity.set(0.0, 0.0, 0.0);
	size.min.set(0.0, 0.0, 0.0);
	size.max.set(0.0, 0.0, 0.0);
	reset_frame();

	varp = new variable[gstring::num_vars()];
	memset(varp, 0, gstring::num_vars() * sizeof(variable));

	if (i && !array[i])
	{
		id = i;
	}
	else
	{
		/* Warning: this will lock up if there are 65535 entities */
		do {
			if (++last_id > 65535)
				last_id = 1;
		} while (array[last_id]);

		id = last_id;
	}

	lh.add_tail(this);
	array[id] = this;
}

entity::~entity()
{
	if (array[id] == this)
		array[id] = 0;
	delete [] varp;
}

entity *entity::get_ent(unsigned i)
{
	entity *ret = array[i];

	if (!ret) ret = new entity(i);

	return ret;
}

void entity::delete_all()
{
	entity *e;

	while ((e = lh.first()))
	{
		delete e;
	}
}

void entity::draw(camera *cam, float t)
{
	for (entity *e = lh.first(); e; e = e->next())
	{
		model *mdl;
		matrix m;
		vector p, r;

		if (!(mdl = e->mdl))
			continue;

		e->position(&p, t);

		if (cam->check_bsphere(&p, mdl->radius()))
			continue;

		e->rotation(&r, t);
		m.set(&p, &r);

		glPushMatrix();
		m.glmult();
		mdl->draw(e->k0, e->k1, t);
		glPopMatrix();
	}
}

void entity::tick()
{
	static gstring *s = 0;
	function *fn;
	entity *e, *n;

	for (e = lh.first(); e; e = e->next())
	{
		e->origin.add(&e->velocity);
		e->velocity.set(0.0, 0.0, 0.0);

		e->angles.add(&e->avelocity);
		e->avelocity.set(0.0, 0.0, 0.0);

		e->k0 = e->k1;
	}

	if (!s) s = new_string("tick");

	if ( (fn = s->fn()) )
		fn->call(player_entity()->idx());

	for (n = lh.first(); (e = n); n = e->next())
	{
		if (array[e->id] != e)
			delete e;
	}
}

float entity::intersect(function *fn)
{
	vector vec;
	bbox box0, box1;
	variable *v = fn->vars();

	position(&vec, 1.0);
	box0.add(&size, &vec);

	for (entity *e = lh.first(); e; e = e->next())
	{
		if (e != this)
		{
			e->position(&vec, 1.0);
			box1.add(&e->size, &vec);

			if (box0.intersect(&box1))
			{
				v[2].idx = idx();
				fn->call(e->idx());

				if (v[1].flt)
					return v[1].flt;
			}
		}
	}

	return 0.0;
}
