/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>

#include "gstring.h"
#include "exec.h"
#include "builtins.h"
#include "entity.h"
#include "misc.h"

int load_code(const char *file)
{
	interpretor::instance.open();

	if (interpretor::instance.load("default"))
		return -1;

	if (interpretor::instance.load(file))
		return -1;

	interpretor::instance.close();

	return 0;
}

interpretor::interpretor()
{
}

void interpretor::open()
{
	function *fn;

	delete_entities();

	while((fn = fnlh.first()))
	{
		delete fn;
	}

	reset_variables();

	new fn_new(&fnlh);
	new fn_player(&fnlh);
	new fn_camera(&fnlh);
	new fn_origin(&fnlh);
	new fn_velocity(&fnlh);
	new fn_angles(&fnlh);
	new fn_avelocity(&fnlh);
	new fn_setorigin(&fnlh);
	new fn_setangles(&fnlh);
	new fn_setvelocity(&fnlh);
	new fn_setavelocity(&fnlh);
	new fn_setsize(&fnlh);
	new fn_setmodel(&fnlh);
	new fn_setframe(&fnlh);
	new fn_waterheight(&fnlh);
	new fn_setwatertexture(&fnlh);
	new fn_setwaterheight(&fnlh);
	new fn_setclouds(&fnlh);
	new fn_setredsun(&fnlh);
	new fn_setbluesun(&fnlh);
	new fn_hota(&fnlh);
	new fn_hotb(&fnlh);
	new fn_gradient(&fnlh);
	new fn_intersect(&fnlh);
	new fn_keys(&fnlh);
	new fn_makevectors(&fnlh);
	new fn_v_forward(&fnlh);
	new fn_v_left(&fnlh);
	new fn_v_up(&fnlh);
	new fn_normalize(&fnlh);
	new fn_precache_sound(&fnlh);
	new fn_play_sound(&fnlh);
	new fn_play_sound_loop(&fnlh);
	new fn_stop_sound(&fnlh);
}

interpretor interpretor::instance;

void interpretor::close()
{
	static gstring *start = 0;
	function *fn;
	entity *player;

	player = player_entity();
	camera_entity();

	if (!start) start = new_string("start");

	if ( (fn = start->fn()) )
		fn->call(player->idx());
}

int interpretor::load(const char *file)
{
	FILE *fp;
	char name[1024];
	int ret;

	sprintf(name, "code/%s.lost", file);

	if ( !(fp = fopen(name, "rb")) )
		return -1;

	ret = load(fp);

	fclose(fp);

	return ret;
}

int interpretor::load(FILE *fp)
{
	unsigned i, n;
	char magic[4];

	if ( fread(magic, 1, 4, fp) != 4 ||
	     strncmp(magic, "CJPO", 4) )
		return -1;

	/* strings */

	if (signed(n = get_int(fp)) < 0)
		return -1;

	gstring **strs = (gstring **)alloca((n + 1) * sizeof(gstring *));

	strs[0] = new_string("");

	for (i = 1; i <= n; i++)
	{
		if (!(strs[i] = load_string(fp)))
			return -1;
	}

	/* variables */

	if (signed(n = get_int(fp)) < 0)
		return -1;

	unsigned char vars[256];
	unsigned vi = 0;

	for (i = 0; i < n; i++)
	{
		unsigned c, s;

		if ((c = fgetc(fp)) > 255)
			return -1;
		if (signed(s = get_int(fp)) < 0)
			return -1;
		if (strs[s]->set_var_type(c))
			return -1;

		unsigned si = strs[s]->var_index(); 

		vars[vi++] = si;
		if (c == T_VECTOR)
		{
			vars[vi++] = si + 1;
			vars[vi++] = si + 2;
		}
	}

	/* functions */

	if (signed(n = get_int(fp)) < 0)
		return -1;

	for (i = 0; i < n; i++)
	{
		function *fn = new function(&fnlh);

		if (fn->load(fp, strs, vars))
			return -1;
	}

	return 0;
}

function::function(list<function> *lh) : node<function>(this)
{
	lh->add_tail(this);

	name = 0;
	code = 0;
	varp = 0;
	nparam = 0;
	nvar = 0;
}

function::~function()
{
	if (code) delete [] code;
	if (varp) delete [] varp;
}

void function::init(const char *str, unsigned n)
{
	name = new_string(str);
	nparam = n++;
	nvar = n;

	if (varp) delete [] varp;
	varp = new variable[n];
	memset(varp, 0, n * sizeof(variable));

	name->set_fn(this);
}

int function::load(FILE *fp, gstring **strs, unsigned char *vars)
{
	unsigned i, j, n;
	function *fn;
	
	if (signed(n = get_int(fp)) < 0)
		return -1;

	if (signed(j = get_int(fp)) < 0)
		return -1;

	name = strs[n];

	if ( (fn = name->fn()) && fn->nparam != j)
		return -1;

	if (signed(n = get_int(fp)) < 0)
		return -1;

	nparam = j++;
	nvar = n;

	if (varp) delete [] varp;
	varp = new variable[n];
	memset(varp, 0, n * sizeof(variable));

	/* constants */

	if (signed(n = get_int(fp)) < 0)
		return -1;

	for (i = 0; i < n; i++)
	{
		unsigned c;

		if ((c = fgetc(fp)) > 255)
			return -1;

		switch(c)
		{
		case T_FLOAT:
			if (get_float(fp, &varp[j].flt))
				return -1;
			j ++;
			break;

		case T_VECTOR:
			if (get_float(fp, &varp[j + 0].flt))
				return -1;
			if (get_float(fp, &varp[j + 1].flt))
				return -1;
			if (get_float(fp, &varp[j + 2].flt))
				return -1;
			j += 3;
			break;

		case T_STRING:
		case T_FUNCTION:
			if (signed(c = get_int(fp)) < 0)
				return -1;
			varp[j].idx = strs[c]->idx();
			j ++;
			break;
		}
	}

	/* code */

	if (signed(n = get_int(fp)) <= 0)  // 0 is not allowed
		return -1;

	if (code) delete [] code;
	code = new instruction[n];

	for (i = 0; i < n; i++)
	{
		code[i].load(fp, vars);
	}

	if (code[n-1].op)
		return -1;

	name->set_fn(this);

	return 0;
}

int instruction::load(FILE *fp, unsigned char *vi)
{
	int i;

	if ((op = fgetc(fp)) > 255)
		return -1;

	if ((i = get_int(fp)) < 0)
		return -1;
	a = i;

	if ((i = get_int(fp)) < 0)
		return -1;

	if (op >= OP_LOAD_EF && op <= OP_STORE_EE)
	{
		b = vi[i];
	}
	else
	{
		b = i;
	}

	if ((i = get_int(fp)) < 0)
		return -1;
	c = i;

	return 0;
}

void function::call(unsigned ent)
{
	if (varp[0].idx)
		return;

	varp[0].idx = ent;

	gstring **strs = string_array();
	entity **ents = entity_array();

	unsigned pc = 0;
	unsigned char op;

	while ((op = code[pc].op))
	{
		unsigned a = code[pc].a;
		unsigned b = code[pc].b;
		unsigned c = code[pc].c;
		pc++;

		switch(op)
		{
		case OP_LOAD_EF:
			struct entity *e;

			if ((e = ents[varp[a].idx]))
			{
				varp[c].flt = e->vars()[b].flt;
				break;
			}
			varp[c].flt = 0.0;
			break;
		case OP_LOAD_EV:
			if ((e = ents[varp[a].idx]))
			{
				variable *v = e->vars();

				varp[c + 0].flt = v[b + 0].flt;
				varp[c + 1].flt = v[b + 1].flt;
				varp[c + 2].flt = v[b + 2].flt;
				break;
			}
			varp[c + 0].flt = 0.0;
			varp[c + 1].flt = 0.0;
			varp[c + 2].flt = 0.0;
			break;
		case OP_LOAD_ES:
		case OP_LOAD_EN:
		case OP_LOAD_EE:
			if ((e = ents[varp[a].idx]))
			{
				varp[c].idx = e->vars()[b].idx;
				break;
			}
			varp[c].idx = 0;
			break;
		case OP_STORE_EF:
			if ((e = ents[varp[a].idx]))
			{
				e->vars()[b].flt = varp[c].flt;
			}
			break;
		case OP_STORE_EV:
			if ((e = ents[varp[a].idx]))
			{
				variable *v = e->vars();

				v[b + 0].flt = varp[c + 0].flt;
				v[b + 1].flt = varp[c + 1].flt;
				v[b + 2].flt = varp[c + 2].flt;
			}
			break;
		case OP_STORE_ES:
		case OP_STORE_EN:
		case OP_STORE_EE:
			if ((e = ents[varp[a].idx]))
			{
				e->vars()[b].idx = varp[c].idx;
			}
			break;

		case OP_LOAD_NF:
			function *fn;
			if ((fn = strs[varp[a].idx]->fn()))
			{
				varp[c].flt = fn->varp[b].flt;
				break;
			}
			varp[c].flt = 0.0;
			break;
		case OP_LOAD_NV:
			if ((fn = strs[varp[a].idx]->fn()))
			{
				varp[c + 0].flt = fn->varp[b + 0].flt;
				varp[c + 1].flt = fn->varp[b + 1].flt;
				varp[c + 2].flt = fn->varp[b + 2].flt;
				break;
			}
			varp[c + 0].flt = 0.0;
			varp[c + 1].flt = 0.0;
			varp[c + 2].flt = 0.0;
			break;
		case OP_LOAD_NS:
		case OP_LOAD_NN:
		case OP_LOAD_NE:
			if ((fn = strs[varp[a].idx]->fn()))
			{
				varp[c].idx = fn->varp[b].idx;
				break;
			}
			varp[c].idx = 0;
			break;
		case OP_STORE_NF:
			if ((fn = strs[varp[a].idx]->fn()))
			{
				fn->varp[b].flt = varp[c].flt;
			}
			break;
		case OP_STORE_NV:
			if ((fn = strs[varp[a].idx]->fn()))
			{
				fn->varp[b + 0].flt = varp[c + 0].flt;
				fn->varp[b + 1].flt = varp[c + 1].flt;
				fn->varp[b + 2].flt = varp[c + 2].flt;
			}
			break;
		case OP_STORE_NS:
		case OP_STORE_NN:
		case OP_STORE_NE:
			if ((fn = strs[varp[a].idx]->fn()))
			{
				fn->varp[b].idx = varp[c].idx;
			}
			break;

		case OP_IF_F:
			if (varp[a].flt)
				pc = b;
			break;
		case OP_IF_V:
			if (varp[a + 0].flt || varp[a + 1].flt ||
			    varp[a + 2].flt)
				pc = b;
			break;
		case OP_IF_S:
		case OP_IF_N:
			if (varp[a].idx)
				pc = b;
			break;
		case OP_IF_E:
			if (ents[varp[a].idx])
				pc = b;
			break;
		case OP_IFNOT_F:
			if (!varp[a].flt)
				pc = b;
			break;
		case OP_IFNOT_V:
			if (!(varp[a + 0].flt || varp[a + 1].flt ||
			      varp[a + 2].flt))
				pc = b;
			break;
		case OP_IFNOT_S:
		case OP_IFNOT_N:
			if (!varp[a].idx)
				pc = b;
			break;
		case OP_IFNOT_E:
			if (!ents[varp[a].idx])
				pc = b;
			break;

		case OP_GOTO:
			pc = b;
			break;
		case OP_CALL:
			if ((fn = strs[varp[b].idx]->fn()))
			{
				fn->call(varp[a].idx);
			}
			break;
		case OP_RETURN:
			varp[0].idx = 0;
			return;

		case OP_MOVE_F:
			varp[c].flt = varp[a].flt;
			break;
		case OP_MOVE_V:
			varp[c + 0].flt = varp[a + 0].flt;
			varp[c + 1].flt = varp[a + 1].flt;
			varp[c + 2].flt = varp[a + 2].flt;
			break;
		case OP_MOVE_S:
		case OP_MOVE_N:
		case OP_MOVE_E:
			varp[c].idx = varp[a].idx;
			break;

		case OP_MUL_F:
			varp[c].flt = varp[a].flt * varp[b].flt;
			break;
		case OP_MUL_V:
			varp[c].flt =
				(varp[a + 0].flt * varp[b + 0].flt) +
				(varp[a + 1].flt * varp[b + 1].flt) +
				(varp[a + 2].flt * varp[b + 2].flt);
			break;
		case OP_MUL_FV:
			varp[c + 0].flt = varp[a].flt * varp[b + 0].flt;
			varp[c + 1].flt = varp[a].flt * varp[b + 1].flt;
			varp[c + 2].flt = varp[a].flt * varp[b + 2].flt;
			break;
		case OP_MUL_VF:
			varp[c + 0].flt = varp[a + 0].flt * varp[b].flt;
			varp[c + 1].flt = varp[a + 1].flt * varp[b].flt;
			varp[c + 2].flt = varp[a + 2].flt * varp[b].flt;
			break;
		case OP_DIV_F:
			varp[c].flt = varp[a].flt / varp[b].flt;
			break;
		case OP_DIV_VF:
			varp[c + 0].flt = varp[a + 0].flt / varp[b].flt;
			varp[c + 1].flt = varp[a + 1].flt / varp[b].flt;
			varp[c + 2].flt = varp[a + 2].flt / varp[b].flt;
			break;
		case OP_ADD_F:
			varp[c].flt = varp[a].flt + varp[b].flt;
			break;
		case OP_ADD_V:
			varp[c + 0].flt = varp[a + 0].flt + varp[b + 0].flt;
			varp[c + 1].flt = varp[a + 1].flt + varp[b + 1].flt;
			varp[c + 2].flt = varp[a + 2].flt + varp[b + 2].flt;
			break;
		case OP_SUB_F:
			varp[c].flt = varp[a].flt - varp[b].flt;
			break;
		case OP_SUB_V:
			varp[c + 0].flt = varp[a + 0].flt - varp[b + 0].flt;
			varp[c + 1].flt = varp[a + 1].flt - varp[b + 1].flt;
			varp[c + 2].flt = varp[a + 2].flt - varp[b + 2].flt;
			break;
		case OP_NEG_F:
			varp[c].flt = - varp[a].flt;
			break;
		case OP_NEG_V:
			varp[c + 0].flt = - varp[a + 0].flt;
			varp[c + 1].flt = - varp[a + 1].flt;
			varp[c + 2].flt = - varp[a + 2].flt;
			break;

		case OP_EQ_F:
			varp[c].flt = (varp[a].flt == varp[b].flt);
			break;
		case OP_EQ_V:
			varp[c].flt = ((varp[a + 0].flt == varp[b + 0].flt) &&
				       (varp[a + 1].flt == varp[b + 1].flt) &&
				       (varp[a + 2].flt == varp[b + 2].flt));
			break;
		case OP_EQ_S:
		case OP_EQ_N:
			varp[c].flt = (varp[a].idx == varp[b].idx);
			break;
		case OP_EQ_E:
			varp[c].flt = (ents[varp[a].idx] == ents[varp[b].idx]);
			break;
		case OP_NE_F:
			varp[c].flt = (varp[a].flt != varp[b].flt);
			break;
		case OP_NE_V:
			varp[c].flt = ((varp[a + 0].flt != varp[b + 0].flt) ||
				       (varp[a + 1].flt != varp[b + 1].flt) ||
				       (varp[a + 2].flt != varp[b + 2].flt));
			break;
		case OP_NE_S:
		case OP_NE_N:
			varp[c].flt = (varp[a].idx != varp[b].idx);
			break;
		case OP_NE_E:
			varp[c].flt = (ents[varp[a].idx] != ents[varp[b].idx]);
			break;
		case OP_LE_F:
			varp[c].flt = (varp[a].flt <= varp[b].flt);
			break;
		case OP_GE_F:
			varp[c].flt = (varp[a].flt >= varp[b].flt);
			break;
		case OP_LT_F:
			varp[c].flt = (varp[a].flt < varp[b].flt);
			break;
		case OP_GT_F:
			varp[c].flt = (varp[a].flt > varp[b].flt);
			break;

		case OP_AND_FF:
			varp[c].flt = (varp[a].flt && varp[b].flt);
			break;
		case OP_AND_FV:
			varp[c].flt = (varp[a].flt &&
				       (varp[b + 0].flt || varp[b + 1].flt ||
					varp[b + 2].flt));
			break;
		case OP_AND_FS:
		case OP_AND_FN:
			varp[c].flt = (varp[a].flt && varp[b].idx);
			break;
		case OP_AND_FE:
			varp[c].flt = (varp[a].flt && ents[varp[b].idx]);
			break;
		case OP_AND_VF:
			varp[c].flt = ((varp[a + 0].flt || varp[a + 1].flt ||
					varp[a + 2].flt) && varp[b].flt);
			break;
		case OP_AND_VV:
			varp[c].flt = ((varp[a + 0].flt || varp[a + 1].flt ||
					varp[a + 2].flt) && 
				       (varp[b + 0].flt || varp[b + 1].flt ||
					varp[b + 2].flt));
			break;
		case OP_AND_VS:
		case OP_AND_VN:
			varp[c].flt = ((varp[a + 0].flt || varp[a + 1].flt ||
					varp[a + 2].flt) && varp[b].idx);
			break;
		case OP_AND_VE:
			varp[c].flt = ((varp[a + 0].flt || varp[a + 1].flt ||
					varp[a + 2].flt) && ents[varp[b].idx]);
			break;
		case OP_AND_SF:
		case OP_AND_NF:
			varp[c].flt = (varp[a].idx && varp[b].flt);
			break;
		case OP_AND_SV:
		case OP_AND_NV:
			varp[c].flt = (varp[a].idx &&
				       (varp[b + 0].flt || varp[b + 1].flt ||
					varp[b + 2].flt));
			break;
		case OP_AND_SS:
		case OP_AND_SN:
		case OP_AND_NS:
		case OP_AND_NN:
			varp[c].flt = (varp[a].idx && varp[b].idx);
			break;
		case OP_AND_SE:
		case OP_AND_NE:
			varp[c].flt = (varp[a].idx && ents[varp[b].idx]);
			break;
		case OP_AND_EF:
			varp[c].flt = (ents[varp[a].idx] && varp[b].flt);
			break;
		case OP_AND_EV:
			varp[c].flt = (ents[varp[a].idx] &&
				       (varp[b + 0].flt || varp[b + 1].flt ||
					varp[b + 2].flt));
			break;
		case OP_AND_ES:
		case OP_AND_EN:
			varp[c].flt = (ents[varp[a].idx] && varp[b].idx);
			break;
		case OP_AND_EE:
			varp[c].flt = (ents[varp[a].idx] && ents[varp[b].idx]);
			break;
		case OP_OR_FF:
			varp[c].flt = (varp[a].flt || varp[b].flt);
			break;
		case OP_OR_FV:
			varp[c].flt = (varp[a].flt || varp[b + 0].flt ||
				       varp[b + 1].flt || varp[b + 2].flt);
			break;
		case OP_OR_FS:
		case OP_OR_FN:
			varp[c].flt = (varp[a].flt || varp[b].idx);
			break;
		case OP_OR_FE:
			varp[c].flt = (varp[a].flt || ents[varp[b].idx]);
			break;
		case OP_OR_VF:
			varp[c].flt = (varp[a + 0].flt || varp[a + 1].flt ||
				       varp[a + 2].flt || varp[b].flt);
			break;
		case OP_OR_VV:
			varp[c].flt = (varp[a + 0].flt || varp[a + 1].flt ||
				       varp[a + 2].flt || varp[b + 0].flt ||
				       varp[b + 1].flt || varp[b + 2].flt);
			break;
		case OP_OR_VS:
		case OP_OR_VN:
			varp[c].flt = (varp[a + 0].flt || varp[a + 1].flt ||
				       varp[a + 2].flt || varp[b].idx);
			break;
		case OP_OR_VE:
			varp[c].flt = (varp[a + 0].flt || varp[a + 1].flt ||
				       varp[a + 2].flt || ents[varp[b].idx]);
			break;
		case OP_OR_SF:
		case OP_OR_NF:
			varp[c].flt = (varp[a].idx || varp[b].flt);
			break;
		case OP_OR_SV:
		case OP_OR_NV:
			varp[c].flt = (varp[a].idx || varp[b + 0].flt ||
				       varp[b + 1].flt || varp[b + 2].flt);
			break;
		case OP_OR_SS:
		case OP_OR_SN:
		case OP_OR_NS:
		case OP_OR_NN:
			varp[c].flt = (varp[a].idx || varp[b].idx);
			break;
		case OP_OR_SE:
		case OP_OR_NE:
			varp[c].flt = (varp[a].idx || ents[varp[b].idx]);
			break;
		case OP_OR_EF:
			varp[c].flt = (ents[varp[a].idx] || varp[b].flt);
			break;
		case OP_OR_EV:
			varp[c].flt = (ents[varp[a].idx] || varp[b + 0].flt ||
				       varp[b + 1].flt || varp[b + 2].flt);
			break;
		case OP_OR_ES:
		case OP_OR_EN:
			varp[c].flt = (ents[varp[a].idx] || varp[b].idx);
			break;
		case OP_OR_EE:
			varp[c].flt = (ents[varp[a].idx] || ents[varp[b].idx]);
			break;
		case OP_NOT_F:
			varp[c].flt = !varp[a].flt;
			break;
		case OP_NOT_V:
			varp[c].flt = !(varp[a + 0].flt || varp[a + 1].flt ||
					varp[a + 2].flt);
			break;
		case OP_NOT_S:
		case OP_NOT_N:
			varp[c].flt = !varp[a].idx;
			break;
		case OP_NOT_E:
			varp[c].flt = !ents[varp[a].idx];
			break;

		case OP_BITAND_F:
			varp[c].flt = int(varp[a].flt) & int(varp[b].flt);
			break;
		case OP_BITOR_F:
			varp[c].flt = int(varp[a].flt) | int(varp[b].flt);
			break;	
		case OP_BITEOR_F:
			varp[c].flt = int(varp[a].flt) ^ int(varp[b].flt);
			break;
		case OP_BITNOT_F:
			varp[c].flt = ~ int(varp[a].flt);
			break;
		}
	}

	varp[0].idx = 0;
}
