/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <float.h>

#include "camera.h"
#include "landscape.h"
#include "texture.h"
#include "gstring.h"
#include "config.h"

landscape::landscape()
{
	wtex = 0;
	wz = 0.0;
	ntex = 0;
	above = 0;
	below = 0;
}

landscape::~landscape()
{
}

int landscape::load(const char *file)
{
	FILE *fp;
	unsigned i, j;
	vector vec;
	char name[1024];

	sprintf(name, "maps/%s", file);

	if ( !(fp = fopen(name, "rb")) )
		return -1;

	if ( fread(name, 1, 4, fp) != 4 ||
	     strncmp(name, "CJPL", 4) )
	{
		fclose(fp);
		return -1;
	}

	if ( (ntex = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	for (i = 1; i <= ntex; i++)
	{
		if (tex[i].load(fp))
		{
			fclose(fp);
			return -1;
		}
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			vec.set(float(i*16), float(j*16), 0.0);
			if (sqr[j][i].load(fp, &vec, tex))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	fclose(fp);

	return 0;
}

l_square::l_square()
{
	ntri = 0;
	tris = 0;
}

l_square::~l_square()
{
	if (tris) delete [] tris;
}

int l_square::load(FILE *fp, vector *vec, l_texture *tex)
{
	unsigned i, j;

	if ( (i = fgetc(fp)) > 255)
		return -1;
	vec->set_z(float(i));

	if ( (i = fgetc(fp)) > 255)
		return -1;

	if (tris) delete [] tris;
	tris = new l_triangle[ntri = i];

	for (j = 0; j < i; j++)
	{
		if (tris[j].load(fp, vec, tex))
			return -1;
	}

	make_bsphere();

	return 0;
}

void l_square::make_bsphere()
{
	bbox box;
	unsigned i;

	box.min.set(4096.0, 4096.0, 384.0);
	box.max.set(0.0, 0.0, -128.0);

	for (i = 0; i < ntri; i++)
	{
		tris[i].min_max(&box);
	}

	min_z = box.min.z();
	max_z = box.max.z();

	center.add(&box.min, &box.max);
	center.div(2.0);
	radius = 0.0;

	for (unsigned i = 0; i < ntri; i++)
	{
		tris[i].radius(&center, &radius);
	};
}

int l_triangle::load(FILE *fp, const vector *vec, l_texture *tp)
{
	unsigned i;

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	tex = &tp[i];

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].load(fp, vec))
			return -1;
	}

	make_normal();

	return 0;
}

void l_triangle::make_normal()
{
	vector a, b;

	a.sub(&vtx[1], &vtx[0]);
	b.sub(&vtx[2], &vtx[0]);
	norm.cross(&a, &b);
	norm.unit();
}

void l_triangle::min_max(bbox *box)
{
	for (unsigned i = 0; i < 3; i++)
	{
		box->min_max(&vtx[i]);
	}
}

void l_triangle::radius(const vector *vec, float *rad)
{
	for (unsigned i = 0; i < 3; i++)
	{
		vector v;
		float f;

		v.sub(&vtx[i], vec);
		f = v.val();

		if ((*rad) < f) *rad = f;
	}
}

int l_vertex::load(FILE *fp, const vector *vec)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	if (fread(&data, 1, 5, fp) != 5)
		return -1;

	set(float(data.x), float(data.y), float(data.z));
	add(vec);

	u = float(data.u) / 64.0;
	v = float(data.v) / 64.0;

	return 0;
}

l_texture::l_texture()
{
	first = 0;
}

l_texture::~l_texture()
{
}

int l_texture::load(FILE *fp)
{
	gstring *gstr;

	if (!(gstr = load_string(fp)))
		return -1;

	if (!(tex = gstr->tex()))
		return -1;

	return 0;
}

void landscape::draw(camera *cam)
{
	int xl = int(floor(cam->box.min.x() / 16.0));
	int yl = int(floor(cam->box.min.y() / 16.0));
	int xh = int(ceil(cam->box.max.x() / 16.0));
	int yh = int(ceil(cam->box.max.y() / 16.0));

	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh > 255) xh = 255;
	if (yh > 255) yh = 255;

	for (int y = yl; y <= yh; y++)
	{
		for (int x = xl ; x <= xh; x++)
		{
			sqr[y][x].draw(this, cam);
		}
	}

	for (unsigned i = 1; i <= ntex; i++)
	{
		tex[i].draw();
	}
}

void l_square::draw(landscape *land, camera *cam)
{
	if (cam->check_bsphere(&center, radius))
		return;

	for (unsigned i = 0; i < ntri; i++)
	{
		tris[i].queue(land, cam);
	}
}

void l_triangle::queue(landscape *land, camera *cam)
{
	if (tex->has_alpha())
	{
		vector c, d;

		c.set(&vtx[0]);
		c.add(&vtx[1]);
		c.add(&vtx[2]);
		cam->backward(&d);
		dist = d.dot(&c);

		land->queue(this, c.z() / 3.0);
	}
	else
	{
		next = tex->first;
		tex->first = this;
	}
}

void landscape::queue(l_triangle *tri, float z)
{
	l_triangle **ptr;
	l_triangle *next;

	ptr = (z < wz) ? &below : &above;
	
	while ((next = *ptr))
	{
		if (next->dist > tri->dist)
			break;
		ptr = &next->next;
	}

	tri->next = next;
	*ptr = tri;
}

void l_texture::draw()
{
	l_triangle *tri;

	if ( (tri = first) )
	{
		glbind();

		glBegin(GL_TRIANGLES);

		do
		{
			tri->draw();
		}
		while ( (tri = tri->next) );

		glEnd();

		first = 0;
	}
}

void l_triangle::draw()
{
	norm.glnormal();

	for (unsigned i = 0; i < 3; i++)
	{
		vtx[i].gltexture();
		vtx[i].glvertex();
	}
}

void landscape::draw_water(camera *cam)
{
	l_triangle *tri;
	vector pos;

	cam->position(&pos);

	for (tri = (pos.z() < wz) ? above : below; tri; tri = tri->next)
	{
		tri->glbind();
		glBegin(GL_TRIANGLES);
		tri->draw();
		glEnd();

		tri = tri->next;
	}

	if (wtex)
	{
		wtex->glbind();

		float xl = floor(cam->box.min.x() / 16.0);
		float yl = floor(cam->box.min.y() / 16.0);
		float xh = ceil(cam->box.max.x() / 16.0);
		float yh = ceil(cam->box.max.y() / 16.0);

		float u = xh - xl;
		float v = yh - yl;

		xl *= 16.0;
		yl *= 16.0;
		xh *= 16.0;
		yh *= 16.0;

		glBegin(GL_QUADS);

		glNormal3f(0.0, 0.0, 1.0);

		glTexCoord2f(0.0, 0.0);
		glVertex3f(xl, yl, wz);

		glTexCoord2f(u, 0.0);
		glVertex3f(xh, yl, wz);

		glTexCoord2f(u, v);
		glVertex3f(xh, yh, wz);

		glTexCoord2f(0.0, v);
		glVertex3f(xl, yh, wz);

		glEnd();
	}

	for (tri = (pos.z() < wz) ? below : above; tri; tri = tri->next)
	{
		tri->glbind();
		glBegin(GL_TRIANGLES);
		tri->draw();
		glEnd();
	}

	above = 0;
	below = 0;
}

float landscape::height_above(float *ptr, const vector *vec)
{
	int x, y;

	*ptr = 1.0;

	x = int(floor(vec->x() / 16.0));
	if (x < 0 || x > 255)
		return 65536.0;

	y = int(floor(vec->y() / 16.0));
	if (y < 0 || y > 255)
		return 65536.0;

	return sqr[y][x].height_above(ptr, vec);
}

float l_square::height_above(float *ptr, const vector *vec)
{
	float ret = 65536.0;

	for (unsigned i = 0; i < ntri; i++)
	{
		tris[i].height_above(&ret, ptr, vec);
	}

	return ret;
}

void l_triangle::height_above(float *ptr, float *slope, const vector *vec)
{
	vector a, b;
	float x, y, z;
	float i, j;

	if (norm.z() <= 0.0)
		return;

	a.sub(&vtx[1], &vtx[0]);
	b.sub(&vtx[2], &vtx[0]);

	x = vtx[0].x() - vec->x();
	y = vtx[0].y() - vec->y();

	if (!(z = (a.x() * b.y()) - (a.y() * b.x())))
		return;

	if ((i = ((b.y() * x) - (b.x() * y)) / z) < 0.0)
		return;

	if ((j = ((a.x() * y) - (a.y() * x)) / z) < 0.0)
		return;

	if (i + j > 1.0)
		return;

	if ((z = vtx[0].z() + (i * a.z()) + (j * b.z())) > vec->z())
		return;

	if (z > *ptr)
	{
		*ptr = z;
		*slope = norm.z();
	}
}

float landscape::height_below(float *ptr, const vector *vec)
{
	int x, y;

	*ptr = 1.0;

	x = int(floor(vec->x() / 16.0));
	if (x < 0 || x > 255)
		return -65536.0;

	y = int(floor(vec->y() / 16.0));
	if (y < 0 || y > 255)
		return -65536.0;

	return sqr[y][x].height_below(ptr, vec);
}

float l_square::height_below(float *ptr, const vector *vec)
{
	float ret = -65536.0;

	for (unsigned i = 0; i < ntri; i++)
	{
		tris[i].height_below(&ret, ptr, vec);
	}

	return ret;
}

void l_triangle::height_below(float *ptr, float *slope, const vector *vec)
{
	vector a, b;
	float x, y, z;
	float i, j;

	if (norm.z() <= 0.0)
		return;

	a.sub(&vtx[1], &vtx[0]);
	b.sub(&vtx[2], &vtx[0]);

	x = vec->x() - vtx[0].x();
	y = vec->y() - vtx[0].y();

	if (!(z = (a.x() * b.y()) - (a.y() * b.x())))
		return;

	if ((i = ((b.y() * x) - (b.x() * y)) / z) < 0.0)
		return;

	if ((j = ((a.x() * y) - (a.y() * x)) / z) < 0.0)
		return;

	if (i + j > 1.0)
		return;

	if ((z = vtx[0].z() + (i * a.z()) + (j * b.z())) > vec->z())
		return;

	if (z > *ptr)
	{
		*ptr = z;
		*slope = norm.z();
	}
}

int landscape::intersect(const bbox *box)
{
	int xl = int(floor(box->min.x() / 16.0));
	int yl = int(floor(box->min.y() / 16.0));
	int xh = int(ceil(box->max.x() / 16.0));
	int yh = int(ceil(box->max.y() / 16.0));

	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh > 255) xh = 255;
	if (yh > 255) yh = 255;

	for (int y = yl; y <= yh; y++)
	{
		for (int x = xl ; x <= xh; x++)
		{
			if (sqr[y][x].intersect(box))
				return 1;
		}
	}

	return 0;
}

int l_square::intersect(const bbox *box)
{
	if (box->min.z() > max_z)
		return 0;

	if (box->max.z() < min_z)
		return 0;

	for (unsigned i = 0; i < ntri; i++)
	{
		if (tris[i].intersect(box))
			return 1;
	}
	return 0;
}

int l_triangle::intersect(const bbox *box)
{
	float f;

	f = box->min.x();
	if ((vtx[0].x() < f) && (vtx[1].x() < f) && (vtx[2].x() < f))
		return 0;
	f = box->max.x();
	if ((vtx[0].x() > f) && (vtx[1].x() > f) && (vtx[2].x() > f))
		return 0;

	f = box->min.y();
	if ((vtx[0].y() < f) && (vtx[1].y() < f) && (vtx[2].y() < f))
		return 0;
	f = box->max.y();
	if ((vtx[0].y() > f) && (vtx[1].y() > f) && (vtx[2].y() > f))
		return 0;

	f = box->min.z();
	if ((vtx[0].z() < f) && (vtx[1].z() < f) && (vtx[2].z() < f))
		return 0;
	f = box->max.z();
	if ((vtx[0].z() > f) && (vtx[1].z() > f) && (vtx[2].z() > f))
		return 0;

	if (box->intersect(&vtx[0], &vtx[1]))
		return 1;
	if (box->intersect(&vtx[1], &vtx[2]))
		return 1;
	if (box->intersect(&vtx[2], &vtx[0]))
		return 1;

	if (intersect(&box->min, &box->max))
		return 1;

	vector a, b;

	a.set(box->max.x(), box->min.y(), box->min.z());
	b.set(box->min.x(), box->max.y(), box->max.z());
	if (intersect(&a, &b))
		return 1;

	a.set(box->min.x(), box->max.y(), box->min.z());
	b.set(box->max.x(), box->min.y(), box->max.z());
	if (intersect(&a, &b))
		return 1;

	a.set(box->min.x(), box->min.y(), box->max.z());
	b.set(box->max.x(), box->max.y(), box->min.z());
	if (intersect(&a, &b))
		return 1;

	return 0;
}

int l_triangle::intersect(const vector *l0, const vector *l1)
{
	vector vn, vd, va, vb;
	float a, b, t, n;

	va.sub(&vtx[1], &vtx[0]);
	vb.sub(&vtx[2], &vtx[0]);
	vn.cross(&va, &vb);

	vd.sub(l1, l0);
	
	if (fabs(n = vn.dot(&vd)) < FLT_EPSILON)
		return 0;

	if ( (t = (vn.dot(&vtx[0]) - vn.dot(l0)) / n) < 0.0 )
		return 0;

	if ( t > 1.0)
		return 0;

	va.cross(&vd);
	if ( (a = (va.dot(&vtx[0]) - va.dot(l0)) / n) < 0.0 )
		return 0;

	vb.cross(&vd);
	if ( (b = (vb.dot(l0) - vb.dot(&vtx[0])) / n) < 0.0 )
		return 0;

	if ( (a + b) > 1.0 )
		return 0;

	return 1;
}
