/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>

#include <GL/glut.h>

#include "config.h"
#include "camera.h"
#include "sky.h"
#include "landscape.h"
#include "audio.h"
#include "entity.h"
#include "misc.h"
#include "exec.h"
#include "font.h"

config cfg;
camera cam;
stars sky;
landscape map;
font fnt;
float fps;

unsigned keys;

static timeval tv0,tv1;

static void reshape(int, int);
static void display();
static void keyboard(unsigned char, int, int);
static void keydown(int, int, int);
static void keyup(int, int, int);

static void calc_fps(timeval *);

int main(int argc, char **argv)
{
	const char *dir;

	if ( (dir = cfg.get_string("game_dir")) )
	{
		chdir(dir);
	}

	//glutInitWindowPosition(0, 0);
	glutInitWindowSize(cfg.get_int("width"), cfg.get_int("height"));
	glutInit(&argc, argv);
	
	audio_init(&argc, argv);

	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow("Lost");

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(keydown);
	glutSpecialUpFunc(keyup);
	glutSetCursor(GLUT_CURSOR_NONE);

	cam.init();
	sky.init();

	if (map.load("start.map"))
		return 1;

	if (load_code("start.map"))
		return 1;

	if (fnt.load("default.font"))
		return 1;

	if (glutExtensionSupported("GL_NV_fog_distance"))
	{
		glFogi(GL_FOG_MODE, GL_LINEAR);
		glFogi(GL_FOG_DISTANCE_MODE_NV, GL_EYE_PLANE);
		glFogf(GL_FOG_START, -cam.get_near());
		glFogf(GL_FOG_END, -cam.get_far());
	}
	else
	{
		glFogi(GL_FOG_MODE, GL_LINEAR);
		glFogf(GL_FOG_START, cam.get_near());
		glFogf(GL_FOG_END, cam.get_far());
	}

	static const float one[4] = { 1.0, 1.0, 1.0, 1.0 };

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, one);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	glEnable(GL_TEXTURE_2D);

	if (glutExtensionSupported("GL_EXT_texture_lod_bias"))
	{
		glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT,
			  GL_TEXTURE_LOD_BIAS_EXT,
			  cfg.get_float("lod_bias"));
	}

	gettimeofday(&tv1, 0);

	glutMainLoop();

	return 0;
}

static void reshape(int width, int height)
{
	glViewport(0, 0, width, height);
	cam.set_size(width, height);
}

static void display()
{
	vector v;
	entity *player;

	gettimeofday(&tv0, 0);
	calc_fps(&tv0);

	float t = ((float(tv0.tv_sec - tv1.tv_sec) * 10.0) +
		   (float(tv0.tv_usec - tv1.tv_usec) / 100000.0));

	if (t >= 1.0)
	{
		if ((t -= 1.0) >= 1.0)
		{
			t = 1.0;

			tv1.tv_sec = tv0.tv_sec;
			if ((tv1.tv_usec = tv0.tv_usec - 100000) < 0)
			{
				tv1.tv_usec += 1000000;
				tv1.tv_sec -= 1;
			}
		}
		else
		{
			if ((tv1.tv_usec += 100000) > 1000000)
			{
				tv1.tv_usec -= 1000000;
				tv1.tv_sec += 1;
			}
		}

		tick_entities();
		keys &= KEY_MASK;

		cam.tgt.set(&player_entity()->origin);
		cam.pos.set(&camera_entity()->origin);
		cam.make();
		audio_update(&cam);

		gettimeofday(&tv0, 0);
	}

	player = player_entity();
	player->position(&v, t);
	cam.tgt.set(&v);

	camera_entity()->position(&v, t);
	cam.pos.set(&v);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	cam.perspective();
	cam.load();

//	glEnable(GL_LIGHTING);
//	glEnable(GL_FOG);

	glDepthMask(GL_FALSE);
	glEnable(GL_CULL_FACE);

	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_BLEND);
	sky.draw(&cam);

	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);

	glDisable(GL_BLEND);
	map.draw(&cam);
	draw_entities(&cam, t);

	glDisable(GL_CULL_FACE);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	map.draw_water(&cam);

	cam.ortho();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_FOG);

	fnt.glbind();
	char str[16];

	sprintf(str, "%ux%u", unsigned(player->origin.x()/16.0),
		unsigned(player->origin.y()/16.0));
	fnt.draw(str, cam.o_width() - fnt.width(str), 0.0);

	sprintf(str, "%.1f fps", fps);
	fnt.draw(str, 0.0, 0.0);

	glutSwapBuffers();
	glutPostRedisplay();
}

void calc_fps(timeval *tv)
{
	static timeval last;
	static unsigned count = 0;
	static unsigned usec = 0;

	count++;
	usec += tv->tv_usec - last.tv_usec;

	if (tv->tv_sec != last.tv_sec)
	{
		usec += 1000000;

		fps = float(count) * 1000000.0 / float(usec);

		count = 0;
		usec = 0;
	}

	last = *tv;
}

static void keyboard(unsigned char key, int, int)
{
	switch (key)
	{
	case '\033':
		exit(0);
		break;
	case 's':
		screenshot(unsigned(cam.get_width()),
			   unsigned(cam.get_height()));
		break;
	case 'w':
		glutReshapeWindow(cfg.get_int("width"), cfg.get_int("height"));
		break;
	case 'f':
		glutFullScreen();
		break;
	case ' ':
		key |= KEY_USE;
		break;
	}
}

static void keydown(int key, int, int)
{
	switch (key)
	{
	case GLUT_KEY_LEFT:
		keys |= KEY_LEFT;
		break;
	case GLUT_KEY_UP:
		keys |= KEY_UP;
		break;
	case GLUT_KEY_RIGHT:
		keys |= KEY_RIGHT;
		break;
	case GLUT_KEY_DOWN:
		keys |= KEY_DOWN;
		break;
	}
}

static void keyup(int key, int, int)
{
	switch (key)
	{
	case GLUT_KEY_LEFT:
		keys &= ~KEY_LEFT;
		break;
	case GLUT_KEY_UP:
		keys &= ~KEY_UP;
		break;
	case GLUT_KEY_RIGHT:
		keys &= ~KEY_RIGHT;
		break;
	case GLUT_KEY_DOWN:
		keys &= ~KEY_DOWN;
		break;
	}
}
