/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>

#include "model.h"
#include "gstring.h"
#include "misc.h"

model::model()
{
	bone = 0;
	tri = 0;
}

model::~model()
{
	if (bone) delete [] bone;
	if (tri) delete [] tri;
}

int model::load(const char *file)
{
	FILE *fp;
	unsigned i;
	gstring *gstr;
	char name[1024];

	sprintf(name, "models/%s", file);

	if ( !(fp = fopen(name, "rb")) )
		return -1;

	if ( fread(name, 1, 4, fp) != 4 ||
	     strncmp(name, "CJPM", 4) )
	{
		fclose(fp);
		return -1;
	}

	if (!(gstr = load_string(fp)))
	{
		fclose(fp);
		return -1;
	}

	if (!(tex = gstr->tex()))
	{
		fclose(fp);
		return -1;
	}

	if (get_fixed(fp, &rad))
	{
		fclose(fp);
		return -1;
	}

	if (signed(nfra = get_int(fp)) < 0)
	{
		fclose(fp);
		return -1;
	}

	if (signed(i = get_int(fp)) < 0)
	{
		fclose(fp);
		return -1;
	}

	if (bone) delete [] bone;
	bone = new m_bone[nbone = i];

	for (i = 0; i < nbone; i++)
	{
		if (bone[i].load(fp, nfra))
		{
			fclose(fp);
			return -1;
		}
	}

	if (signed(i = get_int(fp)) < 0)
	{
		fclose(fp);
		return -1;
	}

	if (tri) delete [] tri;
	tri = new m_triangle[ntri = i];

	for (i = 0; i < ntri; i++)
	{
		if (tri[i].load(fp, this))
		{
			fclose(fp);
			return -1;
		}
	}

	fclose(fp);

	return 0;
}

m_bone::m_bone()
{
	nvtx = 0;
	vtx = 0;
	fra = 0;
}

m_bone::~m_bone()
{
	if (vtx) delete [] vtx;
	if (fra) delete [] fra;
}

int m_bone::load(FILE *fp, unsigned n)
{
	unsigned i;

	if (fra) delete [] fra;
	fra = new m_frame[n];

	for (i = 0; i < n; i++)
	{
		if (fra[i].load(fp))
			return -1;
	}

	if (signed(i = get_int(fp)) < 0)
		return -1;

	if (vtx) delete [] vtx;
	vtx = new m_vertex[nvtx = i];

	for (i = 0; i < nvtx; i++)
	{
		if (vtx[i].load(fp))
			return -1;
	}

	return 0;
}

int m_frame::load(FILE *fp)
{
	if (pos.load(fp))
		return -1;

	if (rot.load(fp))
		return -1;

	return 0;
}

int m_triangle::load(FILE *fp, model *mdl)
{
	for (unsigned i = 0; i < 3; i++)
	{
		if (vtx[i].load(fp, mdl))
			return -1;
	}

	return 0;
}

int m_vertex_index::load(FILE *fp, model *mdl)
{
	unsigned i;
	const m_bone *b;

	if ((i = get_int(fp)) >= mdl->bones())
		return -1;
	b = mdl->find_bone(i);

	if ((i = get_int(fp)) >= b->vertices())
		return -1;
	vtx = b->find_vertex(i);

	if (get_fixed(fp, &u))
		return -1;

	if (get_fixed(fp, &v))
		return -1;

	return 0;
}

int m_vertex::load(FILE *fp)
{
	if (o_vert.load(fp))
		return -1;

	if (o_norm.load(fp))
		return -1;

	o_norm.unit();

	return 0;
}

void model::draw(unsigned k0, unsigned k1, float t)
{
	unsigned i;

	tex->glbind();

	for (i = 0; i < nbone; i++)
	{
		bone[i].draw(k0, k1, t);
	}

	glBegin(GL_TRIANGLES);

	for (i = 0; i < ntri; i++)
	{
		tri[i].draw();
	}

	glEnd();
}

void m_bone::draw(unsigned k0, unsigned k1, float t)
{
	matrix m;

	if (k0 == k1 || !t)
	{
		m.set(fra[k0].position(), fra[k0].rotation());
	}
	else
	{
		vector p, r;
		const vector *v;

		v = fra[k0].position();
		p.sub(fra[k1].position(), v);
		p.mul(t);
		p.add(v);

		v = fra[k0].rotation();
		r.sub(fra[k1].rotation(), v);
		r.mul(t);
		r.add(v);

		m.set(&p, &r);
	}

	for (unsigned i = 0; i < nvtx; i++)
	{
		vtx[i].transform(&m);
	}
}

void m_triangle::draw()
{
	for (unsigned i = 0; i < 3; i++)
	{
		vtx[i].gltexture();
		vtx[i].glnormal();
		vtx[i].glvertex();
	}
}
