/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/time.h>
#include <stdlib.h>

#include "sky.h"
#include "texture.h"
#include "camera.h"

stars::stars() : red(4.0), blue(8.0)
{
	tex = 0;
}

stars::~stars()
{
}

void stars::init()
{
	static const float ambient[4] = { 0.375, 0.375, 0.375, 1.0 };

	red.init(GL_LIGHT0, 32.0, 24.0, 16.0);
	blue.init(GL_LIGHT1, 16.0, 24.0, 32.0);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
}

star::star(float z)
{
	tex = 0;
	dist = z;
}

star::~star()
{
//	if (tex) glDeleteTextures(1, &tex);
}

#define RAD ((STAR_TEX_WIDTH-1.0)/2.0)
#define RAD2 (RAD*RAD)

void star::init(GLenum li, float red, float green, float blue)
{
	float diffuse[4];
	unsigned char *pixels, *p;
	unsigned i, j;
	float f;

	light = li;

	pixels = (unsigned char *)alloca(STAR_TEX_WIDTH * STAR_TEX_WIDTH * 3);

	p = pixels;

	for (i=0; i < STAR_TEX_WIDTH; i++)
	{
		double x = (double)(i) - RAD;

		for (j=0; j < STAR_TEX_WIDTH; j++)
		{
			double r, g, b;

			double y = (double)(j) - RAD;

			double z = ((x * x) + (y * y)) / RAD2;
			z = (1.0 / z) - 1.0;
	
			if (z < 0) z = 0.0;

			r = z * red;
			g = z * green;
			b = z * blue;

			if (r > 255.0) r = 255.0;
			if (g > 255.0) g = 255.0;
			if (b > 255.0) b = 255.0;

			*(p++) = (int)r;
			*(p++) = (int)g;
			*(p++) = (int)b;
		}
	}

	if (!tex) glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D, tex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, STAR_TEX_WIDTH, STAR_TEX_WIDTH,
		     0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glEnable(light);

	f = red;
	if (green > f) f = green;
	if (blue > f) f = blue;

	diffuse[0] = red / f;
	diffuse[1] = green / f;
	diffuse[2] = blue / f;
	diffuse[3] = 0.0;
	glLightfv(light, GL_DIFFUSE, diffuse);

	set(0.0, 0.0);
}

void star::set(float a, float b)
{
	matrix m;
	vector v;

	m.set();
	m.rotate_z(a);
	m.rotate_y(b);

	v.set(-1.0, -1.0, dist);
	vec[0].rotate(&v, &m);

	v.set(-1.0, 1.0, dist);
	vec[1].rotate(&v, &m);

	v.set(1.0, 1.0, dist);
	vec[2].rotate(&v, &m);

	v.set(1.0, -1.0, dist);
	vec[3].rotate(&v, &m);

	v.set(0.0, 0.0, 1.0);
	vec[4].rotate(&v, &m);
}

void stars::draw(camera *cam)
{
	glPushMatrix();
	glTranslatef(cam->pos.x(), cam->pos.y(), cam->pos.z());

//	glDisable(GL_FOG);
//	glDisable(GL_LIGHTING);

	red.draw();
	blue.draw();

	glEnable(GL_FOG);
	glEnable(GL_LIGHTING);

	if (tex)
	{
		tex->glbind();

		float max = cam->get_far();
		float top = max / td;
#if 0

/* We have two methods for drawing the sky.    *
 * The first is to simple draw one big square. */

		glBegin(GL_QUADS);
		glNormal3f(0.0, 0.0, 1.0);

		glTexCoord2f(0.0, 0.0);
		glVertex3f(-max, -max, top);

		glTexCoord2f(0.0, tn);
		glVertex3f(-max, max, top);

		glTexCoord2f(tn, tn);
		glVertex3f(max, max, top);

		glTexCoord2f(tn, 0.0);
		glVertex3f(max, -max, top);

		glEnd();
#else

/* However to reduce the problems with OpenGL implentations without  *
 * a proper fog implentation we have a second method that draws four *
 * triangles.                                                        */

		float tn2 = tn / 2.0;

		glBegin(GL_TRIANGLES);
		glNormal3f(0.0, 0.0, 1.0);


		glTexCoord2f(0.0, 0.0);
		glVertex3f(-max, -max, top);

		glTexCoord2f(0.0, tn);
		glVertex3f(-max, max, top);

		glTexCoord2f(tn2, tn2);
		glVertex3f(0, 0, top);


		glTexCoord2f(0.0, tn);
		glVertex3f(-max, max, top);

		glTexCoord2f(tn, tn);
		glVertex3f(max, max, top);

		glTexCoord2f(tn2, tn2);
		glVertex3f(0, 0, top);


		glTexCoord2f(tn, tn);
		glVertex3f(max, max, top);

		glTexCoord2f(tn, 0.0);
		glVertex3f(max, -max, top);

		glTexCoord2f(tn2, tn2);
		glVertex3f(0, 0, top);


		glTexCoord2f(tn, 0.0);
		glVertex3f(max, -max, top);

		glTexCoord2f(0.0, 0.0);
		glVertex3f(-max, -max, top);

		glTexCoord2f(tn2, tn2);
		glVertex3f(0, 0, top);

		glEnd();
#endif
	}

	glPopMatrix();
}

void star::draw()
{
	float position[4];

	glBindTexture(GL_TEXTURE_2D, tex);

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0);
	vec[0].glvertex();

	glTexCoord2f(0.0, 1.0);
	vec[1].glvertex();

	glTexCoord2f(1.0, 1.0);
	vec[2].glvertex();

	glTexCoord2f(1.0, 0.0);
	vec[3].glvertex();

	glEnd();

	position[0] = vec[4].x();
	position[1] = vec[4].y();
	position[2] = vec[4].z();
	position[3] = 0.0;
	glLightfv(light, GL_POSITION, position);
}
