/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <png.h>

#include <GL/glu.h>

#include "texture.h"
#include "config.h"
#include "misc.h"

texture::texture()
{
	tex = 0;
	alpha = 0;
}

texture::~texture()
{
	if (tex) glDeleteTextures(1, &tex);
}

int texture::load(const char *file)
{
	FILE *fp;
	char fname[1024];

	sprintf(fname, "textures/%s", file);

	if ( !(fp = fopen(fname, "rb")) )
		return -1;

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	png_infop info_ptr = png_create_info_struct(png_ptr);

	if (setjmp(png_ptr->jmpbuf))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, 0);
		fclose(fp);
		return -1;
	}

	png_init_io(png_ptr, fp);

	png_read_info(png_ptr, info_ptr);
	png_set_strip_16(png_ptr);
	png_set_expand(png_ptr);
	png_read_update_info(png_ptr, info_ptr);

	png_uint_32 width = png_get_image_width(png_ptr, info_ptr);
	png_uint_32 height = png_get_image_height(png_ptr, info_ptr);
	png_uint_32 rowbytes = png_get_rowbytes(png_ptr, info_ptr);
	png_byte channels = png_get_channels(png_ptr, info_ptr);

	GLenum fmt;

	switch(channels)
	{
	case 1:
		fmt = GL_LUMINANCE;
		break;
	case 2:
		fmt = GL_LUMINANCE_ALPHA;
		break;
	case 3:
		fmt = GL_RGB;
		break;
	case 4:
		fmt = GL_RGBA;
		break;
	default:
		png_destroy_read_struct(&png_ptr, &info_ptr, 0);
		fclose(fp);
		return -1;
	}

	png_bytep pixels = (png_bytep)alloca(height * rowbytes);
	png_bytep *rows = (png_bytep *)alloca(height * sizeof(png_bytep));

	for (png_uint_32 row = 0; row < height; row++)
	{
		rows[row] = &pixels[row * rowbytes];
	}

	png_read_image(png_ptr, rows);

	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, 0);

	fclose(fp);

	if (!tex) glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D, tex);

	int filt = cfg.get_int("filter");
	GLint min_filt, mag_filt;

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	if (cfg.get_int("mipmap"))
	{
		gluBuild2DMipmaps(GL_TEXTURE_2D, channels, width, height,
				  fmt, GL_UNSIGNED_BYTE, pixels);

		switch (filt)
		{
		case 0:
			min_filt = GL_NEAREST_MIPMAP_NEAREST;
			mag_filt = GL_NEAREST;
			break;
			
		case 1:
			min_filt = GL_LINEAR_MIPMAP_NEAREST;
			mag_filt = GL_LINEAR;
			break;

		default:
			min_filt = GL_LINEAR_MIPMAP_LINEAR;
			mag_filt = GL_LINEAR;
			break;
		}

	}
	else
	{
		glTexImage2D(GL_TEXTURE_2D, 0, channels, width, height, 0,
			     fmt, GL_UNSIGNED_BYTE, pixels);

		if (filt)
		{
			mag_filt = min_filt = GL_LINEAR;
		}
		else
		{
			mag_filt = min_filt = GL_NEAREST;
		}
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filt);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filt);

	alpha = !(channels & 1);

	return 0;
}
