/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __VECTOR_H__
#define __VECTOR_H__

#include <math.h>
#include <stdio.h>

#include <GL/gl.h>
#include <AL/al.h>

class matrix;

class vector
{
	friend class matrix;

 private:
	float v[3];

 public:
	inline float x() const { return v[0]; }
	inline float y() const { return v[1]; }
	inline float z() const { return v[2]; }

	inline void set(const vector *p)
		{ v[0] = p->v[0]; v[1] = p->v[1]; v[2] = p->v[2]; }
	inline void set(float x, float y, float z)
		{ v[0] = x; v[1] = y ; v[2] = z; }
	inline void set_x(float x) { v[0] = x; }
	inline void set_y(float y) { v[1] = y; }
	inline void set_z(float z) { v[2] = z; }

	inline float val2() const
		{ return (v[0] * v[0]) + (v[1] * v[1]) + (v[2] * v[2]); }

	inline float val() const
		{ return sqrt(val2()); }

	inline void neg()
		{ v[0] = - v[0]; v[1] = - v[1]; v[2] = - v[2]; }

	inline void neg(const vector *p)
		{ v[0] = - p->v[0]; v[1] = - p->v[1]; v[2] = - p->v[2]; }

	inline void add(const vector *p)
		{ v[0] += p->v[0]; v[1] += p->v[1]; v[2] += p->v[2]; }

	inline void add(const vector *l, const vector *r)
		{ set(l->v[0] + r->v[0], l->v[1] + r->v[1], l->v[2] + r->v[2]); }

	inline void sub(const vector *p)
		{ v[0] -= p->v[0]; v[1] -= p->v[1]; v[2] -= p->v[2]; }

	inline void sub(const vector *l, const vector *r)
		{ set(l->v[0] - r->v[0], l->v[1] - r->v[1], l->v[2] - r->v[2]); }

	inline void mul(float s)
		{ v[0] *= s; v[1] *= s; v[2] *= s; }

	inline void mul(const vector *v, float s)
		{ set(v->v[0] * s, v->v[1] * s, v->v[2] * s); }

	inline void div(float s)
		{ v[0] /= s; v[1] /= s; v[2] /= s; }

	inline void div(const vector *p, float s)
		{ set(p->v[0] / s, p->v[1] / s, p->v[2] / s); }

	inline void unit()
		{ float s = val2(); if (s) div(sqrt(s)); }

	inline float dot(const vector *p) const
		{ return (v[0] * p->v[0]) + (v[1] * p->v[1]) + (v[2] * p->v[2]); }

	inline void cross(const vector *p)
		{ set(v[1] * p->v[2] - v[2] * p->v[1],
		      v[2] * p->v[0] - v[0] * p->v[2],
		      v[0] * p->v[1] - v[1] * p->v[0]); }

	inline void cross(const vector *l, const vector *r)
		{ set(l->v[1] * r->v[2] - l->v[2] * r->v[1],
		      l->v[2] * r->v[0] - l->v[0] * r->v[2],
		      l->v[0] * r->v[1] - l->v[1] * r->v[0]); }

	inline void glvertex() const { glVertex3fv(v); }
	inline void glnormal() const { glNormal3fv(v); }

	inline void alsource(ALuint sid) /* const */
		{ alSourcefv(sid, AL_POSITION, v); }
	inline void allistener() /* const */
		{ alListenerfv(AL_POSITION, v); }

	void get_x(const matrix *);
	void get_y(const matrix *);
	void get_z(const matrix *);
	void get_w(const matrix *);

	void transform(const matrix *);
	void transform(const vector *, const matrix *);

	void rotate(const matrix *);
	void rotate(const vector *, const matrix *);

	int load(FILE *);
	int load_fixed(FILE *);
	int load_float(FILE *);
};

#endif
