/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

class landscape;

struct Tag
{
        char *token;
        int (landscape::*func)(char *s);
};

class texture
{
public:
	texture *next;
	char *name;
	unsigned idx;

	texture(const char *);
	~texture();

	int save(FILE *);
};

class vector 
{
public:
	float x, y, z;

	inline void set(float i, float j, float k)
		{ x = i; y = j; z = k; }
	inline void set(vector *v)
		{ x = v->x; y = v->y; z = v->z; }
	inline void add(vector *v)
		{ x += v->x; y += v->y; z += v->z; }
};

class vertex
{
public:
	int x, y, z;
	int u, v;

	int set(vector *, float, float);
	int save(FILE *, int);
};

class triangle
{
public:
	triangle *next;
	texture *tex;
	vertex vtx[3];

	int save(FILE *, int);

	void min_max(int *, int *);
};

class square 
{
private:
	triangle *tris;
	int z;

public:
	square();
	~square();

	inline void add_triangle(triangle *t)
		{ t->next = tris; tris = t; }

	int save(FILE *);

	void fix_heights();
};

class landscape
{
private:
	texture *texs;
	square sqr[256][256];
	unsigned n_ply;
	unsigned n_tri;
	unsigned n_out;
	unsigned n_bad;
	unsigned n_sml;

public:
	landscape();
	~landscape();

	int load(const char *);
	int save(const char *);

	void fix_heights();
	void add_triangle(triangle *);
	texture *find_texture(const char *);
	void summary();

private:

/* for ac3d loading */

	FILE *fp;
	unsigned last_num_kids;
	char *last_name;
	texture *last_texture;
	vector *last_vertices;
	vector current_loc;

	static Tag top_tags[];
	static Tag obj_type_tags[];
	static Tag object_tags[];
	static Tag surf_tag[];
	static Tag surface_tags[];
	
	int search(Tag *, char *);
	char *read_line();
	void skip_spaces(char **);
	void skip_quotes(char **);

	int do_material(char *);
	int do_object(char *);

	int do_obj_world(char *);
	int do_obj_poly(char *);
	int do_obj_group(char *);

	int do_name(char *);
	int do_data(char *);
	int do_texture(char *);
	int do_texrep(char *);
	int do_rot(char *);
	int do_loc(char *);
	int do_url(char *);
	int do_numvert(char *);
	int do_numsurf(char *);
	int do_kids(char *);

	int do_surf(char *s);

	int do_mat(char *s);
	int do_refs(char *s);
};

static char *progname;
static int tile = 0;

int main(int argc, char **argv)
{
        landscape map;

	progname = argv[0];

        if (argc < 3 || argc > 4)
        {
                fprintf(stderr, "Usgae: %s <src> <dst> [<exp>]\n", argv[0]);
                return 1;
        }

	if (argc > 3)
	{
		unsigned e = atoi(argv[3]);
		if (e > 2) e = 0;
		tile = 1U << e;
	}

        if (map.load(argv[1]))
        {
                fprintf(stderr, "Failed to load map \"%s\"\n", argv[1]);
                return 1;
        }

	map.fix_heights();

        if (map.save(argv[2]))
        {
                fprintf(stderr, "Failed to save map \"%s\"\n", argv[2]);
                return 1;
        }

	map.summary();

        return 0;
}

landscape::landscape()
{
	last_name = 0;
	last_vertices = 0;
	n_ply = 0;
	n_tri = 0;
	n_out = 0;
	n_bad = 0;
	n_sml = 0;
}

landscape::~landscape()
{
}

/*
 * Load an ac3d file
 *
 * This is based upon the ac3d loader written by Steve Baker.
 */

#define OBJ_WORLD  0
#define OBJ_POLY   1
#define OBJ_GROUP  2

Tag landscape::top_tags[] =
{
	{ "MATERIAL", &landscape::do_material },
	{ "OBJECT"  , &landscape::do_object   },
	{ 0, 0 }
};

Tag landscape::obj_type_tags[] =
{
	{ "world", &landscape::do_obj_world },
	{ "poly" , &landscape::do_obj_poly  },
	{ "group", &landscape::do_obj_group },
	{ 0, 0 }
};

Tag landscape::object_tags[] =
{
	{ "name"   , &landscape::do_name    },
	{ "data"   , &landscape::do_data    },
	{ "texture", &landscape::do_texture },
	{ "texrep" , &landscape::do_texrep  },
	{ "rot"    , &landscape::do_rot     },
	{ "loc"    , &landscape::do_loc     },
	{ "url"    , &landscape::do_url     },
	{ "numvert", &landscape::do_numvert },
	{ "numsurf", &landscape::do_numsurf },
	{ "kids"   , &landscape::do_kids    },
	{ 0, 0 }
};

Tag landscape::surf_tag[] =
{
	{ "SURF", &landscape::do_surf },
	{ 0, 0 }
};

Tag landscape::surface_tags[] =
{
	{ "mat" , &landscape::do_mat  },
	{ "refs", &landscape::do_refs },
	{ 0, 0 }
};

int landscape::load(const char *file)
{
	int firsttime = 1;
	char *s;

	last_num_kids = 0;
	current_loc.set(0.0, 0.0, 0.0);
	

        if ( !(fp = fopen(file, "r")) )
                return -1;

	while ( (s = read_line()) )
	{
		if (firsttime)
		{
			firsttime = 0;

			if ( strncasecmp(s, "AC3D", 4) )
			{
				fprintf(stderr, "%s: Input is not an AC3D format file.\n", progname);
				exit(1);
			}
		}
		else
		{
			search(top_tags, s);
		}
	}

	return 0;
}

int landscape::search(Tag *tags, char *s)
{
	skip_spaces(&s);

	for (int i = 0; tags[i].token; i++)
	{
		if ( !strncasecmp(tags[i].token, s, strlen(tags[i].token)) )
		{
			s += strlen(tags[i].token);

			skip_spaces(&s);

			return (*(this->*tags[i].func))(s);
		}

	}
		
	fprintf(stderr, "%s: Unrecognised token '%s'\n", progname, s);
	exit(1);

	return 1;
}

int landscape::do_material(char *s)
{
// ignore material.
	return 0;
}

int landscape::do_object(char *s)
{
	vector v;
	char *t;

	v.set(&current_loc);

	search(obj_type_tags, s);

	last_name = 0;
	last_texture = 0;

	while ( (t = read_line()) )
	{
		if ( search(object_tags, t) )
			break;
	}

	if (last_vertices)
	{
		delete [] last_vertices;
		last_vertices = 0;
	}

	unsigned num_kids = last_num_kids;

	for (unsigned i = 0; i < num_kids; i++)
	{
		t = read_line();
		search(top_tags, t);
	}

	current_loc.set(&v);

	return 0;
}

int landscape::do_obj_world(char *s)
{
	return OBJ_WORLD;
} 

int landscape::do_obj_poly(char *s)
{
	return OBJ_POLY;
}

int landscape::do_obj_group(char *s)
{
	return OBJ_GROUP;
}

int landscape::do_name(char *s)
{
	skip_quotes(&s);

	if (last_name) delete last_name;
	strcpy(last_name = new char[strlen(s) + 1], s);

	return 0;
}

int landscape::do_data(char *s)
{
	int len = strtol(s, 0, 0);

	for (int i = 0; i < len; i++)
	{
		fgetc(fp);
	}

	return 0;
}

int landscape::do_texture(char *s)
{
	char *p;

	skip_quotes(&s);

	if ( (p = strrchr(s, '/')) )
	     s = p + 1;

	last_texture = find_texture(s);

	return 0;
}

int landscape::do_texrep(char *s)
{
// ignore texrep.
	return 0;
}

int landscape::do_rot(char *s)
{
	fprintf(stderr, "%s: Warning, rot record not implemented.\n",
		progname);

	return 0;
}

int landscape::do_loc(char *s)
{
	vector v;

	if (sscanf(s, "%f %f %f", &v.x, &v.y, &v.z) != 3)
	{
		fprintf(stderr, "%s: Illegal loc record:.\n", progname);
		exit(1);
	}

	current_loc.add(&v);

	return 0;
}

int landscape::do_url(char *s)
{
// ignore url.
	return 0;
}

int landscape::do_numvert(char *s)
{
	unsigned num_vert = strtol(s, 0, 0 );

	if (num_vert)
	{
		last_vertices = new vector[num_vert];

		for (unsigned i = 0; i < num_vert ; i++)
		{
			char *t;

			if ( !(t = read_line()) )
			{
				fprintf(stderr, "%s: Unexpected end of file\n",
					progname);
				exit(1);
			}

			vector v;

			if (sscanf(t, "%f %f %f", &v.x, &v.y, &v.z) != 3)
			{
				fprintf(stderr, "%s: Illegal vertex record.\n",
					progname);
				exit(1);
			}

			v.add(&current_loc);

			last_vertices[i].set(&v);
		}
	}

	return 0;
}

int landscape::do_numsurf(char *s)
{
	int ns = strtol(s, 0, 0);

	for (int i = 0; i < ns; i++)
	{
		char *t;

		if ( !(t = read_line()) )
		{
			fprintf(stderr, "%s: Unexpected end of file\n",
				progname);
			exit(1);
		}

		search(surf_tag, t);
	}

	return 0;
}

int landscape::do_kids(char *s)
{
	last_num_kids = strtol(s, 0, 0);

	return 1;
}

int landscape::do_surf(char *s)
{
	char *t;

	while ( (t = read_line()) )
	{
		if (search(surface_tags, t))
			break;
	}

	return 0;
}

int landscape::do_mat(char *s)
{
// ignore mat.
	return 0;
}

int landscape::do_refs(char *s)
{
	triangle *tri = 0;
	int nrefs = strtol(s, 0, 0);

	n_ply++;

	if (nrefs == 3)
	{
		tri = new triangle;
		tri->tex = last_texture;
	}
	else
	{
		n_tri++;
	}

	for (int i = 0; i < nrefs; i++)
	{
		char *t;

		if ( !(t = read_line()) )
		{
			fprintf(stderr, "%s: Unexpected end of file\n",
				progname);
			exit(1);
		}

		int n;
		float u, v;

		if (sscanf(t, "%d %f %f", &n, &u, &v) != 3)
		{
			fprintf(stderr, "%s: Illegal ref record.\n",
				progname);
			exit(1);
		}

		if (tri)
		{
			if (tri->vtx[i].set(&last_vertices[n], u, 1.0 - v))
			{
				delete tri;
				tri = 0;
				n_out++;
			}
		}
	}

	if (tri)
	{
		add_triangle(tri);
	}

	return 1;
}

char *landscape::read_line()
{
	static char buffer[1024];

	while (fgets(buffer, 1024, fp))
	{
		char *s;

		if ( (s = strchr(buffer, '\n')) )
			*s = '\0';
			
		s = buffer;

		skip_spaces(&s);

		if ( *s && *s != '#' || *s != ';' )
			return s;
	}

	return 0;
}

void landscape::skip_spaces(char **s)
{
	while ( **s == ' ' || **s == '\t' )
		(*s)++;
}

void landscape::skip_quotes(char **s)
{
	skip_spaces(s);

	if (**s == '\"')
	{
		(*s)++;

		char *t = *s;

		while ( *t != '\0' && *t != '\"' )
			t++;

		if (*t != '\"')
		{
			fprintf(stderr, "%s: Mismatched double-quote ('\"') in '%s'\n", *s, progname);
			exit(1);
		}

		*t = '\0';
	}
	else
	{
		fprintf(stderr, "%s: Expected double-quote ('\"') in '%s'\n", *s, progname);
		exit(1);
	}
}

/*
 * Now for a version of my map code.
 */

int landscape::save(const char *file)
{
	int i, j;
	texture *t;

	if ( !(fp = fopen(file, "w")) )
		return -1;

	if (fwrite("CJPL", 1, 4, fp) != 4)
	{
		fclose(fp);
		return -1;
	}

	i = 0;
	for (t = texs; t; t = t->next)
	{
		t->idx = ++i;
	}

	if (fputc(i, fp) != int(i))
		return -1;

	for (t = texs; t; t = t->next)
	{
		if (t->save(fp))
		{
			fclose(fp);
			return -1;
		}
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (sqr[j][i].save(fp))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	fclose(fp);

	return 0;
}


void landscape::fix_heights()
{
	unsigned i, j;

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			sqr[j][i].fix_heights();
		}
	}
}

void landscape::add_triangle(triangle *tri)
{
	int x, y;
	int x0, y0, z0;
	int x1, y1, z1;
	int x2, y2, z2;
	unsigned i;

	x1 = tri->vtx[1].x - tri->vtx[0].x;
	y1 = tri->vtx[1].y - tri->vtx[0].y;
	z1 = tri->vtx[1].z - tri->vtx[0].z;

	x2 = tri->vtx[2].x - tri->vtx[0].x;
	y2 = tri->vtx[2].y - tri->vtx[0].y;
	z2 = tri->vtx[2].z - tri->vtx[0].z;

	x0 = y1 * z2 - z1 * y2;
	y0 = z1 * x2 - x1 * z2;
	z0 = x1 * y2 - y1 * x2;

	if (x0 == 0 && y0 == 0 && z0 == 0)
	{
		delete tri;
		n_sml++;
		return;
	}

	if (tile)
	{
		x1 = abs(x0);
		y1 = abs(y0);
		z1 = abs(z0);

		for (i = 0; i < 3; i++)
		{
			tri->vtx[i].u = tri->vtx[i].x;
			tri->vtx[i].v = tri->vtx[i].y;
		}
		x = z0;

		if (x1 > y1)
		{
			if (x1 > z1)
			{
				for (i = 0; i < 3; i++)
				{
					tri->vtx[i].u = tri->vtx[i].y;
					tri->vtx[i].v = -tri->vtx[i].z;
				}
				x = x0;
			}
		}
		else
		{
			if (y1 > z1)
			{
				for (i = 0; i < 3; i++)
				{
					tri->vtx[i].u = tri->vtx[i].x;
					tri->vtx[i].v = -tri->vtx[i].z;
				}
				x = y0;
			}
		}

		if (x >= 0)
		{
			for (i = 0; i < 3; i++)
			{
				tri->vtx[i].u = -tri->vtx[i].u;
			}
		}

		for (i = 0; i < 3; i++)
		{
			tri->vtx[i].u = (tri->vtx[i].u * 4) / tile;
			tri->vtx[i].v = (tri->vtx[i].v * 4) / tile;
		}

		x = tri->vtx[0].u;
		y = tri->vtx[0].v;

		for (i = 1; i < 3; i++)
		{
			if (tri->vtx[i].u < x) x = tri->vtx[i].u;
			if (tri->vtx[i].v < y) y = tri->vtx[i].v;
		}

		x &= -64;
		y &= -64;

		for (i = 0; i < 3; i++)
		{
			tri->vtx[i].u -= x;
			tri->vtx[i].v -= y;
		}
	}

	x = ((tri->vtx[0].x + tri->vtx[1].x + tri->vtx[2].x) / 3) & -16;
	y = ((tri->vtx[0].y + tri->vtx[1].y + tri->vtx[2].y) / 3) & -16;

	if (x0 >= 0 &&
	    (tri->vtx[0].x & 15) == 0 &&
	    tri->vtx[1].x == tri->vtx[0].x &&
	    tri->vtx[2].x == tri->vtx[0].x)
	{
		x -= 16;
	}

	if (y0 >= 0 &&
	    (tri->vtx[0].y & 15) == 0 &&
	    tri->vtx[1].y == tri->vtx[0].y &&
	    tri->vtx[2].y == tri->vtx[0].y)
	{
		y -= 16;
	}

	if (x < 0) x = 0;
	if (y < 0) x = 0;

	if (x > 4080) x = 4080;
	if (y > 4080) y = 4080;

	for (i = 0; i < 3; i++)
	{
		tri->vtx[i].x -= x; tri->vtx[i].y -= y;
	}

	for (i = 0; i < 3; i++)
	{
		if (tri->vtx[i].x < 0 || tri->vtx[i].x > 16 ||
		    tri->vtx[i].y < 0 || tri->vtx[i].y > 16)
		{
			delete tri;
			n_bad++;
			return;
		}
	}

	sqr[y>>4][x>>4].add_triangle(tri);
}

texture *landscape::find_texture(const char *s)
{
	texture *t;

	for (t = texs; t; t = t->next)
	{
		if (!strcmp(t->name, s))
			return t;
	}

	t = new texture(s);
	t->next = texs;
	texs = t;

	return t;
}

void landscape::summary()
{
	printf("%u polygons\n", n_ply);
	printf("%u dropped (%u not triangles)\n",
	       n_tri + n_out + n_bad + n_sml, n_tri);
        printf("           (%u outside map)\n", n_out);
	printf("           (%u intersect grid)\n", n_bad);
	printf("           (%u too small)\n", n_sml);
}

texture::texture(const char *s)
{
	strcpy(name = new char[strlen(s) + 1], s);
}

texture::~texture()
{
	delete name;
}

int texture::save(FILE *fp)
{
	unsigned i;

	i = strlen(name);

	if (fputc(i, fp) != int(i))
		return -1;

	if (fwrite(name, 1, i, fp) != i)
		return -1;

	return 0;
}

square::square()
{
	tris = 0;
	z = 0;
}

square::~square()
{
	while (tris)
	{
		triangle *t = tris;
		tris = t->next;
		delete t;
	}
}

int square::save(FILE *fp)
{
	unsigned i;
	triangle *t;

	if (fputc(z, fp) != z)
                return -1;

	i = 0;
	for (t = tris; t; t = t->next)
	{
		++i;
	}

	if (fputc(i, fp) != int(i))
                return -1;

	for (t = tris; t; t = t->next)
	{
		if (t->save(fp, z))
			return -1;
	}

	return 0;
}

void square::fix_heights()
{
	triangle *t;
	int min = z;
	int max = z;

	for (t = tris; t; t = t->next)
	{
		t->min_max(&min, &max);
	}

	if ((max - z) > 127) z = max - 127;
	if ((min - z) < -128) z = min + 128;
}

int triangle::save(FILE *fp, int z)
{
	unsigned i;

	i = tex->idx;

	if (fputc(i, fp) != int(i))
		return -1;

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].save(fp, z))
			return -1;
	}

	return 0;
}

int vertex::set(vector *p, float i, float j)
{
	x = int(floor(((p->x + 128.0) * 16.0) + 0.5));
	y = int(floor(((128.0 - p->z) * 16.0) + 0.5));
	z = int(floor((p->y * 16.0) + 0.5));
	u = int(floor((i * 64.0) + 0.5));
	v = int(floor((j * 64.0) + 0.5));

	if (x < 0 || x > 4096 ||
	    y < 0 || y > 4096 ||
	    z < -128 || z > 382)
		return -1;

	return 0;
}

void triangle::min_max(int *minp, int *maxp)
{
	int min = *minp;
	int max = *maxp;
	unsigned i;

	for (i = 0; i < 3; i++)
	{
		int z = vtx[i].z;

		if (z < min) min = z;
		if (z > max) max = z;
	}

	*minp = min;
	*maxp = max;
}

int vertex::save(FILE *fp, int h)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	data.x = x;
	data.y = y;
	data.z = z - h;
	data.u = u;
	data.v = v;

	if (fwrite(&data, 1, 5, fp) != 5)
		return -1;

	return 0;
}
