%{

/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "compile.h"

%}

%union {
	int val;
	float flt;
	string *str;
	type *typ;
	expression *exp;
	statement *stmt;
	variable *var;
	function *fn;
}

%start prog

%token <flt> NUMBER
%token <str> TEXT
%token <str> ID

%token IF
%token UNLESS
%token ELSE
%token FOR
%token DO
%token WHILE
%token UNTIL
%token RETURN
%token THIS
%token VOID
%token FLOAT
%token VECTOR
%token STRING
%token ENTITY

%token AND
%token OR
%token EQ
%token NE
%token LE
%token GE

%type <val> base_type
%type <typ> complex_type
%type <var> var_list
%type <var> variable
%type <fn> func
%type <var> params
%type <var> param_list
%type <var> param
%type <stmt> block
%type <stmt> stmnts
%type <stmt> stmnt
%type <exp> expr
%type <exp> list_expr
%type <exp> assignment_expr
%type <exp> logical_or_expr
%type <exp> logical_and_expr 
%type <exp> inclusive_or_expr
%type <exp> exclusive_or_expr
%type <exp> and_expr
%type <exp> equality_expr
%type <exp> relational_expr
%type <exp> additive_expr
%type <exp> multiplicative_expr
%type <exp> unary_expr
%type <exp> postfix_expr
%type <exp> primary_expr
%type <flt> number

%%

prog : /* empty */ | decl prog ;

decl : complex_type var_list ';'
	{
		add_global_vars($1, $2);
	}
	| func ';'
	{
		this_fn = 0;
	}
	| func block
	{
		this_fn = 0;

		if ($1->var.defd)
			yyerrorf("redefinition of `%s'", $1->var.name->cstr);

		$1->var.used = 1;
		$1->var.defd = 1;
		$1->body = $2;
	}
;

complex_type: base_type
	{
		$$ = new_type($1, 0);
	}
	| complex_type '(' params ')'
	{
		$$ = new_type(T_FUNCTION, new_function($1, 0, $3));
	}
;

base_type: VOID { $$ = T_VOID; }
	| FLOAT { $$ = T_FLOAT; }
	| VECTOR { $$ = T_VECTOR; }
	| STRING { $$ = T_STRING; }
	| ENTITY { $$ = T_ENTITY; }
;

var_list : variable { $$ = $1 }
	| variable ',' var_list { $1->next = $3; $$ = $1; }
;

variable : ID { $$ = new_variable(new_type(T_FLOAT, 0), $1); } ;

func : complex_type ID '(' params ')'
	{
		this_fn = $$ = new_function($1, $2, $4);
	}
;

params : /* empty */ { $$ = 0; }
	| param_list { $$ = new_parameters($1); }
;

param_list : param { $$ = $1 }
	| param ',' param_list { $1->next = $3; $$ = $1; }
;

param : complex_type ID { $$ = new_variable($1, $2); } 
	| complex_type { $$ = new_variable($1, 0); }
;

block : '{' stmnts '}' { $$ = $2 } ;

stmnts : /* empty */ { $$ = 0 }
	| local_vars stmnts { $$ = $2 }
	| stmnt stmnts
	{
		$1->next = $2;
		$$ = $1
	}
;

local_vars : complex_type var_list ';'
	{
		add_local_vars($1, $2);
	}
;

stmnt : block
	{
		$$ = new_statement(S_BLOCK);
		$$->s1 = $1;
	}
	| expr ';'
	{
		$$ = new_statement(S_EXPR);
		$$->e1 = $1;
	}
	| IF '(' expr ')' block ELSE block
	{
		$$ = new_statement(S_IF);
		$$->e1 = $3;
		$$->s1 = $5;
		$$->s2 = $7;
	}
	| IF '(' expr ')' block
	{
		$$ = new_statement(S_IF);
		$$->e1 = $3;
		$$->s1 = $5;
	}
	| UNLESS '(' expr ')' block ELSE block
	{
		$$ = new_statement(S_UNLESS);
		$$->e1 = $3;
		$$->s1 = $5;
		$$->s2 = $7;
	}
	| UNLESS '(' expr ')' block
	{
		$$ = new_statement(S_UNLESS);
		$$->e1 = $3;
		$$->s1 = $5;
	}
	| WHILE '(' expr ')' block
	{
		$$ = new_statement(S_WHILE);
		$$->e1 = $3;
		$$->s1 = $5;
	}
	| UNTIL '(' expr ')' block
	{
		$$ = new_statement(S_UNTIL);
		$$->e1 = $3;
		$$->s1 = $5;
	}
	| DO block WHILE '(' expr ')' ';'
	{
		$$ = new_statement(S_DO_WHILE);
		$$->s1 = $2;
		$$->e1 = $5;
	}
	| DO block UNTIL '(' expr ')' ';'
	{
		$$ = new_statement(S_DO_UNTIL);
		$$->s1 = $2;
		$$->e1 = $5;
	}
	| FOR '(' expr ';' expr ';' expr ')' block
	{
		$$ = new_statement(S_FOR);
		$$->e1 = $3;
		$$->e2 = $5;
		$$->e3 = $7;
		$$->s1 = $9;
	}
	| RETURN expr ';'
	{
		if (this_fn->var.typ != ($2 ? $2->typ : T_VOID))
			yyerror("type mismatch");			
		$$ = new_statement(S_RETURN);
		$$->e1 = $2;
	}
;

expr : /* empty */ { $$ = 0; }
	| list_expr { $$ = $1; }
;

list_expr :  assignment_expr { $$ = $1; }
	| list_expr ',' assignment_expr
	{
		if ($3->prev)
			$$ = new_nop_expr($3);
		else
			$$ = $3;
		$$->prev = $1;
	}
;

assignment_expr : logical_or_expr { $$ = $1 }
	| postfix_expr '=' assignment_expr
	{
		int func = $1->func;

		if (func == E_DOTX || func == E_DOTY ||func == E_DOTZ)
			func = $1->left->func;
		if (func != E_LOCAL && func != E_GLOBAL)
			yyerror("bad lvalue");
		$$ = new_assign_expr($1, $3);
	}
;

logical_or_expr : logical_and_expr { $$ = $1 }
	| logical_or_expr OR logical_and_expr
	{
		$$ = new_binary_expr(B_LOR, $1, $3);
	}
;

logical_and_expr : inclusive_or_expr { $$ = $1 }
	| logical_and_expr AND inclusive_or_expr
	{
		$$ = new_binary_expr(B_LAND, $1, $3);
	}
;

inclusive_or_expr : exclusive_or_expr { $$ = $1 }
	| inclusive_or_expr '|' exclusive_or_expr
	{
		$$ = new_binary_expr(B_OR, $1, $3);
	}
;

exclusive_or_expr : and_expr { $$ = $1 }
	| exclusive_or_expr '^' and_expr
	{
		$$ = new_binary_expr(B_EOR, $1, $3);
	}
;

and_expr : equality_expr { $$ = $1 }
	| and_expr '&' equality_expr
	{
		$$ = new_binary_expr(B_AND, $1, $3);
	}
;

equality_expr : relational_expr { $$ = $1 }
	| equality_expr EQ relational_expr
	{
		$$ = new_binary_expr(B_EQ, $1, $3);
	}
	| equality_expr NE relational_expr
	{
		$$ = new_binary_expr(B_NE, $1, $3);
	}
;

relational_expr	: additive_expr { $$ = $1 }
	| relational_expr '<' additive_expr
	{
		$$ = new_binary_expr(B_LT, $1, $3);
	}
	| relational_expr '>' additive_expr
	{
		$$ = new_binary_expr(B_GT, $1, $3);
	}
	| relational_expr LE additive_expr
	{
		$$ = new_binary_expr(B_LE, $1, $3);
	}
	| relational_expr GE additive_expr
	{
		$$ = new_binary_expr(B_GE, $1, $3);
	}
;

additive_expr : multiplicative_expr { $$ = $1 }
	| additive_expr '+' multiplicative_expr
	{
		$$ = new_binary_expr(B_ADD, $1, $3);
	}
	| additive_expr '-' multiplicative_expr
	{
		$$ = new_binary_expr(B_SUB, $1, $3);
	}
;

multiplicative_expr : unary_expr { $$ = $1; }
	| multiplicative_expr '*' unary_expr
	{
		$$ = new_binary_expr(B_MUL, $1, $3);
	}
	| multiplicative_expr '/' unary_expr
	{
		$$ = new_binary_expr(B_DIV, $1, $3);
	}
;

unary_expr : postfix_expr { $$ = $1; }
	| '+' unary_expr
	{ 
		if ($2->typ->typ != T_FLOAT && $2->typ->typ != T_VECTOR)
			yyerror("type mismatch");
		$$ = $2;
	}
	| '-' unary_expr
	{ 
		$$ = new_unary_expr(U_NEG, $2);
	}
	| '~' unary_expr
	{
		$$ = new_unary_expr(U_INV, $2);
	}
	| '!' unary_expr
	{ 
		$$ = new_unary_expr(U_NOT, $2);
	}
;

postfix_expr : primary_expr { $$ = $1 }
	| postfix_expr '.' ID
	{
		switch($1->typ->typ)
		{
		case T_ENTITY:
			$$ = new_variable_expr($1, $3);	
			break;
		case T_VECTOR:
			$$ = new_vecdot_expr($1, $3);
			break;
		default:
			yyerrorf("request for member `%s' in non entity",
				 $3->cstr);
			$$ = $1;
			break;
		}
	}
	| postfix_expr '(' expr ')'
	{
		$$ = new_call_expr($1, $3);
	}
;

primary_expr : ID
	{
		$$ = new_variable_expr(0, $1);
	}
	| THIS
	{
		$$ = new_const_expr(new_this());
	}
	| NUMBER
	{
		$$ = new_const_expr(new_float_const($1));
	}
	| '\'' number number number '\'' 
	{
		$$ = new_const_expr(new_vector_const($2, $3, $4));
	}
	| TEXT
	{
		$$ = new_const_expr(new_string_const($1));
	}
	| '(' list_expr ')'
	{
		$$ = $2;
	}
;

number : NUMBER { $$ = $1; }
	| '+' NUMBER { $$ = $2; }
	| '-' NUMBER { $$ = - $2; }
;

%%

#define YY_NO_UNPUT
#include "compile.yy.c"

static const char *progname;
static int error = 0;

int main(int argc, char **argv)
{
	int fds[2], fd = -1;
	pid_t pid;

	progname = argv[0];

	if (argc > 3)
	{
		fprintf(stderr, "Usage: %s <input> <output>\n", progname);
		return 1;
	}

	if (pipe(fds))
	{
		print_error("pipe()");
		return 1;
	}

	if (argc > 1)
	{
		if ( (fd = open(argv[1], O_RDONLY)) < 0)
		{
			print_error(argv[1]);
			return 1;
		}
	}

	switch (pid = fork())
	{
	case -1:
		print_error("fork()");
		return 1;

	case 0:
		close(fds[0]);
		if (fd >= 0) dup2(fd, 0);
		dup2(fds[1], 1);

		execl(CPPPATH, "cpp", "-nostdinc", "-undef", 0);

		print_error(CPPPATH);
		return 1;

	default:
		if (fd >= 0) close(fd);
		close(fds[1]);

		if ( (yyin = fdopen(fds[0], "r")) )
		{
			if (yyparse())
				error = 1;
			fclose(yyin);
		}
		else
		{
			print_error("fdopen()");
		}

		waitpid(pid, 0, 0);
		break;
	}

	if (error)
		return 1;

	compile();

	if (error)
		return 1;

	calc_indexes();

	if (error)
		return 1;

	if (argc == 3)
	{
		FILE *fout;

		if ( (fout = fopen(argv[2], "w")) )
		{
			output(fout);
			fclose(fout);
		}
		else
		{
			print_error(argv[2]);
		}
	}
	else
	{
		output(stdout);
	}

	return error;
}

void yyerror(const char *msg)
{
	if (this_fn)
		fprintf(stderr, "%s: %s(): %s\n", progname,
			this_fn->var.name->cstr, msg);
	else
		fprintf(stderr, "%s: %s\n", progname, msg);
	error = 1;
}

void yyerrorf(const char *fmt, ...)
{
	char msg[1024];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(msg, 1024, fmt, ap);
	va_end(ap);
	
	yyerror(msg);
}

void print_error(char *str)
{
	char msg[1024];

	sprintf(msg, "%s: %s", progname, str);
	perror(msg);

	error = 1;
}
