/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "compile.h"

variable *globals = 0;
function *functions = 0;
function *this_fn = 0;
string *string_hash[256];

static int check_params(variable *v1, variable *v2)
{
	while (v1)
	{
		if (!v2 || v1->typ != v2->typ)
			return 0;

		v1 = v1->next;
		v2 = v2->next;
	}

	return (v2 == 0);
}

function *new_function(type *typ, string *name, variable *params)
{
	function *fn, **ptr = &functions;

	while ((fn = *ptr))
	{
		if (fn->var.name == name)
		{
			if (fn->var.typ == typ &&
			    check_params(fn->params, params))
			{
				if (!fn->var.defd)
					fn->params = params;
				return fn;
			}

			if (name)
			{
				yyerrorf("conflicting types for `%s'",
					 fn->var.name->cstr);
				return fn;
			}
		}
		ptr = &fn->next;
	}

	if (name && find_variable(globals, name))
	{
		yyerrorf("conflicting types for `%s'", name->cstr);
	}

	*ptr = fn = malloc(sizeof(function));
	fn->next = 0;
	fn->var.next = 0;
	fn->var.name = name;
	fn->var.typ = typ;
	fn->var.cnst = 0;
	fn->var.used = 0;
	fn->var.defd = 0;
	fn->var.idx = 0;
	fn->params = params;
	fn->constants = 0;
	fn->locals = 0;
	fn->tmps = 0;
	fn->body = 0;
	fn->instrs = 0;
	fn->ninstrs = 0;
	fn->nparams = 0;
	fn->nvars = 0;

	return fn;
}

function *find_function(string *name)
{
	function *fn = functions;

	while (fn)
	{
		if (fn->var.name == name)
			return fn;
		fn = fn->next;
	}

	return 0;
}

int fn_check_type(function *fn1, function *fn2)
{
	if (fn1->var.typ != fn2->var.typ)
		return 1;

	return !check_params(fn1->params, fn2->params);
}


statement *new_statement(int func)
{
	statement *stmt = malloc(sizeof(statement));

	stmt->next = 0;
	stmt->func = func;
	stmt->e1 = 0;
	stmt->e2 = 0;
	stmt->e3 = 0;
	stmt->s1 = 0;
	stmt->s2 = 0;

	return stmt;
}


variable *new_variable(type *typ, string *name)
{
	variable *v = malloc(sizeof(variable));

	if (typ->typ == T_VOID)
	{
		yyerrorf("variable `%s' declared void", name?name->cstr:"");
	}

	v->next = 0;
	v->name = name;
	v->typ = typ;
	v->cnst = 0;
	v->used = 0;
	v->defd = 0;
	v->idx = 0;
	v->u.fn = typ->fn;

	return v;
}

variable *new_this(void)
{
	static variable *this_var = 0;

	if (!this_var)
		this_var = new_variable(new_type(T_ENTITY, 0), 0);

	return this_var;
}

variable *new_parameters(variable *vars)
{
	variable *var;

	for (var = vars; var; var = var->next)
	{
		var->used = 1;

		if (var->name && find_variable(var->next, var->name))
		{
			yyerrorf("redeclaration of `%s'", var->name->cstr);
		}
	}

	return vars;
}

void add_global_vars(type *typ, variable *vars)
{
	variable *var, *next = vars;

	while ( (var = next) )
	{
		next = var->next;

		if (typ->typ == T_VOID)
		{
			yyerrorf("variable `%s' declared void", 
				 var->name->cstr);
		}

		var->next = 0;
		var->typ = typ;
		var->u.fn = typ->fn;

		add_global_var(var);
	}
}

void add_global_var(variable *new)
{
	variable *var, **ptr = &globals;

	while ((var = *ptr))
	{
		if (var->name == new->name)
		{
			if (var->typ != new->typ)
			{
				yyerrorf("conflicting types for `%s'",
					 new->name->cstr);
			}

			free(new);

			return;
		}
		ptr = &var->next;
	}

	if (find_function(new->name))
	{
		yyerrorf("conflicting types for `%s'", new->name->cstr);
	}

	*ptr = new;	
}

void add_local_vars(type *typ, variable *vars)
{
	variable *var, *next = vars;

	while ( (var = next) )
	{
		next = var->next;

		if (typ->typ == T_VOID)
		{
			yyerrorf("variable `%s' declared void", 
				 var->name->cstr);
		}

		var->next = 0;
		var->typ = typ;
		var->u.fn = typ->fn;

		add_local_var(var);
	}
}

void add_local_var(variable *new)
{
	variable *var, **ptr = &this_fn->locals;

	while ((var = *ptr))
	{
		if (var->name == new->name)
		{
			yyerrorf("redeclaration of `%s'", new->name->cstr);
			return;
		}
		ptr = &var->next;
	}

	if (find_variable(this_fn->params, new->name))
	{
		yyerrorf("redeclaration of `%s'", new->name->cstr);
	}

	*ptr = new;
}

variable *new_float_const(float f)
{
	variable *var, **ptr = &this_fn->constants;

	while ((var = *ptr))
	{
		if (var->typ->typ == T_FLOAT && var->u.flt == f)
			return var;
		ptr = &var->next;
	}

	*ptr = var = new_variable(new_type(T_FLOAT, 0), 0);
	var->cnst = 1;
	var->u.flt = f;

	return var;
}

variable *new_vector_const(float x, float y, float z)
{
	variable *var, **ptr = &this_fn->constants;

	while ((var = *ptr))
	{
		if (var->typ->typ == T_VECTOR && var->u.vec.x == x &&
		    var->u.vec.y == y && var->u.vec.z == z)
			return var;
		ptr = &var->next;
	}

	*ptr = var = new_variable(new_type(T_VECTOR, 0), 0);
	var->cnst = 1;
	var->u.vec.x = x;
	var->u.vec.y = y;
	var->u.vec.z = z;

	return var;
}

variable *new_string_const(string *str)
{
	variable *var, **ptr = &this_fn->constants;

	while ((var = *ptr))
	{
		if (var->typ->typ == T_STRING && var->u.str == str)
			return var;
		ptr = &var->next;
	}

	*ptr = var = new_variable(new_type(T_STRING, 0), 0);
	var->cnst = 1;
	var->u.str = str;

	return var;
}

variable *new_function_const(function *fn)
{
	variable *var, **ptr = &this_fn->constants;

	while ((var = *ptr))
	{
		if (var->typ->typ == T_FUNCTION && var->u.fn == fn)
			return var;
		ptr = &var->next;
	}

	*ptr = var = new_variable(new_type(T_FUNCTION, fn), 0);
	var->cnst = 1;
	var->u.fn = fn;

	return var;
}

variable *find_variable(variable *var, string *name)
{
	while (var)
	{
		if (var->name == name)
			return var;
		var = var->next;
	}

	return 0;
}


expression *new_expression(type *typ, char func, char op)
{
	expression *exp = malloc(sizeof(expression));

	exp->prev = 0;
	exp->typ = typ;
	exp->func = func;
	exp->op = op;
	exp->var = 0;
	exp->fn = 0;
	exp->left = 0;
	exp->right = 0;

	return exp;
}

static void expr_set_var(expression *exp, variable *var)
{
	exp->var = var;
	exp->typ = var->typ;
	exp->fn = var->u.fn;
}

expression *new_nop_expr(expression *left)
{
	expression *exp = new_expression(left->typ, E_NOP, 0);

	exp->left = left;
	exp->fn = left->fn;

	return exp;
}

expression *new_const_expr(struct variable *var)
{
	expression *exp = new_expression(var->typ, E_CONSTANT, 0);

	expr_set_var(exp, var);

	return exp;
}

expression *new_variable_expr(expression *left, string *name)
{
	expression *exp = new_expression(new_type(T_VOID, 0), 0, 0);
	variable *var;
	function *fn;

	if (!left && (var = find_variable(this_fn->locals, name)))
	{
		exp->func = E_LOCAL;
		expr_set_var(exp, var);
		return exp;
	}

	if (!left && (var = find_variable(this_fn->params, name)))
	{
		exp->func = E_LOCAL;
		expr_set_var(exp, var);
		return exp;
	}

	if ((var = find_variable(globals, name)))
	{
		if (!left) left = new_const_expr(new_this());
		exp->func = E_GLOBAL;
		expr_set_var(exp, var);
		exp->left = left;
		return exp;
	}

	if ((fn = find_function(name)))
	{
		exp->func = E_FUNCTION;
		expr_set_var(exp, new_function_const(fn));
		exp->left = left;
		return exp;
	}

	yyerrorf("undefined symbol `%s'", name->cstr);

	return exp;
}


static struct { char func, op, right, typ; } una_ops[] =
{
        { U_NEG, OP_NEG_F,    T_FLOAT,    T_FLOAT  },
        { U_NEG, OP_NEG_V,    T_VECTOR,   T_VECTOR },
	{ U_INV, OP_BITNOT_F, T_FLOAT,    T_FLOAT  },
	{ U_NOT, OP_NOT_F,    T_FLOAT,    T_FLOAT  },
	{ U_NOT, OP_NOT_V,    T_VECTOR,   T_FLOAT  },
	{ U_NOT, OP_NOT_S,    T_STRING,   T_FLOAT  },
	{ U_NOT, OP_NOT_N,    T_FUNCTION, T_FLOAT  },
	{ U_NOT, OP_NOT_E,    T_ENTITY,   T_FLOAT  },
	{ 0,     0,           0,          0        }
};

expression *new_unary_expr(char func, expression *right)
{
	expression *exp;
	char typ = T_FLOAT, op = 0;
	unsigned n;

	for (n = 0; una_ops[n].func; n++)
	{
		if (una_ops[n].func == func &&
		    una_ops[n].right == right->typ->typ)
		{
			typ = una_ops[n].typ;
			op = una_ops[n].op;
			break;
		}
	}

	if (!op)
	{
		yyerror("type mismatch");
	}

	exp = new_expression(new_type(typ, 0), E_UNARY, op);

	exp->right = right;

	return exp;
}

static struct { char func, op, left, right, typ; } bin_ops[] =
{
	{ B_ADD,  OP_ADD_F,    T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_ADD,  OP_ADD_V,    T_VECTOR,   T_VECTOR,   T_VECTOR },
	{ B_SUB,  OP_SUB_F,    T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_SUB,  OP_SUB_V,    T_VECTOR,   T_VECTOR,   T_VECTOR },
	{ B_MUL,  OP_MUL_F,    T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_MUL,  OP_MUL_V,    T_VECTOR,   T_VECTOR,   T_FLOAT  },
	{ B_MUL,  OP_MUL_FV,   T_FLOAT,    T_VECTOR,   T_VECTOR },
	{ B_MUL,  OP_MUL_VF,   T_VECTOR,   T_FLOAT,    T_VECTOR },
	{ B_DIV,  OP_DIV_F,    T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_DIV,  OP_DIV_VF,   T_VECTOR,   T_FLOAT,    T_VECTOR },
	{ B_EQ,   OP_EQ_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_EQ,   OP_EQ_V,     T_VECTOR,   T_VECTOR,   T_FLOAT  },
	{ B_EQ,   OP_EQ_S,     T_STRING,   T_STRING,   T_FLOAT  },
	{ B_EQ,   OP_EQ_N,     T_FUNCTION, T_FUNCTION, T_FLOAT  },
	{ B_EQ,   OP_EQ_E,     T_ENTITY,   T_ENTITY,   T_FLOAT  },
	{ B_NE,   OP_NE_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_NE,   OP_NE_V,     T_VECTOR,   T_VECTOR,   T_FLOAT  },
	{ B_NE,   OP_NE_S,     T_STRING,   T_STRING,   T_FLOAT  },
	{ B_NE,   OP_NE_N,     T_FUNCTION, T_FUNCTION, T_FLOAT  },
	{ B_NE,   OP_NE_E,     T_ENTITY,   T_ENTITY,   T_FLOAT  },
	{ B_LT,   OP_LT_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_GT,   OP_GT_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_LE,   OP_LE_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_GE,   OP_GE_F,     T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_AND,  OP_BITAND_F, T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_OR,   OP_BITOR_F,  T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_EOR,  OP_BITEOR_F, T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_FF,   T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_FV,   T_FLOAT,    T_VECTOR,   T_FLOAT  },
	{ B_LAND, OP_AND_FS,   T_FLOAT,    T_STRING,   T_FLOAT  },
	{ B_LAND, OP_AND_FN,   T_FLOAT,    T_FUNCTION, T_FLOAT  },
	{ B_LAND, OP_AND_FE,   T_FLOAT,    T_ENTITY,   T_FLOAT  },
	{ B_LAND, OP_AND_VF,   T_VECTOR,   T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_VV,   T_VECTOR,   T_VECTOR,   T_FLOAT  },
	{ B_LAND, OP_AND_VS,   T_VECTOR,   T_STRING,   T_FLOAT  },
	{ B_LAND, OP_AND_VN,   T_VECTOR,   T_FUNCTION, T_FLOAT  },
	{ B_LAND, OP_AND_VE,   T_VECTOR,   T_ENTITY,   T_FLOAT  },
	{ B_LAND, OP_AND_SF,   T_STRING,   T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_SV,   T_STRING,   T_VECTOR,   T_FLOAT  },
	{ B_LAND, OP_AND_SS,   T_STRING,   T_STRING,   T_FLOAT  },
	{ B_LAND, OP_AND_SN,   T_STRING,   T_FUNCTION, T_FLOAT  },
	{ B_LAND, OP_AND_SE,   T_STRING,   T_ENTITY,   T_FLOAT  },
	{ B_LAND, OP_AND_NF,   T_FUNCTION, T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_NV,   T_FUNCTION, T_VECTOR,   T_FLOAT  },
	{ B_LAND, OP_AND_NS,   T_FUNCTION, T_STRING,   T_FLOAT  },
	{ B_LAND, OP_AND_NN,   T_FUNCTION, T_FUNCTION, T_FLOAT  },
	{ B_LAND, OP_AND_NE,   T_FUNCTION, T_ENTITY,   T_FLOAT  },
	{ B_LAND, OP_AND_EF,   T_ENTITY,   T_FLOAT,    T_FLOAT  },
	{ B_LAND, OP_AND_EV,   T_ENTITY,   T_VECTOR,   T_FLOAT  },
	{ B_LAND, OP_AND_ES,   T_ENTITY,   T_STRING,   T_FLOAT  },
	{ B_LAND, OP_AND_EN,   T_ENTITY,   T_FUNCTION, T_FLOAT  },
	{ B_LAND, OP_AND_EE,   T_ENTITY,   T_ENTITY,   T_FLOAT  },
	{ B_LOR,  OP_OR_FF,    T_FLOAT,    T_FLOAT,    T_FLOAT  },
	{ B_LOR,  OP_OR_FV,    T_FLOAT,    T_VECTOR,   T_FLOAT  },
	{ B_LOR,  OP_OR_FS,    T_FLOAT,    T_STRING,   T_FLOAT  },
	{ B_LOR,  OP_OR_FN,    T_FLOAT,    T_FUNCTION, T_FLOAT  },
	{ B_LOR,  OP_OR_FE,    T_FLOAT,    T_ENTITY,   T_FLOAT  },
	{ B_LOR,  OP_OR_VF,    T_VECTOR,   T_FLOAT,    T_FLOAT  },
	{ B_LOR,  OP_OR_VV,    T_VECTOR,   T_VECTOR,   T_FLOAT  },
	{ B_LOR,  OP_OR_VS,    T_VECTOR,   T_STRING,   T_FLOAT  },
	{ B_LOR,  OP_OR_VN,    T_VECTOR,   T_FUNCTION, T_FLOAT  },
	{ B_LOR,  OP_OR_VE,    T_VECTOR,   T_ENTITY,   T_FLOAT  },
	{ B_LOR,  OP_OR_SF,    T_STRING,   T_FLOAT,    T_FLOAT  },
	{ B_LOR,  OP_OR_SV,    T_STRING,   T_VECTOR,   T_FLOAT  },
	{ B_LOR,  OP_OR_SS,    T_STRING,   T_STRING,   T_FLOAT  },
	{ B_LOR,  OP_OR_SN,    T_STRING,   T_FUNCTION, T_FLOAT  },
	{ B_LOR,  OP_OR_SE,    T_STRING,   T_ENTITY,   T_FLOAT  },
	{ B_LOR,  OP_OR_NF,    T_FUNCTION, T_FLOAT,    T_FLOAT  },
	{ B_LOR,  OP_OR_NV,    T_FUNCTION, T_VECTOR,   T_FLOAT  },
	{ B_LOR,  OP_OR_NS,    T_FUNCTION, T_STRING,   T_FLOAT  },
	{ B_LOR,  OP_OR_NN,    T_FUNCTION, T_FUNCTION, T_FLOAT  },
	{ B_LOR,  OP_OR_NE,    T_FUNCTION, T_ENTITY,   T_FLOAT  },
	{ B_LOR,  OP_OR_EF,    T_ENTITY,   T_FLOAT,    T_FLOAT  },
	{ B_LOR,  OP_OR_EV,    T_ENTITY,   T_VECTOR,   T_FLOAT  },
	{ B_LOR,  OP_OR_ES,    T_ENTITY,   T_STRING,   T_FLOAT  },
	{ B_LOR,  OP_OR_EN,    T_ENTITY,   T_FUNCTION, T_FLOAT  },
	{ B_LOR,  OP_OR_EE,    T_ENTITY,   T_ENTITY,   T_FLOAT  },
	{ 0,      0,           0,          0,          0        }
};

expression *new_binary_expr(char func, expression *left, expression *right)
{
	expression *exp;
	char typ = T_FLOAT, op = 0;
	unsigned n;

	for (n = 0; bin_ops[n].func; n++)
	{
		if (bin_ops[n].func == func &&
		    bin_ops[n].left == left->typ->typ &&
		    bin_ops[n].right == right->typ->typ)
		{
			typ = bin_ops[n].typ;
			op = bin_ops[n].op;
			break;
		}
	}

	if (!op)
	{
		yyerror("type mismatch");
	}

	exp = new_expression(new_type(typ, 0), E_BINARY, op);

	exp->left = left;
	exp->right = right;

	return exp;
}

expression *new_assign_expr(expression *left, expression *right)
{
	expression *exp;

	if (left->typ != right->typ)
	{
		yyerror("type mismatch");
	}

	exp = new_expression(left->typ, E_ASSIGN, 0);

	exp->left = left;
	exp->right = right;

	return exp;
}

expression *new_vecdot_expr(expression *left, string *name)
{
	expression *exp = new_expression(new_type(T_FLOAT, 0), 0, 0);

	exp->left = left;

	if (!strcmp(name->cstr, "x"))
	{
		exp->func = E_DOTX;
	}
	else if (!strcmp(name->cstr, "y"))
	{
		exp->func = E_DOTY;
	}
	else if (!strcmp(name->cstr, "z"))
	{
		exp->func = E_DOTZ;
	}
	else
	{
		yyerrorf("request for member `%s' in non entity", name->cstr);
	}

	return exp;
}

static int check_args(variable **ptr, expression *exp, int n)
{
	variable *var = *ptr;

	if (!exp)
		return (var != 0);

	if (exp->prev)
	{
		if (check_args(ptr, exp->prev, 0))
			return 1;
		var = *ptr;
	}

	if (!var || var->typ != exp->typ)
		return 1;

	if ((*ptr = var->next))
		return n;

	return 0;
}

expression *new_call_expr(expression *func, expression *args)
{
	function *fn;
	expression *exp;
	variable *var;
	type *typ;

	if (func->typ->typ != T_FUNCTION)
	{
		yyerror("not a function");
		return func;
	}

	fn = func->fn;
	typ = fn->var.typ;
	exp = new_expression(typ, E_CALL, OP_CALL);

	var = fn->params;
	if (check_args(&var, args, 1))
		yyerror("type mismatch");

	exp->left = func;
	exp->right = args;
	exp->fn = typ->fn;

	return exp;
}


type *new_type(char typ, function *fn)
{
	static type *types = 0;
	type *tp, **ptr;

	for (ptr = &types; (tp = *ptr); ptr = &tp->next)
	{
		if (tp->typ == typ)
		{
			if (typ != T_FUNCTION ||
			    !fn_check_type(tp->fn, fn))
				return tp;
		}
	}

	*ptr = tp = malloc(sizeof(type));
	tp->typ = typ;
	tp->fn = fn;

	return tp;
}


string *new_string(const char *cstr)
{
	string *str, **ptr = &string_hash[(unsigned char)cstr[0]];

	while ((str = *ptr))
	{
		if (!strncmp(str->cstr, cstr, 255))
			return str;
		ptr = &str->next;
	}

	*ptr = str = malloc(sizeof(string));
	str->next = 0;
	str->idx = 0;
	str->cstr = strdup(cstr);

	return str;
}

