/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>

#include "compile.h"

typedef struct oprand
{
	variable *var;
	char offset;
} oprand;


static void fn_compile(function *);
static void stmt_compile(function *, statement *);
static int expr_compile(expression *, oprand *, oprand *, oprand *);
static int expr_compile_one(expression *, oprand *, oprand *, oprand *);
static void expr_compile_args(variable **, expression *, oprand *);

static variable *new_tmp_var(char);
static float get_float(oprand *);
static void push_op(char, oprand *, oprand *, oprand *);
static instruction *push_goto_op(char, oprand *, int, oprand *);

static variable **tmp_ptr;
static instruction *instrs;
static unsigned index;
static int overflow;


static char if_op[] =
{
	0,
	OP_IF_F,
	OP_IF_V,
	OP_IF_S,
	OP_IF_N,
	OP_IF_E
};

static char ifnot_op[] =
{
	0,
	OP_IFNOT_F,
	OP_IFNOT_V,
	OP_IFNOT_S,
	OP_IFNOT_N,
	OP_IFNOT_E
};

static char move_op[] =
{
	0,
	OP_MOVE_F,
	OP_MOVE_V,
	OP_MOVE_S,
	OP_MOVE_N,
	OP_MOVE_E
};

static char load_e_op[] =
{
	0,
	OP_LOAD_EF,
	OP_LOAD_EV,
	OP_LOAD_ES,
	OP_LOAD_EN,
	OP_LOAD_EE
};

static char store_e_op[] =
{
	0,
	OP_STORE_EF,
	OP_STORE_EV,
	OP_STORE_ES,
	OP_STORE_EN,
	OP_STORE_EE
};

static char load_fn_op[] =
{
	0,
	OP_LOAD_NF,
	OP_LOAD_NV,
	OP_LOAD_NS,
	OP_LOAD_NN,
	OP_LOAD_NE
};

static char store_fn_op[] =
{
	0,
	OP_STORE_NF,
	OP_STORE_NV,
	OP_STORE_NS,
	OP_STORE_NN,
	OP_STORE_NE
};


void compile(void)
{
	function *fn = functions;

	while (fn)
	{  
		if (fn->var.defd)
			fn_compile(fn);
		fn = fn->next;
	}
}

static void fn_compile(function *fn)
{
	this_fn = fn;

	tmp_ptr = &fn->tmps;
	instrs = malloc(sizeof(instruction) * 65536);
	index = 0;
	overflow = 0;

	stmt_compile(fn, fn->body);

	push_op(OP_DONE, 0, 0, 0);

	fn->instrs = realloc(instrs, sizeof(instruction) * index);
	fn->ninstrs = index;

	this_fn = 0;
}

static void stmt_compile(function *fn, statement *stmt)
{
	oprand opr;
	instruction *ptr;
	char op;
	unsigned loop;
	variable **tmp = tmp_ptr;

	if (!stmt)
		return;

	switch(stmt->func)
	{
	case S_BLOCK:
		stmt_compile(fn, stmt->s1);
		break;

	case S_EXPR:
		expr_compile(stmt->e1, 0, 0, 0);
		break;

	case S_IF:
		opr.var = 0;
		opr.offset = 0;

		if ((op = ifnot_op[expr_compile(stmt->e1, &opr, 0, 0)]))
		{
			ptr = push_goto_op(op, &opr, 0, 0);
			tmp_ptr = tmp;
			stmt_compile(fn, stmt->s1);
			if (stmt->s2)
			{
				ptr->b.idx = index + 1;
				ptr = push_goto_op(OP_GOTO, 0, 0, 0);
				stmt_compile(fn, stmt->s2);
			}
			ptr->b.idx = index;
		}

		break;

	case S_UNLESS:
		opr.var = 0;
		opr.offset = 0;

		if ((op = if_op[expr_compile(stmt->e1, &opr, 0, 0)]))
		{
			ptr = push_goto_op(op, &opr, 0, 0);
			tmp_ptr = tmp;
			stmt_compile(fn, stmt->s1);
			if (stmt->s2)
			{
				ptr->b.idx = index + 1;
				ptr = push_goto_op(OP_GOTO, 0, 0, 0);
				stmt_compile(fn, stmt->s2);
			}
			ptr->b.idx = index;
		}

		break;

	case S_WHILE:
		loop = index;
		opr.var = 0;
		opr.offset = 0;

		if ((op = ifnot_op[expr_compile(stmt->e1, &opr, 0, 0)]))
		{
			ptr = push_goto_op(op, &opr, 0, 0);
			tmp_ptr = tmp;
			stmt_compile(fn, stmt->s1);
			push_goto_op(OP_GOTO, 0, loop, 0);
			ptr->b.idx = index;
		}

		break;

	case S_UNTIL:
		loop = index;
		opr.var = 0;
		opr.offset = 0;

		if ((op = if_op[expr_compile(stmt->e1, &opr, 0, 0)]))
		{
			ptr = push_goto_op(op, &opr, 0, 0);
			tmp_ptr = tmp;
			stmt_compile(fn, stmt->s1);
			push_goto_op(OP_GOTO, 0, loop, 0);
			ptr->b.idx = index;
		}

		break;

	case S_DO_WHILE:
		loop = index;
		opr.var = 0;
		opr.offset = 0;

		stmt_compile(fn, stmt->s1);
		if ((op = if_op[expr_compile(stmt->e1, &opr, 0, 0)]))
			push_goto_op(op, &opr, loop, 0);
		tmp_ptr = tmp;
		break;

	case S_DO_UNTIL:
		loop = index;
		opr.var = 0;
		opr.offset = 0;

		stmt_compile(fn, stmt->s1);
		if ((op = ifnot_op[expr_compile(stmt->e1, &opr, 0, 0)]))
			push_goto_op(op, &opr, loop, 0);
		tmp_ptr = tmp;
		break;

	case S_FOR:
		expr_compile(stmt->e1, 0, 0, 0);

		loop = index;
		opr.var = 0;
		opr.offset = 0;

		if ((op = ifnot_op[expr_compile(stmt->e2, &opr, 0, 0)]))
		{
			ptr = push_goto_op(op, &opr, 0, 0);
			tmp_ptr = tmp;
			stmt_compile(fn, stmt->s1);
			expr_compile(stmt->e3, 0, 0, 0);
			push_goto_op(OP_GOTO, 0, loop, 0);
			ptr->b.idx = index;
		}

		break;

	case S_RETURN:
		opr.var = &fn->var;
		opr.offset = 0;
		expr_compile(stmt->e1, &opr, 0, 0);
		push_op(OP_RETURN, 0, 0, 0);
		break;
	}

	tmp_ptr = tmp;

	stmt_compile(fn, stmt->next);
}

static int expr_compile(expression *exp, oprand *ptr,
			oprand *ent, oprand *fent)
{
	if (!exp)
		return T_VOID;

	expr_compile(exp->prev, 0, 0, 0);

	return expr_compile_one(exp, ptr, ent, fent);
}

static void expr_compile_args(variable **ptr, expression *exp, oprand *fn)
{
	oprand op1, op2;

	if (!exp)
		return;

	expr_compile_args(ptr, exp->prev, fn);

	op1.var = 0;
	op1.offset = 0;
	expr_compile_one(exp, &op1, 0, 0);

	op2.var = *ptr;
	op2.offset = 0;
	push_op(store_fn_op[exp->typ->typ], fn, &op2, &op1);

	*ptr = op2.var->next;
}

static int expr_compile_one(expression *exp, oprand *ptr,
			    oprand *ent, oprand *fent)
{
	oprand op1, op2, op3;
	variable **tmp = tmp_ptr;

	switch(exp->func)
	{
	case E_NOP:
		expr_compile(exp->left, ptr, ent, fent);
		break;

	case E_FUNCTION:
		expr_compile(exp->left, fent, 0, 0);
		/* fall through */

	case E_CONSTANT:
	case E_LOCAL:
		if (ptr)
		{
			if (ptr->var)
			{
				op1.var = exp->var;
				op1.offset = 0;
				push_op(move_op[exp->typ->typ], &op1, 0, ptr);
			}
			else
			{
				ptr->var = exp->var;
				ptr->offset = 0;
			}
		}
		break;

	case E_GLOBAL:
		if (ptr)
		{
			if (ent)
			{
				expr_compile(exp->left, ent, 0, 0);
				ptr->var = exp->var;
				ptr->offset = 0;
			}
			else if (fent)
			{
				expr_compile(exp->left, fent, 0, 0);
				tmp_ptr = tmp;

				if (!ptr->var)
				{
					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;
				}

				op1.var = exp->var;
				op1.offset = 0;
				push_op(load_e_op[exp->typ->typ], fent, &op1, ptr);
			}
			else
			{
				op1.var = 0;
				op1.offset = 0;
				expr_compile(exp->left, &op1, 0, 0);
				tmp_ptr = tmp;

				if (!ptr->var)
				{
					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;
				}

				op2.var = exp->var;
				op2.offset = 0;
				push_op(load_e_op[exp->typ->typ], &op1, &op2, ptr);
			}
		}
		else
		{
			expr_compile(exp->left, 0, 0, 0);
		}
		break;

	case E_DOTX:
		if (ptr)
		{
			if (ptr->var)
			{
				op1.var = 0;
				expr_compile(exp->left, &op1, 0, 0);
				op1.offset = 0;
				push_op(move_op[exp->typ->typ], &op1, 0, ptr);
			}
			else
			{
				op1.var = 0;
				op2.var = 0;
				op2.offset = 0;
				expr_compile(exp->left, &op1, &op2, 0);
				op1.offset = 0;

				if (op2.var)
				{
					tmp_ptr = tmp;

					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;

					push_op(load_e_op[exp->typ->typ],
						&op2, &op1, ptr);
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
		}
		else
		{
			expr_compile(exp->left, 0, 0, 0);
		}
		break;

	case E_DOTY:
		if (ptr)
		{
			if (ptr->var)
			{
				op1.var = 0;
				expr_compile(exp->left, &op1, 0, 0);
				op1.offset = 1;
				push_op(move_op[exp->typ->typ], &op1, 0, ptr);
			}
			else
			{
				op1.var = 0;
				op2.var = 0;
				op2.offset = 0;
				expr_compile(exp->left, &op1, &op2, 0);
				op1.offset = 1;

				if (op2.var)
				{
					tmp_ptr = tmp;

					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;

					push_op(load_e_op[exp->typ->typ],
						&op2, &op1, ptr);
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
		}
		else
		{
			expr_compile(exp->left, 0, 0, 0);
		}
		break;

	case E_DOTZ:
		if (ptr)
		{
			if (ptr->var)
			{
				op1.var = 0;
				expr_compile(exp->left, &op1, 0, 0);
				op1.offset = 2;
				push_op(move_op[exp->typ->typ], &op1, 0, ptr);
			}
			else
			{
				op1.var = 0;
				op2.var = 0;
				op2.offset = 0;
				expr_compile(exp->left, &op1, &op2, 0);
				op1.offset = 2;

				if (op2.var)
				{
					tmp_ptr = tmp;

					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;

					push_op(load_e_op[exp->typ->typ],
						&op2, &op1, ptr);
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
		}
		else
		{
			expr_compile(exp->left, 0, 0, 0);
		}
		break;

	case E_UNARY:
		if (ptr)
		{
			op1.var = 0;
			op1.offset = 0;
			expr_compile(exp->right, &op1, 0, 0);
			tmp_ptr = tmp;

			if (op1.var->cnst)
			{
				switch(exp->op)
				{
				case OP_NEG_F:
					op1.var = new_float_const
						(-get_float(&op1));
					break;
				case OP_NEG_V:
					op1.var = new_vector_const
						(-op1.var->u.vec.x,
						 -op1.var->u.vec.y,
						 -op1.var->u.vec.z);
					break;
				case OP_BITNOT_F:
					op1.var = new_float_const
						(~((int)get_float(&op1)));
					break;
				case OP_NOT_F:
					op1.var = new_float_const
						(!get_float(&op1));
					break;
				case OP_NOT_V:
					op1.var = new_float_const
						(!(op1.var->u.vec.x ||
						   op1.var->u.vec.y ||
						   op1.var->u.vec.z));
					break;
				case OP_NOT_S:
					op1.var = new_float_const
						(!op1.var->u.str->cstr[0]);
					break;
				case OP_NOT_E:
					op1.var = new_float_const(!0);
					break;
				}
				op1.offset = 0;

				if (ptr->var)
				{
					push_op(move_op[exp->typ->typ],
						&op1, 0, ptr);
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
			else
			{
				if (!ptr->var)
				{
					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;
				}

				push_op(exp->op, &op1, 0, ptr);
			}
		}
		else
		{
			expr_compile(exp->right, 0, 0, 0);
		}
		break;

	case E_BINARY:
		if (ptr)
		{
			op1.var = 0;
			op1.offset = 0;
			expr_compile(exp->left, &op1, 0, 0);
			op2.var = 0;
			op2.offset = 0;
			expr_compile(exp->right, &op2, 0, 0);
			tmp_ptr = tmp;

			if (op1.var->cnst && op2.var->cnst)
			{
				float f;

				switch(exp->op)
				{
				case OP_ADD_F:
					op1.var = new_float_const
						(get_float(&op1) +
						 get_float(&op2));
					break;
				case OP_ADD_V:
					op1.var = new_vector_const
						(op1.var->u.vec.x +
						 op2.var->u.vec.x,
						 op1.var->u.vec.y +
						 op2.var->u.vec.y,
						 op1.var->u.vec.z +
						 op2.var->u.vec.z);
					break;
				case OP_SUB_F:
					op1.var = new_float_const
						(get_float(&op1) -
						 get_float(&op2));
					break;
				case OP_SUB_V:
					op1.var = new_vector_const
						(op1.var->u.vec.x -
						 op2.var->u.vec.x,
						 op1.var->u.vec.y -
						 op2.var->u.vec.y,
						 op1.var->u.vec.z -
						 op2.var->u.vec.z);
					break;
				case OP_MUL_F:
					op1.var = new_float_const
						(get_float(&op1) *
						 get_float(&op2));
					break;
				case OP_MUL_V:
					op1.var = new_float_const
						((op1.var->u.vec.x *
						  op2.var->u.vec.x) +
						 (op1.var->u.vec.y *
						  op2.var->u.vec.y) +
						 (op1.var->u.vec.z *
						  op2.var->u.vec.z));
					break;
				case OP_MUL_FV:
					f = get_float(&op1);
					op1.var = new_vector_const
						(f * op2.var->u.vec.x,
						 f * op2.var->u.vec.y,
						 f * op2.var->u.vec.z);
					break;
				case OP_MUL_VF:
					f = get_float(&op2);
					op1.var = new_vector_const
						(op1.var->u.vec.x * f,
						 op1.var->u.vec.y * f,
						 op1.var->u.vec.z * f);
					break;
				case OP_DIV_F:	
					op1.var = new_float_const
						(get_float(&op1) /
						 get_float(&op2));
					break;
				case OP_DIV_VF:
					f = get_float(&op2);
					op1.var = new_vector_const
						(op1.var->u.vec.x / f,
						 op1.var->u.vec.y / f,
						 op1.var->u.vec.z / f);
					break;	
				case OP_EQ_F:
					op1.var = new_float_const
						(get_float(&op1) ==
						 get_float(&op2));
					break;
				case OP_EQ_V:
					op1.var = new_float_const
						((op1.var->u.vec.x ==
						  op2.var->u.vec.x) &&
						 (op1.var->u.vec.y ==
						  op2.var->u.vec.y) &&
						 (op1.var->u.vec.z ==
						  op2.var->u.vec.z));
					break;
				case OP_EQ_S:
					op1.var = new_float_const
						(op1.var->u.str ==
						 op2.var->u.str);
					break;
				case OP_EQ_E:
					op1.var = new_float_const(0 == 0);
					break;
				case OP_NE_F:
					op1.var = new_float_const
						(get_float(&op1) !=
						 get_float(&op2));
					break;
				case OP_NE_V:
					op1.var = new_float_const
						((op1.var->u.vec.x !=
						  op2.var->u.vec.x) ||
						 (op1.var->u.vec.y !=
						  op2.var->u.vec.y) ||
						 (op1.var->u.vec.z !=
						  op2.var->u.vec.z));
					break;
				case OP_NE_S:
					op1.var = new_float_const
						(op1.var->u.str !=
						 op2.var->u.str);
					break;
				case OP_NE_E:
					op1.var = new_float_const(0 != 0);
					break;
				case OP_LT_F:
					op1.var = new_float_const
						(get_float(&op1) <
						 get_float(&op2));
					break;
				case OP_GT_F:
					op1.var = new_float_const
						(get_float(&op1) >
						 get_float(&op2));
					break;
				case OP_LE_F:
					op1.var = new_float_const
						(get_float(&op1) <=
						 get_float(&op2));
					break;
				case OP_GE_F:
					op1.var = new_float_const
						(get_float(&op1) >=
						 get_float(&op2));
					break;
				case OP_BITAND_F:
					op1.var = new_float_const
						(((int)get_float(&op1)) &
						 ((int)get_float(&op2)));
					break;
				case OP_BITOR_F:
					op1.var = new_float_const
						(((int)get_float(&op1)) |
						 ((int)get_float(&op2)));
					break;
				case OP_BITEOR_F:
					op1.var = new_float_const
						(((int)get_float(&op1)) ^
						 ((int)get_float(&op2)));
					break;
				case OP_AND_FF:
					op1.var = new_float_const
						(get_float(&op1) &&
						 get_float(&op2));
					break;
				case OP_AND_FV:
					op1.var = new_float_const
						(get_float(&op1) &&
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_AND_FS:
					op1.var = new_float_const
						(get_float(&op1) &&
						 op2.var->u.str->cstr[0]);
					break;
				case OP_AND_FE:
					op1.var = new_float_const
						(get_float(&op1) && 0);
					break;
				case OP_AND_VF:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) &&
						 get_float(&op2));
					break;
				case OP_AND_VV:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) &&
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_AND_VS:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) &&
						 op1.var->u.str->cstr[0]);
					break;
				case OP_AND_VE:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) && 0);
					break;
				case OP_AND_SF:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] &&
						 get_float(&op2));
					break;
				case OP_AND_SV:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] &&
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_AND_SS:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] &&
						 op1.var->u.str->cstr[0]);
					break;
				case OP_AND_SE:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] && 0);
					break;
				case OP_AND_EF:
					op1.var = new_float_const
						(0 && get_float(&op2));
					break;
				case OP_AND_EV:
					op1.var = new_float_const
						(0 && (op2.var->u.vec.x ||
						       op2.var->u.vec.y ||
						       op2.var->u.vec.z));
					break;
				case OP_AND_ES:
					op1.var = new_float_const
						(0 && op1.var->u.str->cstr[0]);
					break;
				case OP_AND_EE:
					op1.var = new_float_const(0 && 0);
					break;
				case OP_OR_FF:
					op1.var = new_float_const
						(get_float(&op1) ||
						 get_float(&op2));
					break;
				case OP_OR_FV:
					op1.var = new_float_const
						(get_float(&op1) ||
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_OR_FS:
					op1.var = new_float_const
						(get_float(&op1) ||
						 op2.var->u.str->cstr[0]);
					break;
				case OP_OR_FE:
					op1.var = new_float_const
						(get_float(&op1) || 0);
					break;
				case OP_OR_VF:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) ||
						 get_float(&op2));
					break;
				case OP_OR_VV:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) ||
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_OR_VS:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) ||
						 op1.var->u.str->cstr[0]);
					break;
				case OP_OR_VE:
					op1.var = new_float_const
						((op1.var->u.vec.x ||
						  op1.var->u.vec.y ||
						  op1.var->u.vec.z) || 0);
					break;
				case OP_OR_SF:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] ||
						 get_float(&op2));
					break;
				case OP_OR_SV:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] ||
						 (op2.var->u.vec.x ||
						  op2.var->u.vec.y ||
						  op2.var->u.vec.z));
					break;
				case OP_OR_SS:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] ||
						 op1.var->u.str->cstr[0]);
					break;
				case OP_OR_SE:
					op1.var = new_float_const
						(op1.var->u.str->cstr[0] || 0);
					break;
				case OP_OR_EF:
					op1.var = new_float_const
						(0 || get_float(&op2));
					break;
				case OP_OR_EV:
					op1.var = new_float_const
						(0 || (op2.var->u.vec.x ||
						       op2.var->u.vec.y ||
						       op2.var->u.vec.z));
					break;
				case OP_OR_ES:
					op1.var = new_float_const
						(0 || op1.var->u.str->cstr[0]);
					break;
				case OP_OR_EE:
					op1.var = new_float_const(0 || 0);
					break;
				}
				
				if (ptr->var)
				{
					push_op(move_op[exp->typ->typ],
						&op1, 0, ptr);
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
			else
			{
				if (!ptr->var)
				{
					ptr->var = new_tmp_var(exp->typ->typ);
					ptr->offset = 0;
				}

				push_op(exp->op, &op1, &op2, ptr);
			}
		}
		else
		{
			expr_compile(exp->left, 0, 0, 0);
			expr_compile(exp->right, 0, 0, 0);
		}
		break;

	case E_ASSIGN:
		op1.var = 0;
		op1.offset = 0;

		op2.var = 0;
		op2.offset = 0;

		expr_compile(exp->left, &op1, &op2, 0);

		if (op2.var)
		{
			if (ptr)
			{
				expr_compile(exp->right, ptr, 0, 0);
				push_op(store_e_op[exp->typ->typ],
					&op2, &op1, ptr);
			}
			else
			{
				op3.var = 0;
				op3.offset = 0;

				expr_compile(exp->right, &op3, 0, 0);
				push_op(store_e_op[exp->typ->typ], &op2, &op1, &op3);
			}
		}
		else
		{
			expr_compile(exp->right, &op1, 0, 0);

			if (ptr)
			{
				if (ptr->var)
				{
					push_op(exp->op, &op1, 0, ptr);
					tmp_ptr = tmp;
				}
				else
				{
					ptr->var = op1.var;
					ptr->offset = op1.offset;
				}
			}
		}

		break;

	case E_CALL:
		op1.var = 0;
		op1.offset = 0;
		op2.var = 0;
		op2.offset = 0;

		expr_compile(exp->left, &op1, 0, &op2);
		op3.var = exp->left->fn->params;
		expr_compile_args(&op3.var, exp->right, &op1);
		push_op(OP_CALL, &op2, &op1, 0);

		if (ptr)
		{
			tmp_ptr = tmp;

			if (!ptr->var)
			{
				ptr->var = new_tmp_var(exp->typ->typ);
				ptr->offset = 0;
			}

			op2.var = &exp->left->fn->var;
			op2.offset = 0;
			push_op(load_fn_op[exp->typ->typ], &op1, &op2, ptr);
		}
		break;
	}

	if (!ptr)
		tmp_ptr = tmp;

	return exp->typ->typ;
}

static variable *new_tmp_var(char typ)
{
	variable *var = *tmp_ptr;
	
	if (!var)
	{
		*tmp_ptr = var = new_variable(new_type(T_FLOAT, 0), 0);
		var->used = 1;
	}

	tmp_ptr = &var->next;

	if (typ == T_VECTOR)
	{
		new_tmp_var(T_FLOAT);
		new_tmp_var(T_FLOAT);
	}

	return var;
}

static float get_float(oprand *opr)
{
	variable *var = opr->var;

	switch(var->typ->typ)
	{
	case T_FLOAT:
		return var->u.flt;
	case T_VECTOR:
		switch(opr->offset)
		{
		case 0:
			return var->u.vec.x;
		case 1:
			return var->u.vec.y;
		case 2:
			return var->u.vec.z;
		}
	}
	return 0.0;
}

static void push_op(char op, oprand *a, oprand *b, oprand *c)
{
	instruction *ptr;
	variable *var;
	char offset;

	if (index >= 65535)
	{
		if (!overflow)
			yyerror("instruction buffer overflow");
		overflow = 1;
	}

	ptr = &instrs[index++];

	ptr->op = op;

	if (a)
	{
		if ((var = a->var))
			var->used = 1;
		offset = a->offset;
	}
	else
	{
		var = 0;
		offset = 0;
	}

	ptr->a.var = var;
	ptr->ao = offset;

	if (b)
	{
		if ((var = b->var))
			var->used = 1;
		offset = b->offset;
	}
	else
	{
		var = 0;
		offset = 0;
	}

	ptr->b.var = var;
	ptr->bo = offset;

	if (c)
	{
		if ((var = c->var))
			var->used = 1;
		offset = c->offset;
	}
	else
	{
		var = 0;
		offset = 0;
	}

	ptr->c.var = var;
	ptr->co = offset;
}

static instruction *push_goto_op(char op, oprand *a, int b, oprand *c)
{
	instruction *ptr;
	variable *var;
	char offset;

	if (index >= 65535)
	{
		if (!overflow)
			yyerror("instruction buffer overflow");
		overflow = 1;
		return 0;
	}

	ptr = &instrs[index++];

	ptr->op = op;

	if (a)
	{
		if ((var = a->var))
			var->used = 1;
		offset = a->offset;
	}
	else
	{
		var = 0;
		offset = 0;
	}

	ptr->a.var = var;
	ptr->ao = offset;

	ptr->b.idx = b;
	ptr->bo = 0;

	if (c)
	{
		if ((var = c->var))
			var->used = 1;
		offset = c->offset;
	}
	else
	{
		var = 0;
		offset = 0;
	}

	ptr->c.var = var;
	ptr->co = offset;

	return ptr;
}
