/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>

#include "compile.h"

static void index_variables(variable *, unsigned *);

static void output_strings(FILE *);
static void output_globals(FILE *);
static void output_functions(FILE *);
static void output_constants(FILE *, variable *);
static void output_instructions(FILE *, instruction *, unsigned);

static void put_float(FILE *, float);
static void put_int(FILE *, int);

static unsigned char type_size[] = { 0, 1, 3, 1, 1, 1 };


void calc_indexes(void)
{
	variable *var;
	function *fn;
	unsigned n;

	for (fn = functions; fn; fn = fn->next)
	{
		this_fn = fn;

		n = 1;

		for (var = fn->constants; var; var = var->next)
		{
			if (var->used && var->typ->typ == T_FUNCTION)
			{
				var->u.fn->var.used = 1;
			}
		}	

		index_variables(&fn->var, &n);
		index_variables(fn->params, &n);
		fn->nparams = n - 1;

		index_variables(fn->constants, &n);
		index_variables(fn->locals, &n);
		index_variables(fn->tmps, &n);
		fn->nvars = n;

		if (n > 65535)
		{
			yyerror("too many variables");
		}

		this_fn = 0;
	}

	n = 0;
	index_variables(globals, &n);

	n = 0;
	for (var = globals; var; var = var->next)
	{
		if (var->used && !var->name->idx)
		{
			var->name->idx = ++n;
		}
	}

	for (fn = functions; fn; fn = fn->next)
	{
		if (fn->var.used && fn->var.name && !fn->var.name->idx)
		{
			fn->var.name->idx = ++n;
		}

		for (var = fn->constants; var; var = var->next)
		{
			if (var->used && var->typ->typ == T_STRING &&
			    var->u.str->cstr[0] && !var->u.str->idx)
			{
				var->u.str->idx = ++n;
			}
		}
	}

	if (n > 65535)
	{
		yyerror("too many strings");
	}
}

static void index_variables(variable *var, unsigned *ptr)
{
	while (var)
	{
		if (var->used)
		{
			var->idx = *ptr;
			*ptr += type_size[var->typ->typ];
		}
		var = var->next;
	}
}

void output(FILE *fp)
{
	fwrite("CJPO", 1, 4, fp);
	output_strings(fp);
	output_globals(fp);
	output_functions(fp);
}

static void output_strings(FILE *fp)
{
	unsigned i, n = 0;
	string *str;

	for (i = 0; i < 256; i++)
	{
		for (str = string_hash[i]; str; str = str->next)
		{
			if (str->idx) str->idx = ++n;
		}
	}

	put_int(fp, n);

	for (i = 0; i < 256; i++)
	{
		for (str = string_hash[i]; str; str = str->next)
		{
			if (str->idx)
			{
				n = strlen(str->cstr);

				if (n > 255) n = 255;

				fputc(n, fp);
				fwrite(str->cstr, 1, n, fp);
			}
		}
	}
}

static void output_globals(FILE *fp)
{
	unsigned n = 0;
	variable *var;

	for (var = globals; var; var = var->next)
	{
		if (var->used) n++;
	}

	put_int(fp, n);

	for (var = globals; var; var = var->next)
	{
		if (var->used)
		{
			fputc(var->typ->typ, fp);
			put_int(fp, var->name->idx);
		}
	}
}

static void output_functions(FILE *fp)
{
	function *fn;
	unsigned n = 0;

	for (fn = functions; fn; fn = fn->next)
	{
		if (fn->var.defd) n++;
	}

	put_int(fp, n);

	for (fn = functions; fn; fn = fn->next)
	{
		if (fn->var.defd)
		{
			put_int(fp, fn->var.name->idx);
			put_int(fp, fn->nparams);
			put_int(fp, fn->nvars);

			output_constants(fp, fn->constants);
			output_instructions(fp, fn->instrs, fn->ninstrs);
		}
	}
}

static void output_constants(FILE *fp, variable *constants)
{
	variable *var;
	unsigned n = 0;

	for (var = constants; var; var = var->next)
	{
		if (var->used) n++;
	}

	put_int(fp, n);

	for (var = constants; var; var = var->next)
	{
		if (var->used)
		{
			fputc(var->typ->typ, fp);
			switch(var->typ->typ)
			{
			case T_FLOAT:
				put_float(fp, var->u.flt);
				break;

			case T_VECTOR:
				put_float(fp, var->u.vec.x);
				put_float(fp, var->u.vec.y);
				put_float(fp, var->u.vec.z);
				break;

			case T_STRING:
				put_int(fp, var->u.str->idx);
				break;

			case T_FUNCTION:
				put_int(fp, var->u.fn->var.name->idx);
				break;
			}
		}
	}
}

void output_instructions(FILE *fp, instruction *instrs, unsigned ninstrs)
{
	instruction *instr;
	variable *var;
	unsigned i, n;

	put_int(fp, ninstrs);

	for (n = 0; n < ninstrs; n++)
	{
		instr = &instrs[n];

		fputc(instr->op, fp);

		i = instr->ao;
		if ((var = instr->a.var))
			i += var->idx;
		put_int(fp, i);

		i = instr->bo;
		if ((instr->op >= OP_IF_F) && (instr->op <= OP_GOTO))
		{
			i += instr->b.idx;
		}
		else
		{
			if ((var = instr->b.var))
				i += var->idx;
		}
		put_int(fp, i);

		i = instr->co;
		if ((var = instr->c.var))
			i += var->idx;
		put_int(fp, i);
	}
}

static void put_float(FILE *fp, float f)
{
	union { int i; float f; } u;
	unsigned char c[4];

	u.f = f;
	c[0] = u.i;
	c[1] = (u.i >> 8);
	c[2] = (u.i >> 16);
	c[3] = (u.i >> 24);

	fwrite(&c, 1, 4, fp);
}

static void put_int(FILE *fp, int i)
{
	unsigned char c[2];

	c[0] = i;
	c[1] = (i >> 8);

	fwrite(&c, 1, 2, fp);
}
