/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define T_VOID     0
#define T_FLOAT    1
#define T_VECTOR   2
#define T_STRING   3
#define T_FUNCTION 4
#define T_ENTITY   5

struct variable
{
	unsigned type;
	union {
		float flt;
		char *str;
	} u;
};

struct global
{
	unsigned short type, offset;
	char *name;
};

static int disasm(FILE *);
static int disasm_fn(FILE *);
static void print_oprand(unsigned, unsigned, unsigned);
static char *get_str(FILE *);
static int get_float(FILE *, float *);
static int get_int(FILE *);

static char *strs[65536];
static struct global globs[65536];
static struct variable vars[65536];

static char *types[] =
{
	"void",
	"float",
	"vector",
	"string",
	"function",
	"entity"
};

static int sizes[] = { 0, 1, 3, 1, 1, 1 };

static struct { char *op; unsigned char a, b, c; } opcodes[] =
{
	{ "DONE",     0, 0, 0 }, /* - - - */

	{ "LOAD_EF ", 5, 6, 1 }, /* e x f */
        { "LOAD_EV ", 5, 6, 2 }, /* e x v */
	{ "LOAD_ES ", 5, 6, 3 }, /* e x s */
	{ "LOAD_EN ", 5, 6, 4 }, /* e x n */
	{ "LOAD_EE ", 5, 6, 5 }, /* e x e */
	{ "STORE_EF", 5, 6, 1 }, /* e x f */
	{ "STORE_EV", 5, 6, 2 }, /* e x v */
	{ "STORE_ES", 5, 6, 3 }, /* e x s */
	{ "STORE_EN", 5, 6, 4 }, /* e x n */
	{ "STORE_EE", 5, 6, 5 }, /* e x e */

	{ "LOAD_NF ", 4, 7, 1 }, /* n x f */
	{ "LOAD_NV ", 4, 7, 2 }, /* n x v */
	{ "LOAD_NS ", 4, 7, 3 }, /* n x s */
	{ "LOAD_NN ", 4, 7, 4 }, /* n x n */
	{ "LOAD_NE ", 4, 7, 5 }, /* n x e */
	{ "STORE_NF", 4, 7, 1 }, /* n x f */
	{ "STORE_NV", 4, 7, 2 }, /* n x v */
	{ "STORE_NS", 4, 7, 3 }, /* n x s */
	{ "STORE_NN", 4, 7, 4 }, /* n x n */
	{ "STORE_NE", 4, 7, 5 }, /* n x e */

	{ "IF_F    ", 1, 8, 0 }, /* f a - */
	{ "IF_V    ", 2, 8, 0 }, /* v a - */
	{ "IF_S    ", 3, 8, 0 }, /* s a - */
	{ "IF_N    ", 4, 8, 0 }, /* n a - */
	{ "IF_E    ", 5, 8, 0 }, /* e a - */
	{ "IFNOT_F ", 1, 8, 0 }, /* f a - */
	{ "IFNOT_V ", 2, 8, 0 }, /* v a - */
	{ "IFNOT_S ", 3, 8, 0 }, /* s a - */
	{ "IFNOT_N ", 4, 8, 0 }, /* n a - */
	{ "IFNOT_E ", 5, 8, 0 }, /* e a - */

	{ "GOTO    ", 0, 8, 0 }, /* - a - */
	{ "CALL    ", 5, 4, 0 }, /* e n - */
	{ "RETURN",   0, 0, 0 }, /* - - - */

	{ "MOVE_F  ", 1, 0, 1 }, /* f - f */
	{ "MOVE_V  ", 2, 0, 2 }, /* v - v */
	{ "MOVE_S  ", 3, 0, 3 }, /* s - s */
	{ "MOVE_N  ", 4, 0, 4 }, /* n - n */
	{ "MOVE_E  ", 5, 0, 5 }, /* e - e */

	{ "MUL_F   ", 1, 1, 1 }, /* f f f */
	{ "MUL_V   ", 2, 2, 1 }, /* v v f */
	{ "MUL_FV  ", 1, 2, 2 }, /* f v v */
	{ "MUL_VF  ", 2, 1, 2 }, /* v f v */
	{ "DIV_F   ", 1, 1, 1 }, /* f f f */
	{ "DIV_VF  ", 2, 1, 2 }, /* v f v */
	{ "ADD_F   ", 1, 1, 1 }, /* f f f */
	{ "ADD_V   ", 2, 2, 2 }, /* v v v */
	{ "SUB_F   ", 1, 1, 1 }, /* f f f */
	{ "SUB_V   ", 2, 2, 2 }, /* v v v */
	{ "NEG_F   ", 1, 0, 1 }, /* f - f */
	{ "NEG_V   ", 2, 0, 2 }, /* v - v */

	{ "EQ_F    ", 1, 1, 1 }, /* f f f */
	{ "EQ_V    ", 2, 2, 1 }, /* v v f */
	{ "EQ_S    ", 3, 3, 1 }, /* s s f */
	{ "EQ_N    ", 4, 4, 1 }, /* n n f */
	{ "EQ_E    ", 5, 5, 1 }, /* e e f */
	{ "NE_F    ", 1, 1, 1 }, /* f f f */
	{ "NE_V    ", 2, 2, 1 }, /* v v f */
	{ "NE_S    ", 3, 3, 1 }, /* s s f */
	{ "NE_N    ", 4, 4, 1 }, /* n n f */
	{ "NE_E    ", 5, 5, 1 }, /* e e f */
	{ "LE_F    ", 1, 1, 1 }, /* f f f */
	{ "GE_F    ", 1, 1, 1 }, /* f f f */
	{ "LT_F    ", 1, 1, 1 }, /* f f f */
	{ "GT_F    ", 1, 1, 1 }, /* f f f */

	{ "AND_FF  ", 1, 1, 1 }, /* f f f */
	{ "AND_FV  ", 1, 2, 1 }, /* f v f */
	{ "AND_FS  ", 1, 3, 1 }, /* f s f */
	{ "AND_FN  ", 1, 4, 1 }, /* f n f */
	{ "AND_FE  ", 1, 5, 1 }, /* f e f */
	{ "AND_VF  ", 2, 1, 1 }, /* v f f */
	{ "AND_VV  ", 2, 2, 1 }, /* v v f */
	{ "AND_VS  ", 2, 3, 1 }, /* v s f */
	{ "AND_VN  ", 2, 4, 1 }, /* v n f */
	{ "AND_VE  ", 2, 5, 1 }, /* v e f */
	{ "AND_SF  ", 3, 1, 1 }, /* s f f */
	{ "AND_SV  ", 3, 2, 1 }, /* s v f */
	{ "AND_SS  ", 3, 3, 1 }, /* s s f */
	{ "AND_SN  ", 3, 4, 1 }, /* s n f */
	{ "AND_SE  ", 3, 5, 1 }, /* s e f */
	{ "AND_NF  ", 4, 1, 1 }, /* n f f */
	{ "AND_NV  ", 4, 2, 1 }, /* n v f */
	{ "AND_NS  ", 4, 3, 1 }, /* n s f */
	{ "AND_NN  ", 4, 4, 1 }, /* n n f */
	{ "AND_NE  ", 4, 5, 1 }, /* n e f */
	{ "AND_EF  ", 5, 1, 1 }, /* e f f */
	{ "AND_EV  ", 5, 2, 1 }, /* e v f */
	{ "AND_ES  ", 5, 3, 1 }, /* e s f */
	{ "AND_EN  ", 5, 4, 1 }, /* e n f */
	{ "AND_EE  ", 5, 5, 1 }, /* e e f */
	{ "OR_FF   ", 1, 1, 1 }, /* f f f */
	{ "OR_FV   ", 1, 2, 1 }, /* f v f */
	{ "OR_FS   ", 1, 3, 1 }, /* f s f */
	{ "OR_FN   ", 1, 4, 1 }, /* f n f */
	{ "OR_FE   ", 1, 5, 1 }, /* f e f */
	{ "OR_VF   ", 2, 1, 1 }, /* v f f */
	{ "OR_VV   ", 2, 2, 1 }, /* v v f */
	{ "OR_VS   ", 2, 3, 1 }, /* v s f */
	{ "OR_VN   ", 2, 4, 1 }, /* v n f */
	{ "OR_VE   ", 2, 5, 1 }, /* v e f */
	{ "OR_SF   ", 3, 1, 1 }, /* s f f */
	{ "OR_SV   ", 3, 2, 1 }, /* s v f */
	{ "OR_SS   ", 3, 3, 1 }, /* s s f */
	{ "OR_SN   ", 3, 4, 1 }, /* s n f */
	{ "OR_SE   ", 3, 5, 1 }, /* s e f */
	{ "OR_NF   ", 4, 1, 1 }, /* n f f */
	{ "OR_NV   ", 4, 2, 1 }, /* n v f */
	{ "OR_NS   ", 4, 3, 1 }, /* n s f */
	{ "OR_NN   ", 4, 4, 1 }, /* n n f */
	{ "OR_NE   ", 4, 5, 1 }, /* n e f */
	{ "OR_EF   ", 5, 1, 1 }, /* e f f */
	{ "OR_EV   ", 5, 2, 1 }, /* e v f */
	{ "OR_ES   ", 5, 3, 1 }, /* e s f */
	{ "OR_EN   ", 5, 4, 1 }, /* e n f */
	{ "OR_EE   ", 5, 5, 1 }, /* e e f */
	{ "NOT_F   ", 1, 0, 1 }, /* f - f */
	{ "NOT_V   ", 2, 0, 1 }, /* v - f */
	{ "NOT_S   ", 3, 0, 1 }, /* s - f */
	{ "NOT_N   ", 4, 0, 1 }, /* n - f */
	{ "NOT_E   ", 5, 0, 1 }, /* e - f */

	{ "BITAND_F", 1, 1, 1 }, /* f f f */
	{ "BITOR_F ", 1, 1, 1 }, /* f f f */
	{ "BITEOR_F", 1, 1, 1 }, /* f f f */
	{ "BITNOT_F", 1, 0, 1 }  /* f - f */
};

int main(int argc, char **argv)
{
	FILE *fp;

	switch (argc)
	{
	case 1:
		if (disasm(stdin))
		{
			fprintf(stderr, "%s: error.\n", argv[0]);
			return 1;
		}
		break;

	case 2:
		fp = fopen(argv[1], "r");
		if (disasm(fp))
		{
			fclose(fp);
			fprintf(stderr, "%s: error.\n",
				argv[1]);
			return 1;
		}
		fclose(fp);
		break;

	default:
		fprintf(stderr, "Usage: %s <file>\n", argv[0]);
		return 1;
	}

	return 0;
}

static int disasm(FILE *fp)
{
	unsigned i, n, v;
	char magic[4];

	if ( fread(magic, 1, 4, fp) != 4 ||
	     strncmp(magic, "CJPO", 4) )
		return -1;

	/* strings */

	if ((signed)(n = get_int(fp)) < 0)
		return -1;

	strs[0] = "";

	for (i = 1; i <= n; i++)
	{
		if (!(strs[i] = get_str(fp)))
			return -1;
	}

	/* variables */

	if ((signed)(n = get_int(fp)) < 0)
		return -1;

	v = 0;

	for (i = 0; i < n; i++)
	{
		unsigned c, s, j;

		if ((c = fgetc(fp)) > 5)
			return -1;
		if ((signed)(s = get_int(fp)) < 0)
			return -1;

		printf("%s %s;\n\n", types[c], strs[s]);

		for (j = 0; j < sizes[c]; j++)
		{
			globs[v].type = c;
			globs[v].offset = j;
			globs[v].name = strs[s];
			v++;
		}
	}

	/* functions */

	if ((signed)(n = get_int(fp)) < 0)
		return -1;

	if (!n) return 0;

	for (i = 0; i < n; i++)
	{
		if (disasm_fn(fp))
			return -1;
	}

	return 0;
}

static int disasm_fn(FILE *fp)
{
	unsigned i, j, n;
	
	if ((signed)(i = get_int(fp)) < 0)
		return -1;

	if ((signed)(j = get_int(fp)) < 0)
		return -1;

	if ((signed)(n = get_int(fp)) < 0)
		return -1;

	printf("function %s(%u) %u =\n{\n", strs[i], j, n);

	/* constants */

	for (i = 0; i < n; i++)
	{
		vars[j].type = 0;
	}

	j++;

	if ((signed)(n = get_int(fp)) < 0)
		return -1;

	for (i = 0; i < n; i++)
	{
		unsigned c;

		if ((c = fgetc(fp)) > 255)
			return -1;

		switch(c)
		{
		case T_FLOAT:
			vars[j].type = T_FLOAT;
			if (get_float(fp, &vars[j].u.flt))
				return -1;
			j ++;
			break;

		case T_VECTOR:
			vars[j + 0].type = T_FLOAT;
			if (get_float(fp, &vars[j + 0].u.flt))
				return -1;
			vars[j + 1].type = T_FLOAT;
			if (get_float(fp, &vars[j + 1].u.flt))
				return -1;
			vars[j + 2].type = T_FLOAT;
			if (get_float(fp, &vars[j + 2].u.flt))
				return -1;
			j += 3;
			break;

		case T_STRING:
		case T_FUNCTION:
			vars[j].type = c;
			if ((signed)(c = get_int(fp)) < 0)
				return -1;
			vars[j].u.str = strs[c];
			j ++;
			break;
		}
	}

	/* code */

	if ((signed)(n = get_int(fp)) <= 0)  /* 0 is not allowed */
		return -1;

	for (i = 0; i < n; i++)
	{
		unsigned op, a, b, c;

		if ((op = fgetc(fp)) > 255)
			return -1;

		if ((signed)(a = get_int(fp)) < 0)
			return -1;

		if ((signed)(b = get_int(fp)) < 0)
			return -1;

		if ((signed)(c = get_int(fp)) < 0)
			return -1;

		printf("%u\t%s", i, opcodes[op].op);
		print_oprand(opcodes[op].a, a, opcodes[op].c);
		print_oprand(opcodes[op].b, b, opcodes[op].c);
		print_oprand(opcodes[op].c, c, opcodes[op].c);
		printf("\n");
	}

	printf("}\n\n");

	return 0;
}

static void print_oprand(unsigned type, unsigned val, unsigned t)
{
	switch(type)
	{
	case 1:
		if (vars[val].type == T_FLOAT)
		{
			printf(" %g", vars[val].u.flt);
		}
		else
		{
			printf(" (%u)", val);
		}
		break;

	case 2:
		if (vars[val + 0].type == T_FLOAT &&
		    vars[val + 1].type == T_FLOAT &&
		    vars[val + 2].type == T_FLOAT)
		{
			printf(" '%g %g %g'", vars[val + 0].u.flt,
			       vars[val + 1].u.flt, vars[val + 2].u.flt);
		}
		else
		{
			printf(" (%u)", val);			
		}
		break;

	case 3:
		if (vars[val].type == T_STRING)
		{
			printf(" \"%s\"", vars[val].u.str);
		}
		else
		{
			printf(" (%u)", val);			
		}
		break;

	case 4:
		if (vars[val].type == T_FUNCTION)
		{
			printf(" %s", vars[val].u.str);
		}
		else
		{
			printf(" (%u)", val);			
		}
		break;

	case 5:
		if (val)
		{
			printf(" (%u)", val);
		}
		else
		{
			printf(" this");
		}
		break;

	case 6:	
		if ( (t == 1) && (globs[val].type == T_VECTOR) )
		{
			printf(" %s.%c", globs[val].name,
			       "xyz"[globs[val].offset]);
		}
		else
		{
			printf(" %s", globs[val].name);
		}
		break;

	case 7:
		printf(" (%u)", val);
		break;

	case 8:
		printf(" %u", val);
		break;
	}
}

static char *get_str(FILE *fp)
{
	unsigned i;
	char *str;

	if ( (i = fgetc(fp)) > 255 )
		return 0;

	str = malloc(i + 1);

	if (i)
	{
		if (fread(str, 1, i, fp) != i)
			return 0;
	}
	str[i] = '\0';

	return str;
}

static int get_float(FILE *fp, float *ptr)
{
	union { int i; float f; } u;
	unsigned char c[4];

	if (fread(&c, 1, 4, fp) != 4)
		return -1;

	u.i = (c[3] << 24) + (c[2] << 16) + (c[1] << 8) + c[0];
	*ptr = u.f;

	return 0;
}

static int get_int(FILE *fp)
{
	unsigned char c[2];

	if (fread(c, 1, 2, fp) != 2)
		return -1;
	return ((c[1] << 8) + c[0]);
}
