/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

struct texture
{
	char type;
	float value;
	char *name;
};

static struct texture tex[255];
static unsigned ntex;

static unsigned char *map;
static unsigned width, height;
static unsigned tile;

static void usage(char *);
static void set_texture(char, float, char *);
static void put_square(FILE *, unsigned, unsigned);
static unsigned choose_texture(int, unsigned, unsigned, unsigned);

int main(int argc, char **argv)
{
	FILE *ifp, *ofp;
	unsigned i, j, l;
	char c;

	if (argc < 4)
	{
		usage(argv[0]);
		return 1;
	}

	set_texture(0, 0.0, argv[3]);

	for (i = 4; i < argc; i++)
	{
		char *arg = argv[i];
		char *val;
		float num;

		if (arg[0] != '-')
		{
			usage(argv[0]);
			return 1;
		}

		switch(arg[1])
		{
		case 't':
			if (!(val = argv[++i]))
			{
				usage(argv[0]);
				return 1;
			}

			tile = atoi(val);
			break;

		case 'a':
		case 'b':
		case 's':
			if (!(val = argv[++i]))
			{
				usage(argv[0]);
				return 1;
			}

			num = atof(val);

			if (!(val = argv[++i]))
			{
				usage(argv[0]);
				return 1;
			}

			set_texture(arg[1], num, val);
			break;

		default:
			usage(argv[0]);
			return 1;
		}
	}

	if (tile > 5) tile = 0;
	tile = 1U << tile;

	if (! (ifp = fopen(argv[2], "r")) )
		return 1;

	if (fscanf(ifp, "P5 %u %u %u%c", &width, &height, &l, &c) != 4)
		return 1;

	if (l != 255 || c != '\n')
		return 1;

	if (width < 513 || height < 513)
		return 1;

	map = malloc(l = (width * height));

	if (fread(map, 1, l, ifp) != l)
		return 1;

	fclose(ifp);

	if (! (ofp = fopen(argv[1], "w")) )
		return 1;

	fwrite("CJPL", 1, 4, ofp);

	fputc(ntex, ofp);

	for (i = 0; i < ntex; i++)
	{
		l = strlen(tex[i].name);
		fputc(l, ofp);
		fwrite(tex[i].name, 1, l, ofp);
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			put_square(ofp, i*2, j*2);
		}
	}

	fclose(ofp);

	return 0;
}

void usage(char *prog)
{
	fprintf(stderr,	"Usage: %s <output> <hightmap> <texture> [-t <exp>]\n", prog);
	fprintf(stderr, " [-a <height> <texture>] [-b <height> <texture>] [-s <slope> <texture>]\n");
}

static void set_texture(char type, float value, char *name)
{
	char *s;

	s = strrchr(name, '/');
	s = s ? s+1 : name;

	tex[ntex].type = type;
	tex[ntex].value = value;
	tex[ntex].name = s;

	ntex++;
}

struct vertex
{
	signed char x, y, z;
	unsigned char u, v;
}
vtx[5] =
{
	{ 0,0,0, 0,0 },
	{ 16,0,0, 64,0 },
	{ 0,16,0, 0,64 },
	{ 16,16,0, 64,64 },
	{ 8,8,0, 32,32 }
};

static void put_square(FILE *fp, unsigned x, unsigned y)
{
	unsigned i, j, k;
	int z, min, max;

	i = j = 64 / tile;
	k = 32 / tile;
	i *= (x/2) & (tile-1);
	j *= (y/2) & (tile-1);

	vtx[0].u = i;
	vtx[0].v = j;
	vtx[1].v = j;
	vtx[2].u = i;

	i += k;
	j += k;

	vtx[4].u = i;
	vtx[4].v = j;

	i += k;
	j += k;

	vtx[1].u = i;
	vtx[2].v = j;
	vtx[3].u = i;
	vtx[3].v = j;

	z = map[(x * width) + y];
	min = max = z;
	vtx[0].z = z;

	z = map[((x+2) * width) + y];
	if (z < min) min = z;
	if (z > max) max = z;
	vtx[1].z = z;

	z = map[(x * width) + y+2];
	if (z < min) min = z;
	if (z > max) max = z;
	vtx[2].z = z;

	z = map[((x+2) * width) + y+2];
	if (z < min) min = z;
	if (z > max) max = z;
	vtx[3].z = z;

	z = map[((x+1) * width) + y+1];
	vtx[4].z = z;

	if ((min - z) < -128) z = min + 128;
	if ((max - z) > 127) z = max - 127;

	for (i = 0; i < 5; i++)
	{
		vtx[i].z -= z;
	}

	fputc(z, fp);
	fputc(4, fp);

	i = choose_texture(z, 0, 1, 4);
	fputc(i, fp);
	fwrite(&vtx[0], 1, 5, fp);
	fwrite(&vtx[1], 1, 5, fp);
	fwrite(&vtx[4], 1, 5, fp);

	i = choose_texture(z, 1, 3, 4);
	fputc(i, fp);
	fwrite(&vtx[1], 1, 5, fp);
	fwrite(&vtx[3], 1, 5, fp);
	fwrite(&vtx[4], 1, 5, fp);

	i = choose_texture(z, 3, 2, 4);
	fputc(i, fp);
	fwrite(&vtx[3], 1, 5, fp);
	fwrite(&vtx[2], 1, 5, fp);
	fwrite(&vtx[4], 1, 5, fp);

	i = choose_texture(z, 2, 0, 4);
	fputc(i, fp);
	fwrite(&vtx[2], 1, 5, fp);
	fwrite(&vtx[0], 1, 5, fp);
	fwrite(&vtx[4], 1, 5, fp);
}

unsigned choose_texture(int z, unsigned v1, unsigned v2, unsigned v3)
{
	unsigned i;
	unsigned t = 0;
	int min, max;
	float ax, ay, az;
	float bx, by, bz;
	float nx, ny, nz;
	float slope;

	min = max = vtx[v1].z;
	if (vtx[v2].z < min) min = vtx[v2].z;
	if (vtx[v2].z > max) max = vtx[v2].z;
	if (vtx[v3].z < min) min = vtx[v3].z;
	if (vtx[v3].z > max) max = vtx[v3].z;
	min += z;
	max += z;
	
	ax = vtx[v2].x - vtx[v1].x;
	ay = vtx[v2].y - vtx[v1].y;
	az = vtx[v2].z - vtx[v1].z;

	bx = vtx[v3].x - vtx[v1].x;
	by = vtx[v3].y - vtx[v1].y;
	bz = vtx[v3].z - vtx[v1].z;

	nx = (ay * bz) - (az * by);
	ny = (az * bx) - (ax * bz);
	nz = (ax * by) - (ay * bx);

	nx *= nx;
	ny *= ny;
	nz *= nz;

	nz = nz / (nx + ny + nz);

	slope = sqrt((1.0 - nz) / nz);

	for (i = 0; i < ntex; i++)
	{
		switch(tex[i].type)
		{
		case 'a':
			if (max > tex[i].value)
				t = i;
			break;
		case 'b':
			if (min < tex[i].value)
				t = i;
			break;
		case 's':
			if (slope > tex[i].value)
				t = i;
			break;
		default:
			t = i;
			break;
		}
	}

	return t + 1;
}
