/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* types */

#define T_VOID     0
#define T_FLOAT    1
#define T_VECTOR   2
#define T_STRING   3
#define T_FUNCTION 4
#define T_ENTITY   5

/* input definitions */

static int input(const char *);
static int load(FILE *);

static int get_int(FILE *);
static int get_float(FILE *, float *);

/* output definitions */

static int output(const char *);
static void save(FILE *);

static void put_int(FILE *, int);
static void put_float(FILE *, float);

/* string definitions */

typedef struct string
{
        struct string *next;
        unsigned short idx;
	unsigned char vtype;
	unsigned char vidx;
        const char *cstr;
} string;

static string *new_string(const char *);
static string *load_string(FILE *);
static void save_string(FILE *, string *);
static int string_set_var_type(string *, unsigned char);

static unsigned string_index = 0;
static unsigned string_nvars = 0;
static string *string_hash[256];
static string *strings[63336];

static unsigned variable_index = 0;
static string *variables[256];

/* function definitions */

typedef struct vector
{
	float x, y, z;
} vector;

typedef struct constant
{
	unsigned type;
	union
	{
		unsigned idx;
		float flt;
		vector v;
	} u;
} constant;

typedef struct instruction
{
	unsigned char op;
	unsigned short a, b, c;
} instruction;

typedef struct function
{
	struct function *next;
	struct string *name;
	instruction *code;
	constant *consts;
	unsigned short nparam;
	unsigned short nvar;
	unsigned short ncode;
	unsigned short nconsts;
} function;

static function *functions = 0;
static unsigned function_num = 0;

static function *new_function(string *, unsigned);
static int load_function(FILE *, string **, unsigned char *);
static void save_function(FILE *, function *);

/* main */

int main(int argc, char **argv)
{
	int i, n;

	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s <input>... <output>\n", argv[0]);
                return 1;
	}

	n = argc - 1;

	for (i = 1; i < n; i++)
	{
		if (input(argv[i]))
			return 1;
	}

	output(argv[n]);

	return 0;
}

/* input functions */

static int input(const char *name)
{
	FILE *fp;
	int ret;

	if ( !(fp = fopen(name, "rb")) )
	{
		perror(name);
		return 1;
	}

	if ((ret = load(fp)))
	{
		fprintf(stderr, "%s: input error\n", name);
	}

	if (fclose(fp))
	{
		perror(name);
		return 1;
	}

	return ret;
}	

static int load(FILE *fp)
{
	char magic[4];
	int i, n;

	string **strs;

	unsigned char vars[256];
	unsigned vi = 0;

	if ( fread(magic, 1, 4, fp) != 4 ||
	     strncmp(magic, "CJPO", 4) )
		return -1;

	/* strings */

	if ((n = get_int(fp)) < 0)
		return -1;

	strs = (string **)alloca((n + 1) * sizeof(string *));

	strs[0] = new_string("");

	for (i = 1; i <= n; i++)
	{
		if (!(strs[i] = load_string(fp)))
			return -1;
	}

	/* variables */

	if ((n = get_int(fp)) < 0)
		return -1;

	for (i = 0; i < n; i++)
	{
		unsigned c;
		int s;

		if ((c = fgetc(fp)) > 255)
			return -1;
		if ((s = get_int(fp)) < 0)
			return -1;
		if (string_set_var_type(strs[s], c))
			return -1;

		s = strs[s]->vidx;

		vars[vi++] = s;
		if (c == T_VECTOR)
		{
			vars[vi++] = s + 1;
			vars[vi++] = s + 2;
		}
	}
	
	/* functions */

	if ((n = get_int(fp)) < 0)
		return -1;

	for (i = 0; i < n; i++)
	{
		if (load_function(fp, strs, vars))
			return -1;
	}

	return 0;
}

static int get_int(FILE *fp)
{
	unsigned char c[2];

	if (fread(c, 1, 2, fp) != 2)
		return -1;
	return ((c[1] << 8) + c[0]);
}

static int get_float(FILE *fp, float *ptr)
{
	union { int i; float f; } u;
	unsigned char c[4];

	if (fread(&c, 1, 4, fp) != 4)
		return -1;

	u.i = (c[3] << 24) + (c[2] << 16) + (c[1] << 8) + c[0];
	*ptr = u.f;

	return 0;
}

/* output functions */

static int output(const char *name)
{
	FILE *fp;

	if ( !(fp = fopen(name, "wb")) )
	{
		perror(name);
		return 1;
	}

	save(fp);

	if (fclose(fp))
	{
		perror(name);
		return 1;
	}

	return 0;
}

static void save(FILE *fp)
{
	unsigned i;
	function *fn;

	fwrite("CJPO", 1, 4, fp);

	/* strings */

	put_int(fp, string_index - 1);

	for (i = 1; i < string_index; i++)
	{
		save_string(fp, strings[i]);
	}

	/* variables */
	
	put_int(fp, variable_index);

	for (i = 0; i < variable_index; i++)
	{
		fputc(variables[i]->vtype, fp);
		put_int(fp, variables[i]->idx);
	}

	/* functions */

	put_int(fp, function_num);

	for (fn = functions; fn; fn = fn->next)
	{
		save_function(fp, fn);
	}
}

static void put_int(FILE *fp, int i)
{
	unsigned char c[2];

	c[0] = i;
	c[1] = (i >> 8);

	fwrite(&c, 1, 2, fp);
}

static void put_float(FILE *fp, float f)
{
	union { int i; float f; } u;
	unsigned char c[4];

	u.f = f;
	c[0] = u.i;
	c[1] = (u.i >> 8);
	c[2] = (u.i >> 16);
	c[3] = (u.i >> 24);

	fwrite(&c, 1, 4, fp);
}

/* string functions */

string *new_string(const char *cstr)
{
        string *str, **ptr = &string_hash[(unsigned char)cstr[0]];

	if (string_index >= 65536)
		return 0;

        while ((str = *ptr))
        {
                if (!strncmp(str->cstr, cstr, 255))
                        return str;
                ptr = &str->next;
        }

        *ptr = str = malloc(sizeof(string));
        str->next = 0;
        str->idx = string_index++;
	str->vtype = 0;
	str->vidx = 0;
        str->cstr = strdup(cstr);

	strings[str->idx] = str;

        return str;
}

static string *load_string(FILE *fp)
{
	unsigned i;
	char cstr[256];

	if ( (i = fgetc(fp)) > 255 )
		return 0;
	if (i)
	{
		if (fread(cstr, 1, i, fp) != i)
			return 0;
	}
	cstr[i] = '\0';

	return new_string(cstr);
}

static void save_string(FILE *fp, string *str)
{
	unsigned n = strlen(str->cstr);

	if (n > 255) n = 255;

	fputc(n, fp);
	fwrite(str->cstr, 1, n, fp);
}

static int string_set_var_type(string *str, unsigned char t)
{
	if (str->vtype) return (str->vtype != t);
	
	str->vtype = t;
	str->vidx = string_nvars;

	string_nvars += ((t == T_VECTOR) ? 3 : 1);

	if (variable_index < 256)
		variables[variable_index++] = str;

	return (string_nvars > 256);
}

/* function functions */

static function *new_function(string *name, unsigned nparam)
{
	function *fn, **ptr = &functions;

	while ((fn = *ptr))
	{
		if (fn->name == name)
		{
			if (fn->nparam == nparam)
				return fn;
			return 0;
		}
		ptr = &fn->next;
	}

	*ptr = fn = malloc(sizeof(function));
	fn->next = 0;
	fn->name = name;
	fn->code = 0;
	fn->consts = 0;
	fn->nparam = nparam;
	fn->nvar = 0;
	fn->ncode = 0;
	fn->nconsts = 0;

	if (++function_num >= 65536)
		return 0;

	return fn;
}

static int load_function(FILE *fp, string **strs, unsigned char *vars)
{
	int i, n;
	function *fn;

	if ((i = get_int(fp)) < 0)
		return -1;

	if ((n = get_int(fp)) < 0)
		return -1;

	if (!(fn = new_function(strs[i], n)))
		return -1;

	if ((n = get_int(fp)) < 0)
		return -1;

	fn->nvar = n;

	/* constants */

	if ((n = get_int(fp)) < 0)
		return -1;

	if (fn->consts) free(fn->consts);
	fn->consts = malloc(n * sizeof(constant));
	fn->nconsts = n;

	for (i = 0; i < n; i++)
	{
		constant *cst = &fn->consts[i];
		unsigned c;

		if ((c = fgetc(fp)) > 255)
			return -1;

		cst->type = c;

		switch(c)
		{
		case T_FLOAT:
			if (get_float(fp, &cst->u.flt))
				return -1;
			break;

		case T_VECTOR:
			if (get_float(fp, &cst->u.v.x))
				return -1;
			if (get_float(fp, &cst->u.v.y))
				return -1;
			if (get_float(fp, &cst->u.v.z))
				return -1;
			break;

		case T_STRING:
		case T_FUNCTION:
			if ((c = get_int(fp)) < 0)
				return -1;
			cst->u.idx = strs[c]->idx;
			break;
		}
	}

	/* code */

	if ((n = get_int(fp)) <= 0)  /* 0 is not allowed */
		return -1;

	if (fn->code) free(fn->code);
	fn->code = malloc(n * sizeof(instruction));
	fn->ncode = n;

	for (i = 0; i < n; i++)
	{
		instruction *inst = &fn->code[i];
		int j;

		if ((inst->op = fgetc(fp)) > 255)
			return -1;

		if ((j = get_int(fp)) < 0)
			return -1;

		inst->a = j;

		if ((j = get_int(fp)) < 0)
			return -1;

		if (inst->op >= 1 && inst->op <= 10)
		{
			inst->b = vars[j];
		}
		else
		{
			inst->b = j;
		}

		if ((j = get_int(fp)) < 0)
			return -1;
		inst->c = j;
	}

	if (fn->code[n-1].op)
		return -1;

	return 0;
}

static void save_function(FILE *fp, function *fn)
{
	unsigned i;

	put_int(fp, fn->name->idx);
	put_int(fp, fn->nparam);
	put_int(fp, fn->nvar);

	/* constants */

	put_int(fp, fn->nconsts);

	for (i = 0; i < fn->nconsts; i++)
	{
		constant *cst = &fn->consts[i];
		unsigned c = cst->type;

		fputc(c, fp);

		switch(c)
		{
		case T_FLOAT:
			put_float(fp, cst->u.flt);
			break;

		case T_VECTOR:
			put_float(fp, cst->u.v.x);
			put_float(fp, cst->u.v.y);
			put_float(fp, cst->u.v.z);
			break;

		case T_STRING:
		case T_FUNCTION:
			put_int(fp, cst->u.idx);
			break;
		}
	}

	/* code */

	put_int(fp, fn->ncode);

	for (i = 0; i < fn->ncode; i++)
	{
		instruction *inst = &fn->code[i];

		fputc(inst->op, fp);
		put_int(fp, inst->a);
		put_int(fp, inst->b);
		put_int(fp, inst->c);
	}
}
