%{

/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>

#include "makemodel.h"

void yyerror(const char *);
int yylex(void);

static bone *this_bone = 0;
static model this_model = { 1, 0, 0, 0, 0.0 };

%}

%union {
	unsigned idx;
	double flt;
	char *str;
	vector *vec;
	triangle *tri;
	vertex_reference *vtr;
	vertex *vtx;
	frame *frm;
	bone *bon;
}

%start model

%token <idx> INTEGER
%token <flt> NUMBER
%token <str> ID
%token <str> STRING

%token BONE
%token FRAME
%token POSITION
%token PRINT
%token TEXTURE
%token TRIANGLES
%token VERTICES

%type <flt> number
%type <vec> vector
%type <vec> angles
%type <tri> triangles
%type <tri> triangle
%type <vtr> vertex_reference
%type <vtx> vertex
%type <vtx> vertices
%type <flt> number
%type <frm> frame
%type <bon> bone
%type <bon> bone_start

%%

model : model_item | model model_item;

model_item: TEXTURE STRING ';'
{
	if (this_model.tex)
		yyerror("texture already defined");
	this_model.tex = $2;
}
	| bone 
{
	$1->next = this_model.bones;
	this_model.bones = $1;
}
	| TRIANGLES '{' triangles '}'
{
	if (this_model.tris)
		yyerror("triangles already defined");
	this_model.tris = $3;
}
;

bone : bone_start '{' bone_items '}' { $$ = $1; this_bone = $1->parent; }
;

bone_start : BONE INTEGER
{ 
	$$ = malloc(sizeof(bone));
	$$->next = 0;
	$$->parent = this_bone;
	$$->children = 0;
	$$->idx = $2;
	$$->pos = 0;
	$$->frames = 0;
	$$->verts = 0;
	this_bone = $$;
}
;

bone_items : bone_item bone_items |
;

bone_item : POSITION vector ';'
{
	if (this_bone->pos)
		yyerror("bone position already defined");
	this_bone->pos = $2;
}
	| VERTICES '{' vertices '}'
{
	if (this_bone->verts)
		yyerror("bone vertices already defined");
	this_bone->verts = $3;
}
	| frame
{
	$1->next = this_bone->frames;
	this_bone->frames = $1;
}
	| bone
{
	$1->next = this_bone->children;
	this_bone->children = $1;
}
;

frame : FRAME INTEGER angles ';'
{
	$$ = malloc(sizeof(frame));
	$$->next = 0;
	$$->idx = $2;
	$$->ang = $3;
	$$->pos = 0;
	$$->prt = 0;

	if ($2 >= this_model.frames)
		this_model.frames = $2 + 1;
}
	| FRAME INTEGER angles vector ';'
{
	$$ = malloc(sizeof(frame));
	$$->next = 0;
	$$->idx = $2;
	$$->ang = $3;
	$$->pos = $4;
	$$->prt = 0;

	if ($2 >= this_model.frames)
		this_model.frames = $2 + 1;
}
	| FRAME INTEGER angles PRINT ';'
{
	$$ = malloc(sizeof(frame));
	$$->next = 0;
	$$->idx = $2;
	$$->ang = $3;
	$$->pos = 0;
	$$->prt = 1;

	if ($2 >= this_model.frames)
		this_model.frames = $2 + 1;
}
;

angles: vector
{
	$$ = $1;
	$$->x /= 256.0;
	$$->y /= 256.0;
	$$->z /= 256.0;
}
;

vertices : vertex { $$ = $1; }
	| vertex vertices { ($$ = $1)->next = $2; }
;

vertex: vector vector ';'
{
	$$ = malloc(sizeof(vertex));
	$$->next = 0;
	$$->bone = this_bone;
	$$->vec = $1;
	$$->norm = $2;
}
;

triangles : triangle { $$ = $1; }
	| triangle triangles { ($$ = $1)->next = $2; }
;

triangle : vertex_reference ',' vertex_reference ',' vertex_reference ';'
{
	$$ = malloc(sizeof(triangle));
	$$->next = 0;
	$$->vtx[0] = $1;
	$$->vtx[1] = $3;
	$$->vtx[2] = $5;
}
;

vertex_reference : INTEGER INTEGER number number
{
	$$ = malloc(sizeof(vertex_reference));
	if (!($$->vtx = find_vertex(this_model.bones, $1, $2)))
	{
		yyerror("can't find vertex");
		exit(1);
	}
	$$->u = $3;
	$$->v = $4;
}
;

vector : '(' number number number ')'
{ 
	$$ = new_vector($2, $3, $4);
}
;

number : NUMBER { $$ = $1; }
	| INTEGER { $$ = (double)$1; }
;

%%

#define YY_NO_UNPUT
#include "makemodel.yy.c"

static const char *progname;

int main(int argc, char **argv)
{
	FILE *fp;

	progname = argv[0];

	if (argc != 3)
	{
		fprintf(stderr, "Usage: %s <input> <output>\n", progname);
		return 1;
	}

	if ( !(yyin = fopen(argv[1], "r")) )
	{
		perror(argv[1]);
		return 1;
	}

	if (yyparse())
	{
		fclose(yyin);
		return 1;
	}

	if (fclose(yyin))
	{
		perror(argv[1]);
		return 1;
	}

	calc_model(&this_model);

	if ( !(fp = fopen(argv[2], "w")) )
	{
		perror(argv[2]);
		return 1;
	}

	if (save_model(&this_model, fp))
	{
		perror(argv[2]);
		fclose(fp);
		return 1;
	}

	if (fclose(fp))
	{
		perror(argv[2]);		
	}

	return 0;
}

void yyerror(const char *msg)
{
	fprintf(stderr, "%s: %s\n", progname, msg);
}
