/*
 *  Copyright (C) 2000 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <malloc.h>

#include "makemodel.h"

vertex *find_vertex(bone *bp, unsigned bi, unsigned vi)
{
	vertex *vp;

	while (bp)
	{
		if (bp->idx == bi)
		{
			unsigned i = 0;
			for (vp = bp->verts; vp ; vp = vp->next)
			{
				if (vi == i)
					return vp;
				i++;
			}
		}

		if ( bp->children &&
		     (vp = find_vertex(bp->children, bi, vi)) )
		{
			return vp;
		}

		bp = bp->next;
	}

	return 0;
}

vector *new_vector(float x, float y, float z)
{
	vector *v = malloc(sizeof(vector));

	v->x = x;
	v->y = y;
	v->z = z;
		       
	return v;
}

void calc_model(model *mdl)
{
	bone *b;
	unsigned i;
	
	for (b = mdl->bones; b; b = b->next)
	{
		fix_frames(b, mdl->frames);
	}

	for (i = 0; i < mdl->frames; i++)
	{
		for (b = mdl->bones; b; b = b->next)
		{
			bone *bp;

			if ( !(bp = find_fixed_bone(b, i)) )
			{
				frame *frm = find_frame(bp = b, i);
				vector *v;
				float x, y, z;

				if ( (v = bp->pos) )
				{
					x = v->x;
					y = v->y;
					z = v->z;
				}
				else
				{
					 x = y = z = 0.0;
				}

				frm->pos = new_vector(x, y, z);
			}

			calc_bones(bp, 0, i);
		}
	}
	
	for (b = mdl->bones; b; b = b->next)
	{
		calc_radius(mdl, b);
	}
}

void calc_bones(bone *bp, bone *last, unsigned fi)
{
	frame *frm = find_frame(bp, fi);
	vector *vec = frm->pos;
	frame *f;
	bone *b;
	matrix m;
	vector v;

	if (frm->prt)
	{
		printf("B%u F%u (%.8f %.8f %.8f)\n",
		       bp->idx, frm->idx, vec->x, vec->y, vec->z);
	}

	if ( (b = bp->parent) && (b != last) )
	{
		f = find_frame(b, fi);

		if (!f->pos)
		{
			if (bp->pos)
			{
				set_matrix(&m, f->ang);
				rotate_vector(&v, bp->pos, &m);
				v.x = vec->x - v.x;
				v.y = vec->y - v.y;
				v.z = vec->z - v.z;
			}
			else
			{
				v.x = vec->x;
				v.y = vec->y;
				v.z = vec->z;
			}

			/* round_vector(&v); */
			f->pos = new_vector(v.x, v.y, v.z);
		}

		calc_bones(b, bp, fi);
	}

	set_matrix(&m, frm->ang);

	for (b = bp->children; b; b = b->next)
	{
		if (b == last)
			continue;

		f = find_frame(b, fi);

		if (!f->pos)
		{
			if (b->pos)
			{
				rotate_vector(&v, b->pos, &m);
				v.x += vec->x;
				v.y += vec->y;
				v.z += vec->z;
			}
			else
			{
				v.x = vec->x;
				v.y = vec->y;
				v.z = vec->z;
			}

			/* round_vector(&v); */
			f->pos = new_vector(v.x, v.y, v.z);
		}

		calc_bones(b, bp, fi);
	}
}

void calc_radius(model *mdl, bone *bp)
{
	frame *frm;
	matrix m;
	vector v;

	for (frm = bp->frames; frm; frm = frm->next)
	{
		vertex *vtx;
		vector *vec = frm->pos;

		set_matrix(&m, frm->ang);

		for (vtx = bp->verts; vtx; vtx = vtx->next)
		{
			float r;

			rotate_vector(&v, vtx->vec, &m);
			v.x += vec->x;
			v.y += vec->y;
			v.z += vec->z;

			r = sqrt((v.x * v.x) + (v.y * v.y) + (v.z * v.z));

			if (mdl->radius < r)
				mdl->radius = r;
		}
	}

	for (bp = bp->children; bp; bp = bp->next)
	{
		calc_radius(mdl, bp);
	}
}

void fix_frames(bone *bp, unsigned nf)
{
	frame *frames = bp->frames;
	frame **next = &bp->frames;
	unsigned i;

	*next = 0;

	for (i = 0; i < nf; i++)
	{
		frame *frm, **ptr;

		for (ptr = &frames; (frm = *ptr); ptr = &frm->next)
		{
			if (frm->idx == i)
			{
				*ptr = frm->next;
				goto found;
			}
		}

		frm = malloc(sizeof(frame));

		frm->idx = i;
		frm->ang = new_vector(0.0, 0.0, 0.0);
		frm->pos = 0;
		frm->prt = 0;

	found:
		frm->next = 0;

		*next = frm;
		next = &frm->next;
	}

	for (bp = bp->children; bp; bp = bp->next)
	{
		fix_frames(bp, nf);
	}
}

bone *find_fixed_bone(bone *bp, unsigned fi)
{
	frame *frm = find_frame(bp, fi);

	if (frm->pos)
		return bp;

	for (bp = bp->children; bp; bp = bp->next)
	{
		bone *b;

		if ( (b = find_fixed_bone(bp, fi)) )
			return b;
	}

	return 0;
}

frame *find_frame(bone *bp, unsigned fi)
{
	frame *frm;

	for (frm = bp->frames; frm; frm = frm->next)
	{
		if (frm->idx == fi)
			return frm;
	}

	return 0;
}

void set_matrix(matrix *m, vector *a)
{
	m->x.x = 1.0; m->x.y = 0.0; m->x.z = 0.0;
	m->y.x = 0.0; m->y.y = 1.0; m->y.z = 0.0;
	m->z.x = 0.0; m->z.y = 0.0; m->z.z = 1.0;

	rotate_matrix_z(m, a->z);
	rotate_matrix_y(m, a->y);
	rotate_matrix_x(m, a->x);
}

void rotate_matrix_x(matrix *m, float a)
{
	float s = sin(a * M_PI);
	float c = cos(a * M_PI);
	float i, j;

	i = m->y.x; j = m->z.x;
	m->y.x = (i * c) - (j * s);
	m->z.x = (i * s) + (j * c);
	i = m->y.y; j = m->z.y;
	m->y.y = (i * c) - (j * s);
	m->z.y = (i * s) + (j * c);
	i = m->y.z; j = m->z.z;
	m->y.z = (i * c) - (j * s);
	m->z.z = (i * s) + (j * c);
}

void rotate_matrix_y(matrix *m, float a)
{
	float s = sin(a * M_PI);
	float c = cos(a * M_PI);
	float i, j;

	i = m->z.x; j = m->x.x;
	m->z.x = (i * c) - (j * s);
	m->x.x = (i * s) + (j * c);
	i = m->z.y; j = m->x.y;
	m->z.y = (i * c) - (j * s);
	m->x.y = (i * s) + (j * c);
	i = m->z.z; j = m->x.z;
	m->z.z = (i * c) - (j * s);
	m->x.z = (i * s) + (j * c);
}

void rotate_matrix_z(matrix *m, float a)
{
	float s = sin(a * M_PI);
	float c = cos(a * M_PI);
	float i, j;

	i = m->x.x; j = m->y.x;
	m->x.x = (i * c) - (j * s);
	m->y.x = (i * s) + (j * c);
	i = m->x.y; j = m->y.y;
	m->x.y = (i * c) - (j * s);
	m->y.y = (i * s) + (j * c);
	i = m->x.z; j = m->y.z;
	m->x.z = (i * c) - (j * s);
	m->y.z = (i * s) + (j * c);
}

void rotate_vector(vector *p, vector *v, matrix *m)
{
	p->x = (m->x.x * v->x) + (m->y.x * v->y) + (m->z.x * v->z);
	p->y = (m->x.y * v->x) + (m->y.y * v->y) + (m->z.y * v->z);
	p->z = (m->x.z * v->x) + (m->y.z * v->y) + (m->z.z * v->z);
}

/*
void round_vector(vector *v)
{
	v->x = rint(v->x * 256.0) / 256.0;
	v->y = rint(v->y * 256.0) / 256.0;
	v->z = rint(v->z * 256.0) / 256.0;
}
*/
