#include <stdio.h>
#include <string.h>

class triangle;
class landscape;

class vector
{
 public:
	int x, y, z;
};

class vertex : public vector
{
 public:
	vertex *next;
	int u, v;
	int idx;

	int load(FILE *, vector *);
};

class vertices
{
 private:
	vertex *vts;
	unsigned num;

 public:
	vertices();

	void add(vertex *);

	int save(FILE *);
};

class triangle
{
 public:
	triangle *next;

 private:
	vertex vtx[3];

 public:
	triangle();
	~triangle();

	int load(FILE *, vector *);

	int save(FILE *);

	void add(vertices *v);
};

class texture
{
 private:
	char *name;
	triangle *tris;

 public:
	texture();
	~texture();

	int load(FILE *);
	int save(FILE *);

	inline triangle *new_triangle()
	{
		triangle *t = new triangle;
		t->next = tris;
		return (tris = t);
	}
};

class landscape
{
 private:
	unsigned ntex;
	texture *texs;

 public:
	landscape();
	~landscape();

	int load(const char *);
	int save(const char *);

	int load_square(FILE *, unsigned, unsigned);
};

int main(int argc, char **argv)
{
	landscape map;

	if (argc != 3)
	{
		fprintf(stderr, "Usgae: %s <src> <dst>\n", argv[0]);
		return 1;
	}

	if (map.load(argv[1]))
	{
		fprintf(stderr, "Failed to load map \"%s\"\n", argv[1]);
		return 1;
	}

	if (map.save(argv[2]))
	{
		fprintf(stderr, "Failed to save map \"%s\"\n", argv[2]);
		return 1;
	}

	return 0;
}

landscape::landscape()
{
	ntex = 0;
	texs = 0;
}

landscape::~landscape()
{
	if (texs) delete [] texs;
}

int landscape::load(const char *file)
{
	FILE *fp;
	unsigned i, j;
	char magic[4];

	if ( !(fp = fopen(file, "r")) )
		return -1;

	if ( fread(magic, 1, 4, fp) != 4 ||
	     strncmp(magic, "CJPL", 4) )
	{
		fclose(fp);
		return -1;
	}

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	if (texs) delete [] texs;

	texs = new texture[(ntex = i) + 1];

	for (i = 1; i <= ntex; i++)
	{
		texs[i].load(fp);
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (load_square(fp, i * 16, j * 16))
				return -1;
		}
	}

	if (fclose(fp))
		return -1;

	return 0;
}

int landscape::load_square(FILE *fp, unsigned x, unsigned y)
{
	unsigned i, j, n, z;
	triangle *t;
	vector v;

	if ( (z = fgetc(fp)) > 255)
		return -1;

	if ( (n = fgetc(fp)) > 255)
		return -1;

	v.x = x;
	v.y = y;
	v.z = z;

	for (i = 0; i < n; i++)
	{
		if ( (j = fgetc(fp)) > ntex )
		{
			return -1;
		}

		t = texs[j].new_triangle();

		if (t->load(fp, &v))
			return -1;
	}

	return 0;
}

int landscape::save(const char *file)
{
	FILE *fp;
	unsigned i;

	if ( !(fp = fopen(file, "w")) )
		return -1;

	fprintf(fp, "AC3Db\n");

	fprintf(fp, "OBJECT world\n");

	fprintf(fp, "kids %u\n", ntex);

	for (i = 1; i <= ntex; i++)
	{
		if (texs[i].save(fp))
			return -1;
	}

	if (fclose(fp))
		return -1;

	return 0;
}


texture::texture()
{
	name = 0;
	tris = 0;
}

texture::~texture()
{
	triangle *t;

	if (name) delete name;

	while ((t = tris))
	{
		tris = t->next;
		delete t;
	}
}

int texture::load(FILE *fp)
{
	unsigned i;

	if ( (i = fgetc(fp)) > 255 )
		return -1;

	name = new char[i + 1];

	if (i)
	{
		if (fread(name, 1, i, fp) != i)
			return -1;
	}
	name[i] = '\0';

	return 0;
}

int texture::save(FILE *fp)
{
	triangle *t;
	unsigned n = 0;
	vertices vts;

	fprintf(fp, "OBJECT poly\n");

	fprintf(fp, "texture \"%s\"\n", name);

	for (t = tris; t; t = t->next)
	{
		n++;
	}

	for (t = tris; t; t = t->next)
	{
		t->add(&vts);
	}

	if (vts.save(fp))
		return -1;

	fprintf(fp, "numsurf %u\n", n);

	n = 0;

	for (t = tris; t; t = t->next)
	{
		if (t->save(fp))
			return -1;
		n += 3;
	}

	fprintf(fp, "kids 0\n");

	return 0;
}

vertices::vertices()
{
	vts = 0;
	num = 0;
}

void vertices::add(vertex *vtx)
{
	vertex *v, **p;

	for (p = &vts; (v = *p); p = &v->next)
	{
		if (v->x == vtx->x &&
		    v->y == vtx->y &&
		    v->z == vtx->z)
		{
			vtx->idx = v->idx;
			return;
		}
	}

	vtx->idx = num++;
	vtx->next = 0;
	*p = vtx;
}

int vertices::save(FILE *fp)
{
	vertex *v;

	fprintf(fp, "numvert %u\n", num);

	for (v = vts; v; v = v->next)
	{
		fprintf(fp, "%.4f %.4f %.4f\n",
			(float(v->x) / 16.0) - 128.0,
			(float(v->z) / 16.0),
			128.0 - (float(v->y) / 16.0));
	}

	return 0;
}

triangle::triangle()
{
	next = 0;
}

triangle::~triangle()
{
}

int triangle::load(FILE *fp, vector *v)
{
	unsigned i;

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].load(fp, v))
			return -1;
	}

	return 0;
}

int triangle::save(FILE *fp)
{
	unsigned i;

	fprintf(fp, "SURF 0x00\n");
	fprintf(fp, "refs 3\n");

	for (i = 0; i < 3; i++)
	{
		fprintf(fp, "%u %.6f %.6f\n", vtx[i].idx,
			float(vtx[i].u) / 64.0,
			float(vtx[i].v) / 64.0);
	}

	return 0;
}

void triangle::add(vertices *vts)
{
	unsigned i;

	for (i = 0; i < 3; i++)
	{
		vts->add(&vtx[i]);
	}
}

int vertex::load(FILE *fp, vector *o)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	if (fread(&data, 1, 5, fp) != 5)
		return -1;

	x = data.x + o->x;
	y = data.y + o->y;
	z = data.z + o->z;
	u = data.u;
	v = data.v;

	return 0;
}
