#include <stdio.h>
#include <string.h>

class triangle;
class landscape;

class vector
{
 public:
	int x, y, z;
};

class texture
{
 public:
	texture *next;
	unsigned index;

 private:
	char *name;

 public:
	texture(const char *);
	~texture();

	int save(FILE *);

	inline const char *filename() const { return name; }
};

class vertex : public vector
{
 private:
	int u, v;

 public:
	int load(FILE *, int);
	int save(FILE *, int);
};

class triangle
{
 public:
	triangle *next;

 private:
	vertex vtx[3];
	texture *tex;

 public:
	triangle();
	~triangle();

	int load(FILE *, texture **, int);
	int save(FILE *, int);

	void min_max(int *, int *);
};

class square 
{
 private:
	triangle *tris;
	int z;

 public:
	square();
	~square();

	int load(FILE *, texture **);
	int save(FILE *);

	void fix_heights();
};

class landscape
{
 private:
	texture *tex;
	square sqr[256][256];

 public:
	landscape();
	~landscape();

	int load(const char *);
	int save(const char *);

	texture *load_texture(FILE *);

	void fix_heights();
};

int main(int argc, char **argv)
{
	landscape map;
	int i, c;

	if (argc < 3)
	{
		fprintf(stderr, "Usgae: %s <src> ... <dst>\n", argv[0]);
		return 1;
	}

	c = argc - 1;

	for (i = 1; i < c; i++)
	{
		if (map.load(argv[i]))
		{
			fprintf(stderr, "Failed to load map \"%s\"\n",
				argv[i]);
			return 1;
		}
	}

	map.fix_heights();

	if (map.save(argv[c]))
	{
		fprintf(stderr, "Failed to save map \"%s\"\n", argv[c]);
		return 1;
	}

	return 0;
}

landscape::landscape()
{
	tex = 0;
}

landscape::~landscape()
{
	while (tex)
	{
		texture *t = tex;
		tex = t->next;
		delete t;
	}
}

int landscape::load(const char *file)
{
	FILE *fp;
	unsigned i, j;
	texture *texs[256];
	char magic[4];

	if ( !(fp = fopen(file, "r")) )
		return -1;

	if ( fread(magic, 1, 4, fp) != 4 ||
	     strncmp(magic, "CJPL", 4) )
	{
		fclose(fp);
		return -1;
	}

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	for (j = 1; j <= i; j++)
	{
		if ( !(texs[j] = load_texture(fp)) )
		{
			fclose(fp);
			return -1;
		}
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (sqr[j][i].load(fp, texs))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	if (fclose(fp))
		return -1;

	return 0;
}

int landscape::save(const char *file)
{
	FILE *fp;
	unsigned i, j;
	texture *t;

	if ( !(fp = fopen(file, "w")) )
		return -1;

	if (fwrite("CJPL", 1, 4, fp) != 4)
	{
		fclose(fp);
		return -1;
	}

	i = 0;
	for (t = tex; t; t = t->next)
	{
		t->index = ++i;
	}

	if (fputc(i, fp) != int(i))
	{
		fclose(fp);
		return -1;
	}

	for (t = tex; t; t = t->next)
	{
		if (t->save(fp))
		{
			fclose(fp);
			return -1;
		}
	}

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			if (sqr[j][i].save(fp))
			{
				fclose(fp);
				return -1;
			}
		}
	}

	if (fclose(fp))
		return -1;

	return 0;
}

texture *landscape::load_texture(FILE *fp)
{
	unsigned i;
	char name[256];
	texture *t, **p;

	if ( (i = fgetc(fp)) > 255 )
		return 0;
	if (i)
	{
		if (fread(name, 1, i, fp) != i)
			return 0;
	}
	name[i] = '\0';

	for (p = &tex; (t = *p); p = &t->next)
	{
		if ( !strcmp(t->filename(), name) )
		{
			return t;
		}
	}

	return *p = new texture(name);
}

void landscape::fix_heights()
{
	unsigned i, j;

	for (j = 0; j < 256; j++)
	{
		for (i = 0; i < 256; i++)
		{
			sqr[j][i].fix_heights();
		}
	}
}

square::square()
{
	tris = 0;
	z = 0;
}

square::~square()
{
	while (tris)
	{
		triangle *t = tris;
		tris = t->next;
		delete t;
	}
}

int square::load(FILE *fp, texture **texs)
{
	unsigned i, j, k;
	triangle *t, **p;

	if ( (k = fgetc(fp)) > 255)
		return -1;

	if (!tris) z = k;

	if ( (j = fgetc(fp)) > 255)
		return -1;

	p = &tris; while ((t = *p)) p = &t->next;

	for (i = 0; i < j; i++)
	{
		t = new triangle();
		t->load(fp, texs, k);
		*p = t;
		p = &t->next;
	}

	return 0;
}

int square::save(FILE *fp)
{
	unsigned i;
	triangle *t;

	if (fputc(z, fp) != z)
                return -1;

	i = 0;
	for (t = tris; t; t = t->next)
	{
		++i;
	}

	if (fputc(i, fp) != int(i))
                return -1;

	for (t = tris; t; t = t->next)
	{
		if (t->save(fp, z))
			return -1;
	}

	return 0;
}

void square::fix_heights()
{
	triangle *t;
	int min = z;
	int max = z;

	for (t = tris; t; t = t->next)
	{
		t->min_max(&min, &max);
	}

	if ((max - z) > 127) z = max - 127;
	if ((min - z) < -128) z = min + 128;
}

triangle::triangle()
{
	next = 0;
	tex = 0;
}

triangle::~triangle()
{
}

int triangle::load(FILE *fp , texture **texs, int z)
{
	unsigned i;

	if ( (i = fgetc(fp)) > 255 )
	{
		fclose(fp);
		return -1;
	}

	tex = texs[i];

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].load(fp, z))
			return -1;
	}

	return 0;
}

int triangle::save(FILE *fp, int z)
{
	unsigned i;

	if (fputc(tex->index, fp) != int(tex->index))
		return -1;

	for (i = 0; i < 3; i++)
	{
		if (vtx[i].save(fp, z))
			return -1;
	}

	return 0;
}

void triangle::min_max(int *minp, int *maxp)
{
	int min = *minp;
	int max = *maxp;
	unsigned i;

	for (i = 0; i < 3; i++)
	{
		int z = vtx[i].z;

		if (z < min) min = z;
		if (z > max) max = z;
	}

	*minp = min;
	*maxp = max;
}

int vertex::load(FILE *fp, int h)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	if (fread(&data, 1, 5, fp) != 5)
		return -1;

	x = data.x;
	y = data.y;
	z = data.z + h;
	u = data.u;
	v = data.v;

	return 0;
}

int vertex::save(FILE *fp, int h)
{
	struct {
		signed char x, y, z;
		unsigned char u, v;
	} data;

	data.x = x;
	data.y = y;
	data.z = z - h;
	data.u = u;
	data.v = v;

	if (fwrite(&data, 1, 5, fp) != 5)
		return -1;

	return 0;
}

texture::texture(const char *s)
{
	next = 0;
	index = 0;
	name = new char[strlen(s) + 1];
	strcpy(name, s);
}

texture::~texture()
{
	delete name;
}

int texture::save(FILE *fp)
{
	unsigned n = strlen(name);

	if (fputc(n, fp) != int(n))
		return -1;

	if (fwrite(name, 1, n, fp) != n)
		return -1;

	return 0;
}
